/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2008 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/*******************************************************
* non-smooth coefficients elliptic PDE
* 
* -div(a(x) grad u) = f  x \in \Omega
* u = u0  x \in \partial \Omega
*
* a(x): piesewise constant  a(x) > \alpha > 0
* a(x) = 1 in first and third quadrants of x-y
* a(x) = R in second and fourth quadrants of x-y
* ture solution is func_g
*
* $Id: non-smooth.c,v 1.29 2008/01/18 06:18:23 liuhui Exp $
*******************************************************/

#include "phg.h"
#include <math.h>

/* parameters used by function func_g and func_ax*/
static FLOAT R = 161.44763879;

static void   /* rhs */
func_f(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
{
    FLOAT r, ee;

    r = x * x + y * y + z * z;
    ee = -6.0 * Exp(r) - 4 * r * Exp(r);

    if ((x > 0.0 && y > 0.0) || (x < 0.0 && y < 0.0)){
        *value = R * ee;
    }
    else {
        *value = ee;
    }
}

static void  /* boundary function */
func_g(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
{
    *value = Exp(x * x + y * y + z * z);
}

static void /* coefficent a(x)*/
func_ax(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
{
    if ((x > 0.0 && y > 0.0) || (x < 0.0 && y < 0.0)){
        *value = R;
    }
    else {
        *value = 1.0;
    }
}

static void /* coefficent a(x), 3 dim */
func_ax_3(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
{
    if ((x > 0.0 && y > 0.0) || (x < 0.0 && y < 0.0)){
        *(value++) = R;
        *(value++) = R;
        *(value++) = R;
    }
    else {
        *(value++) = 1.0;
        *(value++) = 1.0;
        *(value++) = 1.0;
    }
}

static FLOAT  /* calculate time */
cal_time(BOOLEAN flag)
{
    static FLOAT t0 = 0.0;
    double et, tt[3];

    phgGetTime(tt);
    et = tt[2] - t0;
    t0 = tt[2];

    if (flag){
        phgPrintf("  %0.4lfs", (double)et);
    }

    return et;
}

/* cal energy norm */
/* (\int a(x) \grad u dot \grad u dx */
static FLOAT 
phgDofNormEnergy(DOF *u, DOF *a)
{
    GRID *g = u->g;
    SIMPLEX *e;
    DOF *grad;
    FLOAT sum = 0.0;
    FLOAT tmp = 0.0;

    grad = phgDofGradient(u, NULL, NULL, "u grad");

    ForAllElements(g, e){
        tmp = *DofElementData(a, e->index);
        sum += tmp * phgQuadDofDotDof(e, grad, grad, -1);
    }
#if USE_MPI
    MPI_Allreduce(&sum, &tmp, 1, PHG_MPI_FLOAT, MPI_SUM, g->comm);
    sum = tmp;
#endif
    phgDofFree(&grad);
    return Sqrt(sum);
}

/* build linear systme */
static void
build_linear_system(SOLVER *solver, DOF *u_h, DOF *f_h, DOF *a)
{
    int N = u_h->type->nbas;    /* number of basis functions in an element */
    GRID *g = u_h->g;
    SIMPLEX *e;
    int i, j;
    FLOAT A[N][N], B[N], buffer[N], *row, tmp;
    INT I[N];

    assert(u_h->dim == 1);

    ForAllElements(g, e) {
    /* compute \int a(x) \grad\phi_i \cdot \grad\phi_j making use of symmetry */
        tmp = *DofElementData(a, e->index);
        for (i = 0; i < N; i++) {
            I[i] = phgSolverMapE2L(solver, 0, e, i);
            for (j = 0; j <= i; j++)
                A[j][i] = A[i][j] =
                tmp * phgQuadGradBasDotGradBas(e, u_h, i, u_h, j, QUAD_DEFAULT);
        }

        /* loop on basis functions */
        for (i = 0; i < N; i++) {
            if (phgDofDirichletBC(u_h, e, i, func_g,
                  buffer, B + i, DOF_PROJ_NONE)) {
                row = buffer;
            }
            else {
                /* right hand side: \int f * phi_i */
                phgQuadDofTimesBas(e, f_h, u_h, i, QUAD_DEFAULT, B + i);
                row = A[i];
            }
            phgSolverAddMatrixEntries(solver, 1, I + i, N, I, row); 
        }
        phgSolverAddRHSEntries(solver, N, I, B);
    }
}

static void  /* solve Au = b */
SolveEqua(OEM_SOLVER *sol, DOF *u_h, DOF *f_h, DOF *ax)
{
    GRID *g = u_h->g;
    SOLVER *solver;
    int nits;
    phgPrintf("DOF: %d ", DofGetDataCountGlobal(u_h));
    phgPrintf("Elements: %d  ", g->nleaf_global);
    phgPrintf("Load imbalance: %lg\n",  (double)g->lif);
                       
    solver = phgSolverCreate(sol, u_h, NULL);
    phgPrintf("build linear system: ");
    cal_time(FALSE);
    build_linear_system(solver, u_h, f_h, ax);
    cal_time(TRUE);
              
    phgPrintf("   solve equation: ");
    cal_time(FALSE);
    nits = phgSolverSolve(solver, TRUE, u_h, NULL);
    cal_time(TRUE);
    phgPrintf(" [%d]\n", nits);
    phgSolverDestroy(&solver);

}

static FLOAT /* calulate space indicator */
estimate_space(DOF *u_h, DOF *f_h, DOF *a, DOF *error)
{
    GRID *g = u_h->g;
    SIMPLEX *e;
    NEIGHBOUR_DATA *aa;
    DOF *jump, *grad, *agrad, *div, *ax3, *temp;
    FLOAT sum, tmp;
    FLOAT h, ak, ae;
    int i;

    ax3 = phgDofNew(g, DOF_P0, 3, "ax3", func_ax_3);
    grad = phgDofGradient(u_h, NULL, NULL, "grad");
    agrad = phgDofMM(MAT_OP_N, MAT_OP_N, 3, 1, 3, 1.0, ax3, 1, grad, 0.0, NULL);
    temp = phgDofDivergence(agrad, NULL, DOF_DGn[agrad->type->order], "div");
    div = phgDofCopy(f_h, NULL, DOF_DGn[u_h->type->order], NULL);
    phgDofAXPY(1.0, temp, &div);
    phgDofFree(&temp);
    jump = phgQuadFaceJump(agrad, DOF_PROJ_DOT, "jump", -1);
    phgDofFree(&grad);

    phgDofSetDataByValue(error, 0.0);
    aa = phgDofGetNeighbourData(a, NULL);  
    ForAllElements(g, e){ /* neighbours are local */
        sum = 0.0;
        tmp = *DofElementData(a, e->index);
        h = phgGeomGetDiameter(g, e);
        sum = h * h / tmp * phgQuadDofDotDof(e, div, div, -1);
        for (i = 0; i < NFace; i++) {
            if (e->bound_type[i] & (DIRICHLET | NEUMANN))
                continue;    /* boundary face */
            ak = *phgDofNeighbourData(aa, e, i, 0, NULL);
            ae = (tmp > ak) ? tmp : ak;
            h = phgGeomGetFaceDiameter(g, e, i);
            sum +=  (*DofFaceData(jump, e->faces[i])) * h * 0.5 / ae;
        }
        *DofElementData(error, e->index) = sum;
    }
    
    phgDofFree(&div);
    phgDofFree(&jump);
    phgDofFree(&ax3);
    phgDofFree(&agrad);
    phgDofReleaseNeighbourData(&aa);

    return phgDofNormInftyVec(error);

}

static FLOAT
estimate_osc(DOF *f_h, DOF *f_t, DOF *osc)
{
    GRID *g = f_h->g;
    SIMPLEX *e;
    FLOAT dd;

    ForAllElements(g, e) {
        dd = 0;
        dd += phgQuadDofDotDof(e, f_h, f_h, 4);
        dd += phgQuadDofDotDof(e, f_t, f_t, 4);
        dd -= 2 * phgQuadDofDotDof(e, f_h, f_t, 4);
        *DofElementData(osc, e->index) = Fabs(dd);
    }
    return  phgDofNormInftyVec(osc);
}

int
main(int argc, char *argv[])
{
    GRID *g;
    DOF *u_h, *u_t, *point = NULL;
    DOF *f_h, *f_t;
    DOF *ax, *error, *osc;
    FLOAT etaspace, etamax, tmp, tp, tmp1, oscmax;
    double tt1[3], tt2[3], tt3[3];
    size_t peakmem;
    static char *fn = "../test/non-smooth.dat";
    static FLOAT tol = 0.1;
    int flag = 0;
    static BOOLEAN export_mesh = FALSE;
    FLOAT t1[3], t2[3];
    
    phgOptionsRegisterFilename("mesh_file", "Mesh file", (char **)&fn);
    phgOptionsRegisterFloat("tol", "Convergence criterion", &tol);
    phgOptionsRegisterFloat("ns", "non-smooth parameter( > 0)", &R);
    phgOptionsRegisterNoArg("export_mesh", "Export mesh", &export_mesh);

    phgInit(&argc, &argv);
    phgGetTime(t1);
    g = phgNewGrid(-1);
    if (!phgImport(g, fn, TRUE))
        phgError(1, "can't read file \"%s\".\n", fn);

    if (R <= 0.)
        phgError(1, "non-smooth parameter ns <= 0.\n");

    u_h = phgDofNew(g, DOF_DEFAULT, 1, "u_h", DofInterpolation);
    u_t = phgDofNew(g, DOF_DEFAULT, 1, "u_t", func_g);
    point = phgDofNew(g, DOF_DEFAULT, 1, "point", DofNoAction);
    f_h = phgDofNew(g, DOF_DEFAULT, 1, "f_h", func_f);
    f_t = phgDofNew(g, DOF_ANALYTIC, 1, "f_t", func_f);
    ax = phgDofNew(g, DOF_P0, 1, "a(x)", func_ax);
    error = phgDofNew(g, DOF_P0, 1, "error indicator", DofNoAction);
    osc = phgDofNew(g, DOF_P0, 1, "data oscillation", DofNoAction);

#if 0
    /* distribute grid to all processes */
    phgRefineAllElements(g, 7);
#endif
    phgGetTime(tt1);
    while(TRUE){
        phgGetTime(tt2);
        flag++;
        phgPrintf("\n*** Pass %d\n", flag);
        if (phgBalanceGrid(g, 1.2, -1, NULL, 0.))
            phgPrintf("Repartition mesh, load imbalance: %lg\n",
                (double)g->lif);

        /* solve the linear system */
        SolveEqua(SOLVER_DEFAULT, u_h, f_h, ax);

        /* calculate error */
        etamax = estimate_space(u_h, f_h, ax, error);
        oscmax = estimate_osc(f_h, f_t, osc);
        etaspace = phgDofNormL1Vec(error);
        phgPrintf("Max error indicator:   %.6lg  Total error: %lg\n",
                        (double)Sqrt(etamax), (double)Sqrt(etaspace));

        phgDofCopy(u_t, &point, DOF_DEFAULT, NULL);
        phgDofAXPY(-1.0, u_h, &point);

        /* calculate L2 norm */
        tmp = phgDofNormL2(u_t);
        tp = phgDofNormL2(point);
        phgPrintf("True L2 error:         %.6lg   ", (double)tp);
        phgPrintf("relative error: %.6lg\n",
                                        (double)(tp / tmp));

        /* pointwise error */
        tp = phgDofNormInftyVec(u_t);
        tmp = phgDofNormInftyVec(point);
        phgPrintf("Max point error:       %.6lg   ", (double)tmp);
        phgPrintf("relative error: %.6lg\n",
                                        (double)(tmp / tp));

        /* calculate energy norm */
        tmp = phgDofNormEnergy(u_t, ax);
        tmp1 = phgDofNormEnergy(point, ax);
        phgPrintf("True energy error:     %.6lg   ", (double)tmp1);
        phgPrintf("relative error: %.6lg\n",
                                        (double)(tmp1 / tmp));
        
        phgMemoryUsage(g, &peakmem);
        if (Sqrt(etaspace) <= tol || peakmem >= 1024 * (size_t)400 * 1024){
            phgGetTime(tt3);
            phgPrintf("Time usage: %lfs\n", (double)(tt3[2] - tt2[2]));
            phgPrintf("Memory usage: %lfMb\n\n",
                                        (double)peakmem / (1024. * 1024.));
            break;
        }
        /* mark elements for refining */
        phgMarkRefine(MARK_DEFAULT, error, 0.8, osc, 0.8, 1, 2, tol);

        phgPrintf("Refine elements:");
        cal_time(FALSE);
        
        /* refine elements */
        phgRefineMarkedElements(g);
        cal_time(TRUE);
        phgGetTime(tt3);
        phgPrintf("\nTime usage: %lfs\n", (double)(tt3[2] - tt2[2]));
        phgMemoryUsage(g, &peakmem);
        phgPrintf("Memory usage: %lf Mb\n\n", (double)peakmem / (1024.*1024.));
    }

    phgPrintf("Total time: %lfs\n", (double)(tt3[2] - tt1[2]));
    if (export_mesh){
        DOF *grad;
        grad = phgDofGradient(u_h, NULL, NULL, "grad_u_h");
        phgPrintf("Final mesh written to \"%s\".\n",
                phgExportVTK(g, "non-smooth.vtk",
                    u_h, grad, ax, error, u_t, point, NULL));
        phgDofFree(&grad);
    }
    
    phgGetTime(t2);
    phgPrintf("total time: %f\n", t2[2] - t1[2]);
    phgDofFree(&u_t);
    phgDofFree(&u_h);
    phgDofFree(&point);
    phgDofFree(&ax);
    phgDofFree(&f_h);
    phgDofFree(&f_t);
    phgDofFree(&error);
    phgDofFree(&osc);

    phgFreeGrid(&g);
    phgFinalize();
    return 0;
}
