/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2008 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: dof.c,v 1.157 2007/12/25 12:08:02 zlb Exp $ */

#include "phg.h"
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <stdarg.h>

DOF_TYPE *DOF_DEFAULT = DOF_P2;

int
phgDofBasisOrder(DOF *u, int ibas)
{
    DOF_TYPE *type = u->type;
    BYTE *orders;
    int n, m;

    if ((orders = type->orders) == NULL || ibas < 0 || ibas >= type->nbas)
	return type->order;
    if (ibas < (n = NVert * type->np_vert))
	return orders[ibas % type->np_vert];
    orders += type->np_vert;
    if ((ibas -= n) < (m = NEdge * type->np_edge))
	return orders[ibas % type->np_edge];
    orders += type->np_edge;
    if ((ibas -= m) < (n = NFace * type->np_face))
	return orders[ibas % type->np_face];
    return orders[type->np_face + ibas - n];
}

/* Special DOFs */

DOF_TYPE DOF_CONSTANT_ = {DofCache, "Constant", NULL, NULL, &DOF_CONSTANT_,
    NULL, NULL, NULL, NULL, NULL, NULL,
    TRUE, FALSE, -1,		/* invariant, free_after_use, id */
    0, 0, 0, 1,			/* nbas, order, cont, dim */
    0, 0, 0, 0
};				/* np_vert, np_edge, np_face, np_elem */

DOF_TYPE DOF_ANALYTIC_ = {DofCache, "Analytic", NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL,
    FALSE, FALSE, -1,		/* invariant, free_after_use, id */
    0, -1, 0, 1,		/* nbas, order, cont, dim */
    0, 0, 0, 0
};				/* np_vert, np_edge, np_face, np_elem */

/* Dummy USER_FUNCs */

void
phgDofUserFuncNoData_(FLOAT x, FLOAT y, FLOAT z, FLOAT *values)
{
    Unused(x);
    Unused(y);
    Unused(z);
    Unused(values);
    phgError(1, "%s:%d, calling dummy DOF function.\n", __FILE__, __LINE__);
    return;
}

void
phgDofUserFuncNoAction_(FLOAT x, FLOAT y, FLOAT z, FLOAT *values)
{
    Unused(x);
    Unused(y);
    Unused(z);
    Unused(values);
    phgError(1, "%s:%d, calling dummy DOF function.\n", __FILE__, __LINE__);
    return;
}

void
phgDofUserFuncInterpolation_(FLOAT x, FLOAT y, FLOAT z, FLOAT *values)
{
    Unused(x);
    Unused(y);
    Unused(z);
    Unused(values);
    phgError(1, "%s:%d, calling dummy DOF function.\n", __FILE__, __LINE__);
    return;
}

void
phgDofUserFuncLambda_(FLOAT x, FLOAT y, FLOAT z, FLOAT *values)
{
    Unused(x);
    Unused(y);
    Unused(z);
    Unused(values);
    phgError(1, "%s:%d, calling dummy DOF function.\n", __FILE__, __LINE__);
    return;
}

/* list of referenced DOF_TYPES */
static struct {
    DOF_TYPE *type;
    size_t refcount;
} *type_list = NULL;
static size_t type_list_count = 0, type_list_allocated = 0;

DOF *
phgDofNew_(GRID *g, DOF_TYPE *type, SHORT dim, const char *name,
	   DOF_USER_FUNC userfunc, const char *srcfile, int srcline)
{
    DOF *dof;
    size_t len;

    FunctionEntry;

    if (phgMasterSlave)		/* Master/slave mode is not yet implemented */
	phgError(1, "phgDofNew does not allow master/slave mode\n");

    if (!phgGridIsValid(g))
	phgError(1, "%s:%d, invalid grid.\n", srcfile == NULL ? "" : srcfile,
			srcline);

    dof = phgCalloc(1, sizeof(*dof));
    dof->magic = MAGIC_DOF;
    if (name == NULL)
	name = "noname";
    phgInfo(2, "creating DOF \"%s\".\n", name);
    dof->name = phgAlloc(strlen(name) + 1);
    strcpy(dof->name, name);
    dof->type = type;
    dof->userfunc = userfunc;
    dof->userfunc_lambda = NULL;
    dof->invariant = FALSE;
    dof->DB_mask = DIRICHLET;
    if (type == DOF_CONSTANT && userfunc != DofNoAction
	&& userfunc != DofNoData) {
	phgPrintf("changing userfunc of constant DOF \"%s\" to DofNoAction\n",
		  dof->name);
	dof->userfunc = DofNoAction;
    }
    if (userfunc == DofInterpolation && type->InterpC2F == NULL) {
	phgPrintf("DOF \"%s\": no interpolation function for type \"%s\", "
		  "changing userfunc to DofNoAction\n", dof->name,
		  type->name);
	dof->userfunc = DofNoAction;
    }
    dof->g = g;
    if ((dof->dim = dim) <= 0)
	phgError(1, "invalid value for dim (%d)", dim);
    dof->srcfile = srcfile;
    dof->srcline = srcline;

    if (!SpecialDofType(type)) {
	if (type->np_vert > 0 && !g->vert_flag)
	    phgError(1, "%s elements has vertex DOFs "
		     "but VERT_FLAG is FALSE.\n", dof->type->name);
	if (type->np_edge > 0 && !g->edge_flag)
	    phgError(1, "%s elements require edge indices "
		     "but EDGE_FLAG is FALSE.\n", dof->type->name);
	if (type->np_face > 0 && !g->face_flag)
	    phgError(1, "%s elements require face indices "
		     "but FACE_FLAG is FALSE.\n", dof->type->name);
	if (type->np_elem > 0 && !g->elem_flag)
	    phgError(1, "%s elements require element indices "
		     "but ELEM_FLAG is FALSE.\n", dof->type->name);
    }

    /* attach the new DOF to g */
    len = 0;
    if (g->dof != NULL) {
	DOF **p;
	for (p = g->dof; *p != NULL; p++, len++);
    }
    g->dof = phgRealloc(g->dof, (len + 2) * sizeof(*(g->dof)));
    g->dof[len] = dof;
    g->dof[len + 1] = NULL;

    /* init dof's cache */
    dof->cache_func = NULL;

    dof->count_vert = dim * (SpecialDofType(type) ? 0 : type->np_vert);
    dof->count_edge = dim * (SpecialDofType(type) ? 0 : type->np_edge);
    dof->count_face = dim * (SpecialDofType(type) ? 0 : type->np_face);
    dof->count_elem = dim * (SpecialDofType(type) ? 0 : type->np_elem);

    /* init DOF values */
    if (dof->userfunc == DofNoData) {
	dof->data = NULL;
	dof->userfunc = DofNoAction;
    }
    else if (userfunc != DofLambdaFunction) {
	phgDofSetDataByFunction(dof, userfunc);
    }

    /* update the list of referenced DOF_TYPE's */
    if (type->id >= 0) {
	assert(type->id < type_list_count
	       && type_list[type->id].type == type);
	type_list[type->id].refcount++;
    }
    else {
	/* add a new entry */
	if (type_list_count >= type_list_allocated) {
	    len = (type_list_allocated += 8) * sizeof(*type_list);
	    type_list = phgRealloc(type_list, len);
	}
	type_list[type_list_count].type = type;
	type_list[type_list_count].refcount = 1;
	type->id = type_list_count++;
    }

    Return dof;
    return dof;		/* to make MIPS cc happy */
}

void
phgDofFree(DOF **dof_ptr)
{
    DOF *dof = (dof_ptr == NULL ? NULL : *dof_ptr);

    MagicCheck(DOF, dof)

    if (dof == NULL)
	return;

    if (dof->g != NULL && !phgDofIsValid(dof->g, dof))
	phgError(1, "phgDofFree: invalid DOF object.\n");

    phgInfo(2, "freeing DOF \"%s\".\n", 
		dof->name == NULL ? "nonname" : dof->name);

    if (dof->type->id >= 0) {
	/* update the list of referenced DOF_TYPEs */
	assert(dof->type->id < type_list_count &&
	       type_list[dof->type->id].type == dof->type);
	if (--type_list[dof->type->id].refcount == 0) {
	    /* remove the type from type_list[] */
	    phgDofClearCache(NULL, NULL, dof->type, TRUE);
	    if (dof->type->id < --type_list_count) {
		int i;
		memmove(type_list + dof->type->id,
			type_list + dof->type->id + 1,
			(type_list_count -
			 dof->type->id) * sizeof(*type_list));
		for (i = dof->type->id; i < type_list_count; i++)
		    type_list[i].type->id = i;
	    }
	    if (dof->type->free_after_use) {
		phgFree((void *)dof->type->name);
		phgFree(dof->type->points);
		phgFree(dof->type);
	    }
	    else
		dof->type->id = -1;
	    if (type_list_count == 0) {
		phgFree(type_list);
		type_list = NULL;
		type_list_allocated = 0;
	    }
	}
    }

    phgFree(dof->name);
    phgFree(dof->data);
    phgDofClearCache(NULL, dof, NULL, TRUE);

    /* detach the dof from the mesh */
    if (phgGridIsValid(dof->g)) {
	DOF **p, **p0 = NULL;
	size_t len = 0;
	for (p = dof->g->dof; p != NULL && *p != NULL; p++, len++) {
	    if (*p == dof) {
		len = 0;
		p0 = p;
	    }
	}
	if (p == NULL || p0 == NULL)
	    phgError(1, "%s:%d: unexpected error (grid freed before DOF?)\n",
		     __FILE__, __LINE__);
	memmove(p0, p0 + 1, sizeof(*p0) * len);
    }

    phgFree(dof);

    *dof_ptr = NULL;

    return;
}

void
phgDofSetFunction(DOF *u, DOF_USER_FUNC func)
{
    assert(u->type != DOF_CONSTANT);
    u->userfunc = func;
    u->userfunc_lambda = NULL;
    phgDofSetDataByFunction(u, func);
}

void
phgDofSetLambdaFunction(DOF *u, DOF_USER_FUNC_LAMBDA func)
{
    assert(u->type != DOF_CONSTANT);
    u->userfunc = DofLambdaFunction;
    u->userfunc_lambda = func;
    phgDofSetDataByLambdaFunction(u, func);
}

BTYPE
phgDofSetDirichletBoundaryMask(DOF *u, BTYPE mask)
{
    BTYPE old_mask = u->DB_mask;
    u->DB_mask = mask;
    return old_mask;
}

BOOLEAN
phgDofIsValid(GRID *g, DOF *dof)
{
    DOF **p;

    if (g == NULL || (p = g->dof) == NULL)
	return FALSE;

    while (*p != dof && *p != NULL)
	p++;

    return *p == dof;
}

DOF *
phgDofIndex2Pointer(GRID *g, int id)
{
    DOF **p;

    if ((p = g->dof) == NULL)
	return NULL;

    while (id-- > 0 && *p != NULL) {
	p++;
    }

    return *p;
}

int
phgDofPointer2Index(DOF *dof)
{
    DOF **p;

    if (dof == NULL)
	return -1;

    p = dof->g->dof;
    while (*p != NULL && *p != dof) {
	p++;
    }

    return (*p == dof) ? p - dof->g->dof : -1;
}

static void
clear_cache0(DOF_TYPE *type, BOOLEAN final)
{
    if (final || !type->invariant) {
	phgQuadClearDofCache(&type->cache_basfunc, NULL, final);
	phgQuadClearDofCache(&type->cache_basgrad, NULL, final);
    }
    phgQuadClearDofCache(&type->cache_gradient, NULL, final);
    phgQuadClearDofCache(&type->cache_curl, NULL, final);
}

void
phgDofClearCache(GRID *g, DOF *dof, DOF_TYPE *type, BOOLEAN final)
/* when 'type' is NULL and dof is not NULL, 
 * when dof is not NULL , the parameter 'g' take no effective and the cache in 'type' will be cleared. In this case, none cache in referenced DOF_TYPE will be cleared.
 * when dof is  NULL, one or all of the referenced DOF's caches or DOF_TYPE's caches will be cleared. */
{
    size_t i;
    DOF **dof1;

    if (dof != NULL) {
	phgQuadClearDofCache(&(dof->cache_func), NULL, final);
	if (type != NULL) {
	    clear_cache0(type, final);
	    return;
	}
    }

    if (g != NULL && g->dof != NULL) {
	for (dof1 = g->dof; *dof1 != NULL; dof1++)
	    phgQuadClearDofCache(&((*dof1)->cache_func), NULL, final);
    }

    if (type != NULL) {
	clear_cache0(type, final);
	return;
    }
    else {
	/* clear all quad caches */
	for (i = 0; i < type_list_count; i++)
	    clear_cache0(type_list[i].type, final);
    }
}

void
phgDofMapEdgeData(const DOF_TYPE *type, const SIMPLEX *e, int edge_no, int *M)
/* returns in M[] ordering of data according to edge direction */
{
    int i;

    if (type->np_edge == 0)
	return;

    if (type->np_edge <= 1 || type->BasFuncs == NULL || type->points == NULL
	|| MapVertex(e, GetEdgeVertex(edge_no, 0))
	< MapVertex(e, GetEdgeVertex(edge_no, 1))) {
	for (i = 0; i < type->np_edge; i++)
	    M[i] = i;
    }
    else {
	for (i = 0; i < type->np_edge; i++)
	    M[i] = type->np_edge - 1 - i;
    }

    return;
}

void
phgDofMapFaceData(const DOF_TYPE *type, const SIMPLEX *e, int face_no, int *M)
/* compute ordering of data according to global indices of face vertices */
{
    int v0, v1, v2;

    if (type->np_face <= 1 || type->BasFuncs == NULL || type->points == NULL) {
	/* Note: for non-point DOF the bases are already ordered according to
	 * global face orientation, return an identical map in this case */
	for (v0 = 0; v0 < type->np_face; v0++)
	    M[v0] = v0;
	return;
    }
    else if (type->np_face != 3) {
	/* TODO: should decompose 'np_face' into 6 * n + 3 * m + l (l = 0 or 1)
	   if 'points' = {n S111 + m S211 + l S3}, where Sxxx denotes the
	   permutation stars of (a,b,c). */
	phgError(1, "%s:%d: unimplemented case (np_face=%d)!\n",
		 __FILE__, __LINE__, type->np_face);
    }

    M[0] = 0;
    M[1] = 0;
    M[2] = 0;
    v0 = MapVertex(e, GetFaceVertex(face_no, 0));
    v1 = MapVertex(e, GetFaceVertex(face_no, 1));
    v2 = MapVertex(e, GetFaceVertex(face_no, 2));
    (v0 > v1) ? M[0]++ : M[1]++;
    (v0 > v2) ? M[0]++ : M[2]++;
    (v1 > v2) ? M[1]++ : M[2]++;

    return;
}

void
phgDofGetBasesOnFace(DOF_TYPE *type, SIMPLEX *e, int face_no, SHORT bases[])
/* returns the list of element bases of DOF_TYPE 'type0' on the element 'e'
 * which are not identically zero on the face 'face_no' (including VERTEX,
 * EDGE, FACE and ELEMENT bases), where:
 * 	type0 = (type->base_type != NULL ? type->base_type : type)
 *
 * The bases are ordered according to the global orientation of the FACE,
 * i.e., global indices of the three vertices of the face, such that the
 * ordering is identical for the two elements sharing the face.
 *
 * Note:
 *   1.	The length of the array bases[] is computed by:
 * 		3 * (type->np_vert + type->np_edge) + type->np_face
 * 	or
 * 		1
 * 	if type is DOF_P0
 *
 *   2. The current implementation assumes only VERTEX/EDGE/FACE bases
 *      on the face are not zero.
 *
 *      TODO: add a flag in DOF_TYPE to indicate whether it has this property.
 */
{
    int i, j, np, m0, m, v0, v1, v[3], *M;

    assert(type != NULL);
    if (type->base_type != NULL)
	type = type->base_type;

    /* Note: the following lines are to be removed in the future, after
     * adding an appropriate flag in DOF_TYPE indicating which bases
     * functions are nonzero on a given face */
    if (type == DOF_P0) {
	bases[0] = 0;
	return;
    }

    M = phgAlloc(sizeof(*M) *
	(type->np_edge > type->np_face ? type->np_edge : type->np_face));
    GetFaceVertices(e, face_no, v[0], v[1], v[2]);

    /* The VERTEX bases */
    np = type->np_vert;
    for (i = 0; i < 3; i++) {
	m = v[i] * np;
	for (j = 0; j < np; j++)
	    *(bases++) = m++;
    }

    /* The EDGE bases */
    m0 = NVert * np;
    np = type->np_edge;
    for (i = 0; i < 3; i++) {
	switch (i) {
	    case 0: 	v0 = v[0]; v1 = v[1]; break;
	    case 1: 	v0 = v[0]; v1 = v[2]; break;
	    case 2: 	v0 = v[1]; v1 = v[2]; break;
	    default:	v0 = v1 = 0;	/* to make gcc happy */
	}
	phgDofMapEdgeData(type, e, j = GetEdgeNo(v0,v1), M);
	m = m0 + np * j;
	for (j = 0; j < np; j++)
	    bases[M[j]] = m++;
	bases += np;
    }

    /* The FACE bases */
    m0 += NEdge * np;
    np = type->np_face;
    phgDofMapFaceData(type, e, face_no, M);
    m = m0 + face_no * np;
    for (j = 0; j < np; j++)
	bases[M[j]] = m++;

    phgFree(M);
}

static struct {
    SIMPLEX *e;			/* the element having the DOF */
    BYTE index;			/* local index of the DOF */
    BYTE flag;			/* flag */
} *tmp, *tmp_vert, *tmp_edge, *tmp_face, *tmp_elem, *tmp_end;
static BOOLEAN flagged_only = FALSE;

static BOOLEAN
  setup_tmp_callback
CB_ARGS(e)
{
    int i;

    if (tmp_vert < tmp_edge) {
	for (i = 0; i < NVert; i++) {
	    tmp = tmp_vert + e->verts[i];
	    if (tmp->flag && (!flagged_only || tmp->e->flag < 1))
		continue;
	    tmp->flag = TRUE;
	    tmp->e = e;
	    tmp->index = i;
	}
    }

    if (tmp_edge < tmp_face) {
	for (i = 0; i < NEdge; i++) {
	    tmp = tmp_edge + e->edges[i];
	    if (tmp->flag && (!flagged_only || tmp->e->flag < 1))
		continue;
	    tmp->flag = TRUE;
	    tmp->e = e;
	    tmp->index = i;
	}
    }

    if (tmp_face < tmp_elem) {
	for (i = 0; i < NFace; i++) {
	    tmp = tmp_face + e->faces[i];
	    if (tmp->flag && (!flagged_only || tmp->e->flag < 1))
		continue;
	    tmp->flag = TRUE;
	    tmp->e = e;
	    tmp->index = i;
	}
    }

    if (tmp_elem < tmp_end) {
	tmp = tmp_elem + e->index;
	if (tmp->flag)
	    return TRUE;
	tmp->flag = TRUE;
	tmp->e = e;
	tmp->index = 0;
    }

    return TRUE;
}

static void
setup_tmp(DOF *dof)
{
    size_t size0 = (dof->type->np_vert > 0) ? dof->g->nvert : 0,
	size1 = (dof->type->np_edge > 0) ? dof->g->nedge : 0,
	size2 = (dof->type->np_face > 0) ? dof->g->nface : 0,
	size3 = (dof->type->np_elem > 0) ? dof->g->nelem : 0;

    if (tmp_vert != NULL)
	phgFree(tmp_vert);
    tmp_vert = phgCalloc(size0 + size1 + size2 + size3, sizeof(*tmp_vert));
    tmp_edge = tmp_vert + size0;
    tmp_face = tmp_edge + size1;
    tmp_elem = tmp_face + size2;
    tmp_end = tmp_elem + size3;

    if (tmp_end > tmp_vert)
	phgTraverseElements(dof->g, setup_tmp_callback);
}

void
phgDofDump(DOF *dof)
{
    char s[1024], *fmt;
    int i, j, k, index = 0;
    GRID *g;
    FLOAT *data, *points = NULL, x, y, z;
    SIMPLEX *e = NULL;
    COORD *p0 = NULL, *p1 = NULL, *p2 = NULL, *p3 = NULL;
    static int *M = NULL, M_size = 0;

    if (dof == NULL)
	return;

    FreeAtExit(M);

    if (dof->type == DOF_ANALYTIC) {
	if (dof->userfunc_lambda != NULL)
	    phgInfo(-1, "Analytic DOF \"%s\", dim = %d, lambda function: %p\n",
		    dof->name, dof->dim, dof->userfunc_lambda);
	else
	    phgInfo(-1, "Analytic DOF \"%s\", dim = %d, xyz function: %p\n",
		    dof->name, dof->dim, dof->userfunc);
	return;
    }
    else if (dof->type == DOF_CONSTANT) {
	if (dof->data == NULL) {
	    phgInfo(-1, "Constant DOF \"%s\", dim = %d, no data\n",
		    dof->name, dof->dim);
	    return;
	}
	phgInfo(-1, "Constant DOF \"%s\", dim = %d, values:\n",
		dof->name, dof->dim);
	for (i = 0, j = 0; i < dof->dim / 4; i++, j += 4) {
	    phgInfo(-1, "    %10.4le %10.4le %10.4le %10.4le\n",
		    (double)dof->data[j], (double)dof->data[j + 1],
		    (double)dof->data[j + 2], (double)dof->data[j + 3]);
	}
	switch (dof->dim % 4) {
	    case 1:
		phgInfo(-1, "    %10.4le\n", (double)dof->data[j]);
		break;
	    case 2:
		phgInfo(-1, "    %10.4le %10.4le\n", (double)dof->data[j],
			(double)dof->data[j + 1]);
		break;
	    case 3:
		phgInfo(-1, "    %10.4le %10.4le %10.4le\n",
			(double)dof->data[j], (double)dof->data[j + 1],
			(double)dof->data[j + 2]);
		break;
	}
	return;
    }

    phgInfo(-1, "DOF \"%s\": dim=%dx%d, "
	    "np_vert=%d, np_edge=%d, np_face=%d, np_elem=%d\n",
	    dof->name, dof->dim, dof->type->dim,
	    dof->type->np_vert, dof->type->np_edge,
	    dof->type->np_face, dof->type->np_elem);
    if ((g = dof->g) == NULL || (data = dof->data) == NULL) {
	phgInfo(-1, "no data.\n");
	return;
    }

    if (dof->type->points != NULL) {
	int np_vert = dof->type->np_vert;	/* backup dof->type->np_vert */
	dof->type->np_vert = 0;	/* don't need tmp_vert */
	setup_tmp(dof);
	dof->type->np_vert = np_vert;	/* restore dof->type->np_vert */
    }

    if ((k = dof->type->np_vert) > 0) {
	fmt = (k <= 1) ? "%5d  " : ((k < 10) ? "%5d[%d]  " : "%5d[%02d]  ");
	phgInfo(-1, "Vertices:\n");
	for (i = 0; i < g->nvert; i++) {
	    if (g->types_vert[i] == UNREFERENCED) {
		data += dof->count_vert;
		continue;
	    }
	    for (j = 0; j < dof->type->np_vert; j++) {
		sprintf(s, fmt, GlobalVertex(g, i), j);
		if (dof->type->points != NULL)
		    sprintf(s + strlen(s), "%s(%6.3lf,%6.3lf,%6.3lf) = ",
			    dof->name, (double)g->verts[i][0],
			    (double)g->verts[i][1], (double)g->verts[i][2]);
		for (k = 0; k < dof->dim; k++)
		    sprintf(s + strlen(s), "%10.4le ", (double)*(data++));
		phgInfo(-1, "%s\n", s);
	    }
	}
    }

    if ((k = dof->type->np_edge) > 0) {
	if (M_size < k)
	    M = phgAlloc((M_size = k) * sizeof(*M));
	fmt = (k <= 1) ? "%5d  " : ((k < 10) ? "%5d[%d]  " : "%5d[%02d]  ");
	phgInfo(-1, "Edges:\n");
	for (i = 0; i < g->nedge; i++) {
	    if (g->types_edge[i] == UNREFERENCED) {
		data += dof->count_edge;
		continue;
	    }
	    if (dof->type->points != NULL) {
		assert(tmp_edge[i].e != NULL);
		e = tmp_edge[i].e;
		index = tmp_edge[i].index;
		p0 = g->verts + e->verts[GetEdgeVertex(index, 0)];
		p1 = g->verts + e->verts[GetEdgeVertex(index, 1)];
		points = dof->type->points + dof->type->np_vert;
	    }
	    phgDofMapEdgeData(dof->type, e, index, M);
	    for (j = 0; j < dof->type->np_edge; j++) {
		sprintf(s, fmt, GlobalEdge(g, i), j);
		if (dof->type->points != NULL) {
		    x = (*p0)[0] * points[0] + (*p1)[0] * points[1];
		    y = (*p0)[1] * points[0] + (*p1)[1] * points[1];
		    z = (*p0)[2] * points[0] + (*p1)[2] * points[1];
		    sprintf(s + strlen(s), "%s(%6.3lf,%6.3lf,%6.3lf) = ",
			    dof->name, (double)x, (double)y, (double)z);
		    points += 2;
		}
		for (k = 0; k < dof->dim; k++)
		    sprintf(s + strlen(s), "%10.4le ",
			    (double)*(data + dof->dim * M[j] + k));
		phgInfo(-1, "%s\n", s);
	    }
	    data += dof->count_edge;
	}
    }

    if ((k = dof->type->np_face) > 0) {
	if (M_size < k)
	    M = phgAlloc((M_size = k) * sizeof(*M));
	fmt = (k <= 1) ? "%5d  " : ((k < 10) ? "%5d[%d]  " : "%5d[%02d]  ");
	phgInfo(-1, "Faces:\n");
	for (i = 0; i < g->nface; i++) {
	    if (g->types_face[i] == UNREFERENCED) {
		data += dof->count_face;
		continue;
	    }
	    if (dof->type->points != NULL) {
		assert(tmp_face[i].e != NULL);
		e = tmp_face[i].e;
		index = tmp_face[i].index;
		p0 = g->verts + e->verts[GetFaceVertex(index, 0)];
		p1 = g->verts + e->verts[GetFaceVertex(index, 1)];
		p2 = g->verts + e->verts[GetFaceVertex(index, 2)];
		points = dof->type->points + dof->type->np_vert +
		    dof->type->np_edge * 2;
	    }
	    phgDofMapFaceData(dof->type, e, index, M);
	    for (j = 0; j < dof->type->np_face; j++) {
		sprintf(s, fmt, GlobalFace(g, i), j);
		if (dof->type->points != NULL) {
		    x = (*p0)[0] * points[0] + (*p1)[0] * points[1] +
			(*p2)[0] * points[2];
		    y = (*p0)[1] * points[0] + (*p1)[1] * points[1] +
			(*p2)[1] * points[2];
		    z = (*p0)[2] * points[0] + (*p1)[2] * points[1] +
			(*p2)[2] * points[2];
		    sprintf(s + strlen(s), "%s(%6.3lf,%6.3lf,%6.3lf) = ",
			    dof->name, (double)x, (double)y, (double)z);
		    points += 3;
		}
		for (k = 0; k < dof->dim; k++)
		    sprintf(s + strlen(s), "%10.4le ",
			    (double)*(data + dof->dim * M[j] + k));
		phgInfo(-1, "%s\n", s);
	    }
	    data += dof->count_face;
	}
    }

    if ((k = dof->type->np_elem) > 0) {
	fmt = (k <= 1) ? "%5d  " : ((k < 10) ? "%5d[%d]  " : "%5d[%02d]  ");
	phgInfo(-1, "Elements:\n");
	for (i = 0; i < g->nelem; i++) {
	    if (g->types_elem[i] == UNREFERENCED) {
		data += dof->count_elem;
		continue;
	    }
	    if (dof->type->points != NULL) {
		assert(tmp_elem[i].e != NULL);
		e = tmp_elem[i].e;
		p0 = g->verts + e->verts[0];
		p1 = g->verts + e->verts[1];
		p2 = g->verts + e->verts[2];
		p3 = g->verts + e->verts[3];
		points = dof->type->points + dof->type->np_vert +
		    dof->type->np_edge * 2 + dof->type->np_face * 3;
	    }
	    for (j = 0; j < dof->type->np_elem; j++) {
		sprintf(s, fmt, GlobalElement(g, i), j);
		if (dof->type->points != NULL) {
		    x = (*p0)[0] * points[0] + (*p1)[0] * points[1] +
			(*p2)[0] * points[2] + (*p3)[0] * points[3];
		    y = (*p0)[1] * points[0] + (*p1)[1] * points[1] +
			(*p2)[1] * points[2] + (*p3)[1] * points[3];
		    z = (*p0)[2] * points[0] + (*p1)[2] * points[1] +
			(*p2)[2] * points[2] + (*p3)[2] * points[3];
		    sprintf(s + strlen(s), "%s(%6.3lf,%6.3lf,%6.3lf) = ",
			    dof->name, (double)x, (double)y, (double)z);
		    points += 4;
		}
		for (k = 0; k < dof->dim; k++)
		    sprintf(s + strlen(s), "%10.4le ", (double)*(data++));
		phgInfo(-1, "%s\n", s);
	    }
	}
    }

    phgFree(tmp_vert);
    tmp_vert = tmp_edge = tmp_face = tmp_elem = tmp_end = NULL;
}

static SIMPLEX *elem;
static FLOAT **pdata = NULL, **cdata = NULL, *cdata0[15], *cdata1[15];
static BOOLEAN cdata0_flag = FALSE, cdata1_flag = FALSE;
static int vmap0[NVert], emap0[NEdge], vmap1[NVert], emap1[NEdge];

static void
c2f_func0(DOF *dof, SIMPLEX *e0, int bno, const FLOAT *lambda, FLOAT *values)
{
    FLOAT lam[Dim + 1];

    Unused(bno);

    lam[0] = lambda[vmap0[0]] + (lam[1] = lambda[3] * 0.5);
    lam[2] = lambda[vmap0[2]];
    lam[3] = lambda[vmap0[3]];

    phgDofEval_(dof, elem, lam, NULL, 0, values, pdata);
}

static void
c2f_func1(DOF *dof, SIMPLEX *e1, int bno, const FLOAT *lambda, FLOAT *values)
{
    FLOAT lam[Dim + 1];

    Unused(bno);

    lam[1] = lambda[vmap1[1]] + (lam[0] = lambda[3] * 0.5);
    lam[2] = lambda[vmap1[2]];
    lam[3] = lambda[vmap1[3]];

    phgDofEval_(dof, elem, lam, NULL, 0, values, pdata);
}

static void
f2c_func(DOF *dof, SIMPLEX *e, int bno, const FLOAT *lambda, FLOAT *values)
{
    int *map;
    FLOAT lam[Dim + 1], **data;

    Unused(bno);

    if ((e->children[1] == NULL) ||
	(lambda[0] >= lambda[1] && e->children[0] != NULL)) {
	/* use child 0 */
	map = vmap0;
	lam[map[0]] = lambda[0] - lambda[1];
	lam[3] = 2. * lambda[1];
	lam[map[2]] = lambda[2];
	lam[map[3]] = lambda[3];
	e = e->children[0];
	data = cdata0;
	if (cdata0_flag == FALSE) {
	    cdata0_flag = TRUE;
	    /* VERTEX data */
	    data[3] = cdata[0];
	    data[map[0]] = pdata[0];
	    data[map[2]] = pdata[2];
	    data[map[3]] = pdata[3];
	    /* EDGE data */
	    data[NVert + GetEdgeNo(map[0], 3)] = cdata[1];
	    data[NVert + GetEdgeNo(map[2], 3)] = cdata[3];
	    data[NVert + GetEdgeNo(map[3], 3)] = cdata[4];
	    data[NVert + GetEdgeNo(map[0], map[2])] = pdata[NVert + 1];
	    data[NVert + GetEdgeNo(map[0], map[3])] = pdata[NVert + 2];
	    data[NVert + GetEdgeNo(map[2], map[3])] = pdata[NVert + 5];
	    /* FACE data */
	    data[NVert + NEdge + map[2]] = cdata[5];
	    data[NVert + NEdge + map[3]] = cdata[7];
	    data[NVert + NEdge + map[0]] = cdata[9];
	    data[NVert + NEdge + 3] = pdata[NVert + NEdge + 1];
	    /* ELEMENT data */
	    data[NVert + NEdge + NFace] = cdata[10];
	}
    }
    else {
	/* use child 1 */
	map = vmap1;
	lam[map[1]] = lambda[1] - lambda[0];
	lam[3] = 2. * lambda[0];
	lam[map[2]] = lambda[2];
	lam[map[3]] = lambda[3];
	e = e->children[1];
	data = cdata1;
	if (cdata1_flag == FALSE) {
	    cdata1_flag = TRUE;
	    /* VERTEX data */
	    data[3] = cdata[0];
	    data[map[1]] = pdata[1];
	    data[map[2]] = pdata[2];
	    data[map[3]] = pdata[3];
	    /* EDGE data */
	    data[NVert + GetEdgeNo(map[1], 3)] = cdata[2];
	    data[NVert + GetEdgeNo(map[2], 3)] = cdata[3];
	    data[NVert + GetEdgeNo(map[3], 3)] = cdata[4];
	    data[NVert + GetEdgeNo(map[1], map[2])] = pdata[NVert + 3];
	    data[NVert + GetEdgeNo(map[1], map[3])] = pdata[NVert + 4];
	    data[NVert + GetEdgeNo(map[2], map[3])] = pdata[NVert + 5];
	    /* FACE data */
	    data[NVert + NEdge + map[2]] = cdata[6];
	    data[NVert + NEdge + map[3]] = cdata[8];
	    data[NVert + NEdge + map[1]] = cdata[9];
	    data[NVert + NEdge + 3] = pdata[NVert + NEdge + 0];
	    /* ELEMENT data */
	    data[NVert + NEdge + NFace] = cdata[11];
	}
    }

    phgDofEval_(dof, e, lam, NULL, 0, values, data);
}

void
phgDofInterpC2FGeneric(DOF *dof, SIMPLEX *e, FLOAT **parent_data,
		       FLOAT **children_data)
/* Default coarse-to-fine interpolation function. It can be used
 * as the 'InterpC2F' member in any DOF_TYPE which has valid 'InitFunc'
 * and 'BasFuncs' members.
 *
 * Note:
 *	For better efficiency specialized functions should be implemented
 *	for specific DOF_TYPEs */
{
    SIMPLEX *e0 = e->children[0], *e1 = e->children[1];
    FLOAT **child0_data, **child1_data;

    phgMapP2C(e, vmap0, emap0, 0);
    phgMapP2C(e, vmap1, emap1, 1);

    /* Note: only need child[01]_data for HBn, HCn and HFEBn with n >= 2 */
    /* FIXME: do we need to add a flag in DOF_TYPE for this test? */
    if (dof->type->order > 1 /* other tests */) {
	/* setup complete list of pointers for children data */
	child0_data = phgAlloc(2 * 15 * sizeof(*child0_data));
	child1_data = child0_data + 15;

	/* the vertices of child0 */
	child0_data[vmap0[0]] = parent_data[0];
	child0_data[vmap0[2]] = parent_data[2];
	child0_data[vmap0[3]] = parent_data[3];
	child0_data[3] = children_data[0];

	/* the vertices of child1 */
	child1_data[vmap1[1]] = parent_data[1];
	child1_data[vmap1[2]] = parent_data[2];
	child1_data[vmap1[3]] = parent_data[3];
	child1_data[3] = children_data[0];

	/* the edges of child0 */
	child0_data[4 + emap0[1]] = parent_data[5];	/* parent edge 0-2 */
	child0_data[4 + emap0[2]] = parent_data[6];	/* parent edge 0-3 */
	child0_data[4 + emap0[5]] = parent_data[9];	/* parent edge 2-3 */
	/* cut edge (v0 mapped to 2 or 0) */
	if (vmap0[0] == 2)
	    child0_data[4 + 5] = children_data[1];	/* cut edge (2-3) */
	else
	    child0_data[4 + 2] = children_data[1];	/* cut edge (0-3) */
	/* new edge on v2 (v2 mapped to 1) */
	child0_data[4 + 4] = children_data[3];		/* new edge (1-3) */
	/* new edge on v3 (v3 mapped to 0 or 2) */
	if (vmap0[3] == 0)
	    child0_data[4 + 2] = children_data[4];	/* new edge (0-3) */
	else
	    child0_data[4 + 5] = children_data[4];	/* new edge (2-3) */

	/* the edges of child1 */
	child1_data[4 + emap1[3]] = parent_data[7];	/* parent edge 1-2 */
	child1_data[4 + emap1[4]] = parent_data[8];	/* parent edge 1-3 */
	child1_data[4 + emap1[5]] = parent_data[9];	/* parent edge 2-3 */
	/* cut edge  (v1 mapped to 2 or 0) */
	if (vmap1[1] == 2)
	    child1_data[4 + 5] = children_data[2];	/* cut edge (2-3) */
	else
	    child1_data[4 + 2] = children_data[2];	/* cut edge (0-3) */
	/* new edge on v2 (v2 mapped to 2 or 1) */
	if (vmap1[2] == 2)
	    child1_data[4 + 5] = children_data[3];	/* new edge (2-3) */
	else
	    child1_data[4 + 4] = children_data[3];	/* new edge (1-3) */
	/* new edge on v3 (v3 mapped to 0, 1 or 2) */
	if (vmap1[3] == 0)
	    child1_data[4 + 2] = children_data[4];	/* new edge (0-3) */
	else if (vmap1[3] == 1)
	    child1_data[4 + 4] = children_data[4];	/* new edge (1-3) */
	else
	    child1_data[4 + 5] = children_data[4];	/* new edge (2-3) */

	/* the faces of child0 */
	child0_data[10 + 3] = parent_data[11];		/* parent face 1 */
	child0_data[10 + vmap0[0]] = children_data[9];	/* new face */
	child0_data[10 + vmap0[2]] = children_data[5];	/* cut face */
	child0_data[10 + vmap0[3]] = children_data[7];	/* cut face */

	/* the faces of child1 */
	child1_data[10 + 3] = parent_data[10];		/* parent face 0 */
	child1_data[10 + vmap1[1]] = children_data[9];	/* new face */
	child1_data[10 + vmap1[2]] = children_data[6];	/* cut face */
	child1_data[10 + vmap1[3]] = children_data[8];	/* cut face */

	/* the volume data */
	child0_data[14] = children_data[10];
	child1_data[14] = children_data[11];
    }
    else {
	child0_data = child1_data = NULL;
    }

    elem = e;
    pdata = parent_data;

    if (dof->type->np_vert > 0) {
	/* The new vertex */
	dof->type->InitFunc(dof, e0, VERTEX, 3, NULL,
			    c2f_func0, NULL, children_data[0], child0_data);
    }

    if (dof->type->np_edge > 0) {
	/* The cut-edges */
	dof->type->InitFunc(dof, e0, EDGE, emap0[0], NULL,
			    c2f_func0, NULL, children_data[1], child0_data);
	dof->type->InitFunc(dof, e1, EDGE, emap1[0], NULL,
			    c2f_func1, NULL, children_data[2], child1_data);
	/* The new edges */
	dof->type->InitFunc(dof, e0, EDGE, GetEdgeNo(vmap0[2], 3), NULL,
			    c2f_func0, NULL, children_data[3], child0_data);
	dof->type->InitFunc(dof, e0, EDGE, GetEdgeNo(vmap0[3], 3), NULL,
			    c2f_func0, NULL, children_data[4], child0_data);
    }

    if (dof->type->np_face > 0) {
	/* the cut faces of e0 */
	dof->type->InitFunc(dof, e0, FACE, vmap0[2], NULL,
			    c2f_func0, NULL, children_data[5], child0_data);
	dof->type->InitFunc(dof, e0, FACE, vmap0[3], NULL,
			    c2f_func0, NULL, children_data[7], child0_data);
	/* the cut faces of e1 */
	dof->type->InitFunc(dof, e1, FACE, vmap1[2], NULL,
			    c2f_func1, NULL, children_data[6], child1_data);
	dof->type->InitFunc(dof, e1, FACE, vmap1[3], NULL,
			    c2f_func1, NULL, children_data[8], child1_data);
	/* the new face */
	dof->type->InitFunc(dof, e0, FACE, vmap0[0], NULL,
			    c2f_func0, NULL, children_data[9], child0_data);
    }

    if (dof->type->np_elem > 0) {
	/* volume data on e0 */
	dof->type->InitFunc(dof, e0, ELEMENT, 0, NULL,
			    c2f_func0, NULL, children_data[10], child0_data);
	/* volume data on e1 */
	dof->type->InitFunc(dof, e1, ELEMENT, 0, NULL,
			    c2f_func1, NULL, children_data[11], child1_data);
    }

    phgFree(child0_data);
}

void
phgDofInterpF2CGeneric(DOF *dof, SIMPLEX *e, FLOAT **parent_data,
		       FLOAT **children_data)
/* Default fine-to-coarse interpolation function. It can be used
 * as the 'InterpF2C' member in any DOF_TYPE which has valid 'InitFunc'
 * and 'BasFuncs' members.
 *
 * Note:
 *   1. It is assumed that 'parent_data' already contains valid values at
 *	the locations common to parent and children, only values at edge 0,
 *	faces 2 and 3, and element data need to be computed.
 *
 *   2. For better efficiency specialized functions should be implemented
 *	for specific DOF_TYPEs */
{
#if USE_MPI
    SIMPLEX *e0 = e->children[0], *e1 = e->children[1];
#endif

    pdata = parent_data;
    cdata = children_data;
    cdata0_flag = FALSE;
    cdata1_flag = FALSE;

    phgMapP2C(e, vmap0, emap0, 0);
    phgMapP2C(e, vmap1, emap1, 1);

#if USE_MPI
    if (dof->type->np_vert > 0) {
	if (e0 == NULL)
	    dof->type->InitFunc(dof, e, VERTEX, 0, NULL, f2c_func, NULL,
				parent_data[0], parent_data);
	else if (e1 == NULL)
	    dof->type->InitFunc(dof, e, VERTEX, 1, NULL, f2c_func, NULL,
				parent_data[1], parent_data);
    }
#endif

    if (dof->type->np_edge > 0) {
	dof->type->InitFunc(dof, e, EDGE, 0, NULL, f2c_func, NULL,
			    parent_data[NVert + 0], parent_data);
#if USE_MPI
	if (e0 == NULL) {
	    dof->type->InitFunc(dof, e, EDGE, 1, NULL, f2c_func, NULL,
				parent_data[NVert + 1], parent_data);
	    dof->type->InitFunc(dof, e, EDGE, 2, NULL, f2c_func, NULL,
				parent_data[NVert + 2], parent_data);
	}
	else if (e1 == NULL) {
	    dof->type->InitFunc(dof, e, EDGE, 3, NULL, f2c_func, NULL,
				parent_data[NVert + 3], parent_data);
	    dof->type->InitFunc(dof, e, EDGE, 4, NULL, f2c_func, NULL,
				parent_data[NVert + 4], parent_data);
	}
#endif
    }

    if (dof->type->np_face > 0) {
	dof->type->InitFunc(dof, e, FACE, 2, NULL, f2c_func, NULL,
			    parent_data[NVert + NEdge + 2], parent_data);
	dof->type->InitFunc(dof, e, FACE, 3, NULL, f2c_func, NULL,
			    parent_data[NVert + NEdge + 3], parent_data);
#if USE_MPI
	if (e0 == NULL)
	    dof->type->InitFunc(dof, e, FACE, 1, NULL, f2c_func, NULL,
				parent_data[NVert + NEdge + 1], parent_data);
	else if (e1 == NULL)
	    dof->type->InitFunc(dof, e, FACE, 0, NULL, f2c_func, NULL,
				parent_data[NVert + NEdge + 0], parent_data);
#endif
    }

    if (dof->type->np_elem > 0) {
	dof->type->InitFunc(dof, e, ELEMENT, 0, NULL, f2c_func, NULL,
			    parent_data[NVert + NEdge + NFace], parent_data);
    }
}

void
phgDofInitFuncPoint(DOF *dof, SIMPLEX *e, GTYPE type, int index,
		    DOF_USER_FUNC userfunc,
		    DOF_USER_FUNC_LAMBDA userfunc_lambda,
		    const FLOAT *funcvalues, FLOAT *dofvalues,
		    FLOAT **pdofvalues)
{
    const GRID *g = dof->g;
    COORD *p0 = NULL, *p1 = NULL, *p2 = NULL, *p3;
    FLOAT *points;
    FLOAT x, y, z, lambda[Dim + 1];
    int n, j, v0, v1, v2, bno = 0;

    Unused(pdofvalues);
    assert(dof->type->dim == 1);	/* only 1 is meaningful */

    switch (type) {
	case VERTEX:
	    n = dof->type->np_vert;
	    break;
	case EDGE:
	    n = dof->type->np_edge;
	    break;
	case FACE:
	    n = dof->type->np_face;
	    break;
	case ELEMENT:
	    n = dof->type->np_elem;
	    break;
	default:
	    n = 0;		/* make gcc happy */
    }

    if (funcvalues != NULL) {
	memcpy(dofvalues, funcvalues, n * dof->dim * sizeof(*dofvalues));
	return;
    }

    if (dof->type->points == NULL)
	phgError(1, "%s:%d, not a point DOF.\n", __FILE__, __LINE__);

    if (userfunc == NULL && userfunc_lambda == NULL)
	phgError(1, "%s:%d, invalid user function!\n", __FILE__, __LINE__);

    switch (type) {
	case VERTEX:
	    if (userfunc_lambda != NULL) {
		lambda[0] = lambda[1] = lambda[2] = lambda[3] = 0.;
		lambda[index] = 1.0;
		bno = index * n;
		for (j = 0; j < n; j++) {
		    userfunc_lambda(dof, e, bno + j, lambda, dofvalues);
		    dofvalues += dof->dim;
		}
		break;
	    }
	    x = g->verts[e->verts[index]][0];
	    y = g->verts[e->verts[index]][1];
	    z = g->verts[e->verts[index]][2];
	    for (j = 0; j < n; j++) {
		userfunc(x, y, z, dofvalues);
		dofvalues += dof->dim;
	    }
	    break;
	case EDGE:
	    v0 = GetEdgeVertex(index, 0);
	    v1 = GetEdgeVertex(index, 1);
	    if (userfunc_lambda != NULL) {
		lambda[0] = lambda[1] = lambda[2] = lambda[3] = 0.;
		bno = NVert * dof->type->np_vert + index * n;
	    }
	    else {
		p0 = g->verts + e->verts[v0];
		p1 = g->verts + e->verts[v1];
	    }
	    points = dof->type->points + dof->type->np_vert;
	    if (n > 1) {
		int M[n];
		phgDofMapEdgeData(dof->type, e, index, M);
		if (userfunc_lambda == NULL) {
		    for (j = 0; j < n; j++) {
			x = (*p0)[0] * points[0] + (*p1)[0] * points[1];
			y = (*p0)[1] * points[0] + (*p1)[1] * points[1];
			z = (*p0)[2] * points[0] + (*p1)[2] * points[1];
			userfunc(x, y, z, dofvalues + M[j] * dof->dim);
			points += 2;
		    }
		}
		else {
		    for (j = 0; j < n; j++) {
			lambda[v0] = points[0];
			lambda[v1] = points[1];
			userfunc_lambda(dof, e, bno + j, lambda,
					dofvalues + M[j] * dof->dim);
			points += 2;
		    }
		}
	    }
	    else {
		if (userfunc_lambda == NULL) {
		    for (j = 0; j < n; j++) {
			x = (*p0)[0] * points[0] + (*p1)[0] * points[1];
			y = (*p0)[1] * points[0] + (*p1)[1] * points[1];
			z = (*p0)[2] * points[0] + (*p1)[2] * points[1];
			userfunc(x, y, z, dofvalues);
			points += 2;
			dofvalues += dof->dim;
		    }
		}
		else {
		    for (j = 0; j < n; j++) {
			lambda[v0] = points[0];
			lambda[v1] = points[1];
			userfunc_lambda(dof, e, bno + j, lambda, dofvalues);
			points += 2;
			dofvalues += dof->dim;
		    }
		}
	    }
	    break;
	case FACE:
	    v0 = GetFaceVertex(index, 0);
	    v1 = GetFaceVertex(index, 1);
	    v2 = GetFaceVertex(index, 2);
	    if (userfunc_lambda != NULL) {
		lambda[0] = lambda[1] = lambda[2] = lambda[3] = 0.;
		bno = NVert * dof->type->np_vert + NEdge * dof->type->np_edge +
		      index * n;
	    }
	    else {
		p0 = g->verts + e->verts[v0];
		p1 = g->verts + e->verts[v1];
		p2 = g->verts + e->verts[v2];
	    }
	    points = dof->type->points + dof->type->np_vert +
		dof->type->np_edge * 2;
	    if (n > 1) {
		int M[n];
		phgDofMapFaceData(dof->type, e, index, M);
		if (userfunc_lambda == NULL) {
		    for (j = 0; j < n; j++) {
			x = (*p0)[0] * points[0] + (*p1)[0] * points[1] +
			    (*p2)[0] * points[2];
			y = (*p0)[1] * points[0] + (*p1)[1] * points[1] +
			    (*p2)[1] * points[2];
			z = (*p0)[2] * points[0] + (*p1)[2] * points[1] +
			    (*p2)[2] * points[2];
			userfunc(x, y, z, dofvalues + M[j] * dof->dim);
			points += 3;
		    }
		}
		else {
		    for (j = 0; j < n; j++) {
			lambda[v0] = points[0];
			lambda[v1] = points[1];
			lambda[v2] = points[2];
			userfunc_lambda(dof, e, bno + j, lambda,
					dofvalues + M[j] * dof->dim);
			points += 3;
		    }
		}
	    }
	    else {
		if (userfunc_lambda == NULL) {
		    for (j = 0; j < n; j++) {
			x = (*p0)[0] * points[0] + (*p1)[0] * points[1] +
			    (*p2)[0] * points[2];
			y = (*p0)[1] * points[0] + (*p1)[1] * points[1] +
			    (*p2)[1] * points[2];
			z = (*p0)[2] * points[0] + (*p1)[2] * points[1] +
			    (*p2)[2] * points[2];
			userfunc(x, y, z, dofvalues);
			points += 3;
			dofvalues += dof->dim;
		    }
		}
		else {
		    for (j = 0; j < n; j++) {
			lambda[v0] = points[0];
			lambda[v1] = points[1];
			lambda[v2] = points[2];
			userfunc_lambda(dof, e, bno + j, lambda, dofvalues);
			points += 3;
			dofvalues += dof->dim;
		    }
		}
	    }
	    break;
	case ELEMENT:
	    points = dof->type->points + dof->type->np_vert +
		dof->type->np_edge * 2 + dof->type->np_face * 3;
	    if (userfunc_lambda == NULL) {
		p0 = g->verts + e->verts[0];
		p1 = g->verts + e->verts[1];
		p2 = g->verts + e->verts[2];
		p3 = g->verts + e->verts[3];
		for (j = 0; j < n; j++) {
		    x = (*p0)[0] * points[0] + (*p1)[0] * points[1] +
			(*p2)[0] * points[2] + (*p3)[0] * points[3];
		    y = (*p0)[1] * points[0] + (*p1)[1] * points[1] +
			(*p2)[1] * points[2] + (*p3)[1] * points[3];
		    z = (*p0)[2] * points[0] + (*p1)[2] * points[1] +
			(*p2)[2] * points[2] + (*p3)[2] * points[3];
		    userfunc(x, y, z, dofvalues);
		    dofvalues += dof->dim;
		    points += 4;
		}
	    }
	    else {
		bno = NVert * dof->type->np_vert + NEdge * dof->type->np_edge +
		      NFace * dof->type->np_face + index * n;
		for (j = 0; j < n; j++) {
		    userfunc_lambda(dof, e, bno + j, points, dofvalues);
		    dofvalues += dof->dim;
		    points += 4;
		}
	    }
	    break;
    }
}

void
phgDofSetDataByFunction_(DOF *dof, DOF_USER_FUNC userfunc,
			 DOF_USER_FUNC_LAMBDA userfunc_lambda,
			 FLOAT *funcvalues)
{
    GRID *g = dof->g;
    FLOAT *data;
    INT i;
    size_t size;

    if (dof->type == DOF_ANALYTIC)
	return;

    size = (dof->type == DOF_CONSTANT) ? dof->dim : DofGetDataCount(dof);
    if (dof->data == NULL) {
	dof->data = phgCalloc(size, sizeof(*dof->data));
	dof->data_vert = dof->data;
	dof->data_edge = dof->data_vert + DofGetVertexDataCount(dof);
	dof->data_face = dof->data_edge + DofGetEdgeDataCount(dof);
	dof->data_elem = dof->data_face + DofGetFaceDataCount(dof);
    }

    if (userfunc == DofNoAction ||
	(userfunc == DofInterpolation && userfunc_lambda == NULL))
	return;

    if (dof->type == DOF_CONSTANT) {
	/* constant DOF, using funcvalues */
	if (funcvalues != NULL) {
	    memcpy(dof->data, funcvalues, dof->dim * sizeof(*dof->data));
	}
	else if (userfunc_lambda != NULL) {
	    FLOAT lambda[] = { 0., 0., 0., 1. };
	    userfunc_lambda(dof, dof->g->roots, -1, lambda, dof->data);
	}
	else {
	    assert(userfunc != NULL);
	    userfunc(0., 0., 0., dof->data);
	}
	return;
    }

    if ((userfunc != NULL || userfunc_lambda != NULL)
	&& dof->type->InitFunc == NULL)
	phgError(1, "phgDofSetDataByFunction_: no initialization function!\n");

    data = dof->data;
    setup_tmp(dof);

    if (dof->type->np_vert > 0) {
	size = dof->count_vert;
	for (i = 0; i < g->nvert; i++) {
	    if (tmp_vert[i].flag
		&& (!flagged_only || tmp_vert[i].e->flag < 1))
		dof->type->InitFunc(dof, tmp_vert[i].e, VERTEX,
				    tmp_vert[i].index, userfunc,
				    userfunc_lambda, funcvalues, data, NULL);
	    data += size;
	}
    }

    if (dof->type->np_edge > 0) {
	size = dof->count_edge;
	for (i = 0; i < g->nedge; i++) {
	    if (tmp_edge[i].flag
		&& (!flagged_only || tmp_edge[i].e->flag < 1))
		dof->type->InitFunc(dof, tmp_edge[i].e, EDGE,
				    tmp_edge[i].index, userfunc,
				    userfunc_lambda, funcvalues, data, NULL);
	    data += size;
	}
    }

    if (dof->type->np_face > 0) {
	size = dof->count_face;
	for (i = 0; i < g->nface; i++) {
	    if (tmp_face[i].flag
		&& (!flagged_only || tmp_face[i].e->flag < 1))
		dof->type->InitFunc(dof, tmp_face[i].e, FACE,
				    tmp_face[i].index, userfunc,
				    userfunc_lambda, funcvalues, data, NULL);
	    data += size;
	}
    }

    if (dof->type->np_elem > 0) {
	size = dof->count_elem;
	for (i = 0; i < g->nelem; i++) {
	    if (tmp_elem[i].flag
		&& (!flagged_only || tmp_elem[i].e->flag < 1))
		dof->type->InitFunc(dof, tmp_elem[i].e, ELEMENT,
				    tmp_elem[i].index, userfunc,
				    userfunc_lambda, funcvalues, data, NULL);
	    data += size;
	}
    }

    phgFree(tmp_vert);
    tmp_vert = tmp_edge = tmp_face = tmp_elem = tmp_end = NULL;
}

static int dummy_n;
static const FLOAT *dummy_values;
static void
dummy_userfunc_lambda(DOF *dof, SIMPLEX *e, int bno, const FLOAT *lambda,
			FLOAT *values)
{
    int i, n;
    const FLOAT *p;

    Unused(e);
    Unused(lambda);
    Unused(bno);

    n = DofDim(dof);
    for (i = 0, p = dummy_values; i < n; i++) {
	*(values++) = *(p++);
	if (p >= dummy_values + dummy_n)
	    p = dummy_values;
    }
}

void
phgDofSetDataByValue(DOF *dof, FLOAT value)
{
    size_t i, np, nvalues;
    FLOAT *values;

    assert(dof->type != DOF_ANALYTIC);

    if (dof->type != DOF_CONSTANT) {
	nvalues = DofDim(dof);
	if (dof->type->InitFunc != phgDofInitFuncPoint) {
	    if (dof->type->InitFunc == NULL) {
		/* no init function, simply copy the data */
		for (i = 0; i < DofGetDataCount(dof); i++)
		    dof->data[i] = value;
		return;
	    }
	    dummy_n = 1;
	    dummy_values = &value;
	    phgDofSetDataByFunction_(dof, NULL, dummy_userfunc_lambda, NULL);
	    return;
	}
	np = dof->type->np_vert;
	if (np < (i = dof->type->np_edge))
	    np = i;
	if (np < (i = dof->type->np_face))
	    np = i;
	if (np < (i = dof->type->np_elem))
	    np = i;
    }
    else {
	nvalues = dof->dim;
	np = 1;
    }
    values = phgAlloc(nvalues * np * sizeof(*values));
    /* make 'np' * 'nvalues' duplicate copies of the value */
    for (i = 0; i < np * nvalues; i++)
	values[i] = value;
    phgDofSetDataByFunction_(dof, NULL, NULL, values);
    phgFree(values);
}

void
phgDofSetDataByValues(DOF *dof, const FLOAT array_of_values[])
{
    size_t i, np, nvalues;
    FLOAT *values;

    assert(dof->type != DOF_ANALYTIC);

    if (dof->type != DOF_CONSTANT) {
	nvalues = DofDim(dof);
	if (dof->type->InitFunc != phgDofInitFuncPoint) {
	    dummy_n = nvalues;
	    dummy_values = array_of_values;
	    phgDofSetDataByFunction_(dof, NULL, dummy_userfunc_lambda, NULL);
	    return;
	}
	np = dof->type->np_vert;
	if (np < (i = dof->type->np_edge))
	    np = i;
	if (np < (i = dof->type->np_face))
	    np = i;
	if (np < (i = dof->type->np_elem))
	    np = i;
    }
    else {
	nvalues = dof->dim;
	np = 1;
    }
    values = phgAlloc(nvalues * np * sizeof(*values));
    /* make 'np' duplicate copies of the values */
    for (i = 0; i < nvalues; i++)
	values[i] = (i < nvalues) ? array_of_values[i] : values[i - nvalues];
    phgDofSetDataByFunction_(dof, NULL, NULL, values);
    phgFree(values);
}

void
phgDofSetDataByValuesV(DOF *dof, FLOAT v0, ...)
/* set DOF data using variable arguments */
{
    int i, n = DofDim(dof);
    FLOAT *values;
    va_list ap;                                                         \

    values = phgAlloc(n * sizeof(*values));
    values[0] = v0;
    va_start(ap, v0);
    for (i = 1; i < n; i++)
	values[i] = va_arg(ap, FLOAT);
    va_end(ap);
    phgDofSetDataByValues(dof, values);
    phgFree(values);
}

void
phgDofScatter(GRID *g)
/* scatters DOFs to slaves (called by phgPartitionGrid) */
{
#if USE_MPI
    int i, n;
    size_t size, size0;
    DOF **dof;
    FLOAT *buffer, *p, *p0;
    BYTE *flags;

    if (g->dof == NULL)
	return;

    /* clear QUAD and DOF caches */
    phgDofClearCache(g, NULL, NULL, FALSE);

    /* count number of DOF arrays */
    for (n = 0, dof = g->dof; *dof != NULL; dof++, n++);
    flags = phgAlloc(n * sizeof(*flags));

    /* flags indicating whether DOFs have valid data */
    if (g->rank == 0) {
	for (i = 0, dof = g->dof; *dof != NULL; dof++, i++)
	    flags[i] = (!SpecialDofType((*dof)->type)
			&& (*dof)->data != NULL);
    }
#if DEBUG
    /* This helps finding out unfreed DOFs */
    i = n;
    MPI_Bcast(&i, 1, MPI_INT, 0, g->comm);
    if (i != n) {
	phgWarning(0, "number of DOFs mismatches: %d against %d on proc 0\n",
		   i, n);
	phgError(0, "forgot freeing some DOFs?.\n");
    }
#endif
    MPI_Bcast(flags, n * sizeof(*flags), MPI_BYTE, 0, g->comm);

    /* compute total size of data */
    size = 0;
    for (i = 0, dof = g->dof; *dof != NULL; dof++, i++) {
	if (!flags[i])
	    continue;
	size += DofGetDataCountGlobal(*dof);
    }
    if (size == 0)
	return;
    buffer = phgAlloc(size * sizeof(*buffer));

    if (g->rank == 0) {
	/* copy data to buffer */
	p0 = buffer;
	for (i = 0, dof = g->dof; *dof != NULL; dof++, i++) {
	    if (!flags[i])
		continue;
	    size0 = DofGetDataCountGlobal(*dof);
	    memcpy(p0, (*dof)->data, size0 * sizeof(FLOAT));
	    phgFree((*dof)->data);
	    (*dof)->data = NULL;
	    p0 += size0;
	}
    }

    /* Broadcast data to all processes */
    MPI_Bcast(buffer, size * sizeof(*buffer), MPI_BYTE, 0, g->comm);

    /* copy data to buffer */
    p0 = buffer;
    for (i = 0, dof = g->dof; *dof != NULL; dof++, i++) {
	if (!flags[i])
	    continue;
	size0 = DofGetDataCount(*dof);
	(*dof)->data = phgAlloc(size0 * sizeof(FLOAT));
	(*dof)->data_vert = (*dof)->data;
	(*dof)->data_edge = (*dof)->data_vert + DofGetVertexDataCount(*dof);
	(*dof)->data_face = (*dof)->data_edge + DofGetEdgeDataCount(*dof);
	(*dof)->data_elem = (*dof)->data_face + DofGetFaceDataCount(*dof);
	p = (*dof)->data;
	if ((*dof)->type->np_vert > 0) {
	    size = (*dof)->count_vert;
	    size0 = size * sizeof(FLOAT);
	    for (n = 0; n < g->nvert; n++) {
		memcpy(p, p0 + size * GlobalVertex(g, n), size0);
		p += size;
	    }
	    p0 += size * g->nvert_global;
	}
	if ((*dof)->type->np_edge > 0) {
	    size = (*dof)->count_edge;
	    size0 = size * sizeof(FLOAT);
	    for (n = 0; n < g->nedge; n++) {
		memcpy(p, p0 + size * GlobalEdge(g, n), size0);
		p += size;
	    }
	    p0 += size * g->nedge_global;
	}
	if ((*dof)->type->np_face > 0) {
	    size = (*dof)->count_face;
	    size0 = size * sizeof(FLOAT);
	    for (n = 0; n < g->nface; n++) {
		memcpy(p, p0 + size * GlobalFace(g, n), size0);
		p += size;
	    }
	    p0 += size * g->nface_global;
	}
	if ((*dof)->type->np_elem > 0) {
	    size = (*dof)->count_elem;
	    size0 = size * sizeof(FLOAT);
	    for (n = 0; n < g->nelem; n++) {
		memcpy(p, p0 + size * GlobalElement(g, n), size0);
		p += size;
	    }
	    p0 += size * g->nelem_global;
	}
    }

    phgFree(buffer);
    phgFree(flags);
#endif

    return;
}

static int level = 0;		/* recursion level */
/* common values and pointers used by the interpolation routines */
static int n0, n1, n2, n3;
static FLOAT *data0, *data1, *data2, *data3;

void
update_dof_data(DOF *dof, SIMPLEX *e, FLOAT **data)
/* places DOF data for element 'e' pointed by 'data' to dof->data */
{
    int i, j;
    int size[4];
    int count[] = { NVert, NEdge, NFace, 1 };
    int *index[4];
    FLOAT *loc[4];
    FLOAT *p;

    size[0] = n0;
    size[1] = n1;
    size[2] = n2;
    size[3] = n3;
    index[0] = e->verts;
    index[1] = e->edges;
    index[2] = e->faces;
    index[3] = &e->index;
    loc[0] = data0;
    loc[1] = data1;
    loc[2] = data2;
    loc[3] = data3;

    for (j = 0; j < 4; j++) {
	for (i = 0; i < count[j]; i++) {
	    p = loc[j] + index[j][i] * size[j];
	    if (size[j] > 0 && p != *data) {
		memcpy(p, *data, size[j] * sizeof(FLOAT));
	    }
	    data++;
	}
    }
}

static void
build_child_pointers(DOF *dof, SIMPLEX *e, FLOAT **old_data, FLOAT **new_data,
		     FLOAT *buffer, int child)
/* construct list of pointers for children */
{
    int i, Vmap[NVert], Emap[NEdge];
    FLOAT *p1, *p2, *p3;

    phgMapC2P(e, Vmap, Emap, child);

    p1 = buffer;
    p2 = p1 + n1 * 4;		/* 4 new edges */
    p3 = p2 + n2 * 5;		/* 5 new faces */

    /* vertices */
    if (n0 > 0) {
	for (i = 0; i < NVert - 1; i++) {
	    *(new_data++) = old_data[Vmap[i]];
	}
	*(new_data++) = data0 + e->children[0]->verts[NVert - 1] * n0;
    }
    else {
	new_data += NVert;
    }

    /* edges */
    if (n1 > 0) {
	for (i = 0; i < NEdge; i++) {
	    switch (Emap[i]) {
		case -1:	/* new edge */
		    (*new_data++) = p1 + (Vmap[GetEdgeVertex(i, 0)]) * n1;
		    break;
		case 0:	/* cut edge */
		    (*new_data++) = p1 + child * n1;
		    break;
		default:	/* old edge */
		    (*new_data++) = old_data[NVert + Emap[i]];
		    break;
	    }
	}
    }
    else {
	new_data += NEdge;
    }

    /* faces */
    if (n2 > 0) {
	for (i = 0; i < NFace; i++) {
	    switch (Vmap[i]) {
		case 0:
		case 1:	/* new face */
		    (*new_data++) = p2 + 4 * n2;
		    break;
		case 2:	/* cut face on F2 */
		    (*new_data++) = p2 + child * n2;
		    break;
		case 3:	/* cut face on F3 */
		    (*new_data++) = p2 + (child + 2) * n2;
		    break;
		case -1:	/* old faces */
		    (*new_data++) = old_data[NVert + NEdge + 1 - child];
		    break;
	    }
	}
    }
    else {
	new_data += NFace;
    }

    /* volume */
    if (n3 > 0) {
	(*new_data) = p3 + child * n3;
    }
}

static void
interp_recursive(DOF *dof, SIMPLEX *e, FLOAT **old_data)
/* recursively interpolates DOFs down the binary tree starting from e */
{
    static FLOAT *p1, *p2, *p3;
    static SIMPLEX *e0, *e1;
    FLOAT *buffer, *new_data[15];

    if ((e0 = e->children[0]) == NULL || (e1 = e->children[1]) == NULL) {
	/* copy data to dof->data */
	update_dof_data(dof, e, old_data);
	return;
    }

    level++;

    /* allocate buffer for new DOFs (only allocate for edge/face/vol data) */
    if (n1 > 0 || n2 > 0 || n3 > 0)
	buffer = phgAlloc((n1 * 4 + n2 * 5 + n3 * 2) * sizeof(*buffer));
    else
	buffer = NULL;

    p1 = buffer;
    p2 = p1 + n1 * 4;
    p3 = p2 + n2 * 5;

    memset(new_data, 0, sizeof(new_data));

    if (dof->type->InterpC2F != NULL && dof->userfunc == DofInterpolation) {
	/* interpolation */
	if (n0 > 0) {
	    /* the new vertex */
	    new_data[0] = data0 + e0->verts[3] * n0;
	}

	if (n1 > 0) {
	    /* the cut edge */
	    new_data[1] = p1;
	    new_data[2] = p1 + n1;
	    /* the new edge on V2 */
	    new_data[3] = p1 + n1 * 2;
	    /* the new edge on V3 */
	    new_data[4] = p1 + n1 * 3;
	}

	if (n2 > 0) {
	    /* the cut faces on F2 */
	    new_data[5] = p2;
	    new_data[6] = p2 + n2;
	    /* the cut faces on F3 */
	    new_data[7] = p2 + n2 * 2;
	    new_data[8] = p2 + n2 * 3;
	    /* the new face */
	    new_data[9] = p2 + n2 * 4;
	}

	if (n3 > 0) {
	    /* volume data for child 0 */
	    new_data[10] = p3;
	    /* volume data for child 1 */
	    new_data[11] = p3 + n3;
	}

	dof->type->InterpC2F(dof, e, old_data, new_data);
    }

    if (e0 != NULL) {
	build_child_pointers(dof, e, old_data, new_data, buffer, 0);
	interp_recursive(dof, e0, new_data);
	e1 = e->children[1];	/* note: e1 changed by recursion */
    }

    if (e1 != NULL) {
	build_child_pointers(dof, e, old_data, new_data, buffer, 1);
	interp_recursive(dof, e1, new_data);
    }

    phgFree(buffer);

    level--;
}

void
phgDofInterpolate(GRID *g, BOOLEAN flag)
/* interpolates DOFs, should be called with flag==FALSE
   before refinement, then with flag==TRUE after refinement. */
{
    static SIMPLEX **elems = NULL, **pe, *e;
    static INT count, nvert, nedge, nface, nelem, nleaf;
    DOF **dof;

    /* check whether interpolation is really needed */
    if (g->dof == NULL)
	return;
    for (dof = g->dof; *dof != NULL; dof++) {
	if (!SpecialDofType((*dof)->type) && (*dof)->data != NULL)
	    break;
    }
    if (*dof == NULL)
	return;

    if ((!flag && elems != NULL) || (flag && elems == NULL))
	phgError(1, "phgDofInterpolate: incorrect calling sequence.\n");

    if (!flag) {
	SIMPLEX *e;
	/* store current leaf elements */
	pe = elems = phgAlloc((count = g->nleaf) * sizeof(*elems));
	ForAllElements(g, e)
	    * (pe++) = e;
	assert(pe - elems == count);
	/* save current DOF counts */
	nvert = g->nvert;
	nedge = g->nedge;
	nface = g->nface;
	nelem = g->nelem;
	nleaf = g->nleaf;

	return;
    }

    /* reset cache, map, etc. */
    /* phgDofClearCache(NULL, FALSE); (not necessary: elements not changed) */

    if (g->nleaf == nleaf)
	goto end;		/* no new elements created */

    /* Clear e->flag */
    ForAllElements(g, e)
	e->flag = 0;

    /* loop on DOFs */
    for (dof = g->dof; *dof != NULL; dof++) {
	size_t size, old_size;
	FLOAT *old_data, *p0, *p1, *p2, *p3;

	if (SpecialDofType((*dof)->type) || (*dof)->data == NULL)
	    continue;

	/* allocate new buffer */
	size = DofGetDataCount(*dof);

	if ((*dof)->userfunc == DofNoAction) {
	    phgFree((*dof)->data);
	    (*dof)->data = phgCalloc(size, sizeof(*(*dof)->data));
	    (*dof)->data_vert = (*dof)->data;
	    (*dof)->data_edge = (*dof)->data_vert +
		DofGetVertexDataCount(*dof);
	    (*dof)->data_face = (*dof)->data_edge + DofGetEdgeDataCount(*dof);
	    (*dof)->data_elem = (*dof)->data_face + DofGetFaceDataCount(*dof);
	    continue;
	}

	old_data = (*dof)->data;
	(*dof)->data = phgCalloc(size, sizeof(*(*dof)->data));
	(*dof)->data_vert = (*dof)->data;
	(*dof)->data_edge = (*dof)->data_vert + DofGetVertexDataCount(*dof);
	(*dof)->data_face = (*dof)->data_edge + DofGetEdgeDataCount(*dof);
	(*dof)->data_elem = (*dof)->data_face + DofGetFaceDataCount(*dof);
	/* copy old data to new buffer, because some entries may be only
	   referenced by non-leaf elements */
	p0 = (*dof)->data;
	p1 = old_data;
	if ((*dof)->count_vert > 0) {
	    old_size = (*dof)->count_vert * nvert;
	    memcpy(p0, p1, old_size * sizeof(*old_data));
	    p0 += (*dof)->count_vert * g->nvert;
	    p1 += old_size;
	}
	if ((*dof)->count_edge > 0) {
	    old_size = (*dof)->count_edge * nedge;
	    memcpy(p0, p1, old_size * sizeof(*old_data));
	    p0 += (*dof)->count_edge * g->nedge;
	    p1 += old_size;
	}
	if ((*dof)->count_face > 0) {
	    old_size = (*dof)->count_face * nface;
	    memcpy(p0, p1, old_size * sizeof(*old_data));
	    p0 += (*dof)->count_face * g->nface;
	    p1 += old_size;
	}
	if ((*dof)->count_elem > 0) {
	    old_size = (*dof)->count_elem * nelem;
	    memcpy(p0, p1, old_size * sizeof(*old_data));
	    p0 += (*dof)->count_elem * g->nelem;
	}

	n0 = (*dof)->count_vert;
	n1 = (*dof)->count_edge;
	n2 = (*dof)->count_face;
	n3 = (*dof)->count_elem;

	data0 = (*dof)->data;
	data1 = data0 + n0 * g->nvert;
	data2 = data1 + n1 * g->nedge;
	data3 = data2 + n2 * g->nface;

	p0 = old_data;
	p1 = p0 + n0 * nvert;
	p2 = p1 + n1 * nedge;
	p3 = p2 + n2 * nface;

	for (pe = elems; pe < elems + count; pe++) {
	    FLOAT *old[15];
	    if (n0 > 0) {
		old[0] = p0 + (*pe)->verts[0] * n0;
		old[1] = p0 + (*pe)->verts[1] * n0;
		old[2] = p0 + (*pe)->verts[2] * n0;
		old[3] = p0 + (*pe)->verts[3] * n0;
	    }
	    if (n1 > 0) {
		old[4] = p1 + (*pe)->edges[0] * n1;
		old[5] = p1 + (*pe)->edges[1] * n1;
		old[6] = p1 + (*pe)->edges[2] * n1;
		old[7] = p1 + (*pe)->edges[3] * n1;
		old[8] = p1 + (*pe)->edges[4] * n1;
		old[9] = p1 + (*pe)->edges[5] * n1;
	    }
	    if (n2 > 0) {
		old[10] = p2 + (*pe)->faces[0] * n2;
		old[11] = p2 + (*pe)->faces[1] * n2;
		old[12] = p2 + (*pe)->faces[2] * n2;
		old[13] = p2 + (*pe)->faces[3] * n2;
	    }
	    if (n3 > 0) {
		old[14] = p3 + (*pe)->index * n3;
	    }
	    (*pe)->flag = 1;	/* element is on old mesh */
	    interp_recursive(*dof, *pe, old);
	    /* TODO: make use of flags[] to avoid redundant computations */
	}
	phgFree(old_data);
	flagged_only = TRUE;
	if ((*dof)->userfunc_lambda != NULL)
	    phgDofSetDataByLambdaFunction(*dof, (*dof)->userfunc_lambda);
	else
	    phgDofSetDataByFunction(*dof, (*dof)->userfunc);
	flagged_only = FALSE;
    }

  end:
    /* Free memory */
    phgFree(elems);
    elems = NULL;
}

#if USE_MPI
static INT *current_map;
static int
sort_map_indices_comp(const void *p0, const void *p1)
{
    return current_map[*((const INT *)p0)] - current_map[*((const INT *)p1)];
}

static INT *
sort_map_indices(INT count, INT *map)
{
    INT i;

    if (count == 0)
	return NULL;

    current_map = map;
    map = phgAlloc(count * sizeof(*map));
    for (i = 0; i < count; i++)
	map[i] = i;
    if (current_map != NULL)
	qsort(map, count, sizeof(*map), sort_map_indices_comp);

    return map;
}

static void
copy_dof_data(DOF *dof, FLOAT *d0, INT n0[4], INT *index0[4], INT *map0[4],
	      BYTE *type0[4],
	      FLOAT *d1, INT n1[4], INT *index1[4], INT *map1[4])
/* copy data at 'd0' indexed with 'index0' and 'map0' to
	data at 'd1' indexed with 'index1' and 'map1'.

   Note: index0 or index1 is allowed to be NULL meaning identity map */
{
    INT *M0, *M1, *I, *J, N0, N1;
    BYTE *T0;
    INT dof_size[4];
    int i, j, i0 = 0, j0, ii;
    size_t size;

    dof_size[0] = dof->type->np_vert;
    dof_size[1] = dof->type->np_edge;
    dof_size[2] = dof->type->np_face;
    dof_size[3] = dof->type->np_elem;

    for (ii = 0; ii < 4; ii++) {
	M0 = *(map0++);
	M1 = *(map1++);
	I = index0 == NULL ? NULL : *(index0++);
	J = index1 == NULL ? NULL : *(index1++);
	N0 = *(n0++);
	N1 = *(n1++);
	T0 = (type0 == NULL) ? NULL : *(type0++);
	size = dof_size[ii] * dof->dim;
	/* copy data */
	j = 0;
	i = 0;
	while (j < N1) {
	    j0 = (J == NULL ? j : J[j]);
	    while (i < N0 && M0[i0 = (I == NULL ? i : I[i])] < M1[j0])
		i++;
	    if (i >= N0)
		break;
	    if (M0[i0] > M1[j0] || (T0 != NULL && T0[i0] == UNREFERENCED)) {
		j++;
		continue;
	    }
	    /* copy data */
	    memcpy(d1 + j0 * size, d0 + i0 * size, size * sizeof(*d0));
	    j++;
	}
	d0 += size * N0;
	d1 += size * N1;
    }
}
#endif

#define IGNORE_NOACTION	1

void
phgDofRedistribute(GRID *g, BOOLEAN flag)
/* redistributes DOFs, should be called with flag==FALSE
 * before the grid is changed, then with flag==TRUE after.
 *
 * Note: the communications may be merged with those in phgRedistributeGrid
 * 	 for better efficiency.
 */
{
#if USE_MPI
    /* old number of elements/vertices/edges/faces */
    static INT nelem = 0, nvert = 0, nedge = 0, nface = 0;
    /* old maps */
    static INT	*L2Gmap_elem = NULL, *L2Gmap_vert = NULL,
		*L2Gmap_edge = NULL, *L2Gmap_face = NULL;
    /* old boundary types */
    static BTYPE *types_elem = NULL, *types_vert = NULL,
		 *types_edge = NULL, *types_face = NULL;

    /* lists of vertices/edges/faces/elements to send.
     * The length of lists is g->nprocs. lists[i] describes data to be sent
     * to proc i, which is an array of INT organized as follows:
     *		lists[i][0] = number of vertices
     *		lists[i][1] = number of edges
     *		lists[i][2] = number of faces
     *		lists[i][3] = number of elements
     * followed by:
     *		list of vertices
     *		list of edges 
     *		list of faces 
     *		list of elements
     * The total length of list[i] is:
     *		4 + lists[i][0] + lists[i][1] + lists[i][2] + lists[i][3])
     */
    static INT **lists = NULL, *list;
    /* sorted old local indices */
    static INT *index0[4];

    /* blksize = max(sizeof(INT), sizeof(FLOAT)) */
    static size_t blksize = (sizeof(INT) > sizeof(FLOAT) ?
						sizeof(INT) : sizeof(FLOAT));

    BTYPE *type0[4];

    /* pointers to old counts and maps */
    INT n0[4], *map0[4];
    /* pointers to new counts, maps and sorted local indices */
    INT n1[4], *map1[4], *index1[4];
    /* pointers to temperary maps */
    INT *map2[4];

    DOF **dof;
    INT i, j;
    int ii, jj, ndof_total, cnt, cnt1;
    /* data count per vertex/edge/face/element (Note: not suitable for h-p) */
    int data_count[4] = {0, 0, 0, 0};
    /* data counts to/from each process */
    int (*sdata_counts)[5], (*rdata_counts)[5];
    int *scounts, *sdispls, *rcounts, *rdispls;
    BYTE *sbuff, *rbuff, *buff;
    size_t size;

    MPI_Datatype type;

    phgDofClearCache(g, NULL, NULL, FALSE);

    for (dof = g->dof, ndof_total = 0;
	 dof != NULL && *dof != NULL;
	 dof++, ndof_total++) {
#if IGNORE_NOACTION
	if ((*dof)->userfunc == DofNoAction)
	    continue;
#endif	/* IGNORE_NOACTION */
	if (SpecialDofType((*dof)->type))
	    continue;
	data_count[0] += (*dof)->count_vert;
	data_count[1] += (*dof)->count_edge;
	data_count[2] += (*dof)->count_face;
	data_count[3] += (*dof)->count_elem;
    }

    /* initialize arrays */
    n0[0] = nvert;
    n0[1] = nedge;
    n0[2] = nface;
    n0[3] = nelem;
    n1[0] = g->nvert;
    n1[1] = g->nedge;
    n1[2] = g->nface;
    n1[3] = g->nelem;
    map0[0] = L2Gmap_vert;
    map0[1] = L2Gmap_edge;
    map0[2] = L2Gmap_face;
    map0[3] = L2Gmap_elem;
    type0[0] = types_vert;
    type0[1] = types_edge;
    type0[2] = types_face;
    type0[3] = types_elem;
    map1[0] = g->L2Gmap_vert;
    map1[1] = g->L2Gmap_edge;
    map1[2] = g->L2Gmap_face;
    map1[3] = g->L2Gmap_elem;

    if (flag == FALSE) {
	BYTE *flags, *flags0;	/* used to mark vertices/edges/faces/elements */
	BYTE v0, v1, v2;
	SIMPLEX *e;
	INT k;

	/* sort old maps */
	index0[0] = sort_map_indices(g->nvert, g->L2Gmap_vert);
	index0[1] = sort_map_indices(g->edge_flag ? g->nedge:0, g->L2Gmap_edge);
	index0[2] = sort_map_indices(g->face_flag ? g->nface:0, g->L2Gmap_face);
	index0[3] = sort_map_indices(g->elem_flag ? g->nelem:0, g->L2Gmap_elem);

	/* construct lists of vertices/edges/faces/elements to send */
	lists = phgCalloc(g->nprocs, sizeof(*lists));
	size = g->nface;
	if (size < g->nvert)
	    size = g->nvert;
	if (size < g->nedge)
	    size = g->nedge;
	if (size < g->nelem)
	    size = g->nelem;
	
	flags = phgAlloc((size + g->nprocs) * sizeof(*flags));
	/* determine to which processes have data to send */
	flags0 = flags + size;
	memset(flags0, 0, g->nprocs * sizeof(*flags0));
	ForAllElements(g, e) {
	    if (e->mark == g->rank)
		continue;
	    flags0[e->mark] = 1;
	}
	/* construct individual lists */
	for (ii = 0; ii < g->nprocs; ii++) {
	    RNEIGHBOUR *rn, *rn1;
	    if (flags0[ii] == 0) {
		lists[ii] = phgCalloc(4, sizeof(*lists[ii]));
		continue;
	    }
	    /* construct list for process ii:
	     *		bit 0/1/2/3 = vert/edge/face/elem */
	    memset(flags, 0, size * sizeof(*flags));
	    ForAllElements(g, e) {
		if (e->mark != ii)
		    continue;
		if (data_count[0] > 0)
		    for (jj = 0; jj < NVert; jj++)
			flags[e->verts[jj]] |= 1;
		if (data_count[1] > 0)
		    for (jj = 0; jj < NEdge; jj++)
			flags[e->edges[jj]] |= 2;
		if (data_count[2] > 0)
		    for (jj = 0; jj < NFace; jj++)
			flags[e->faces[jj]] |= 4;
		if (data_count[3] > 0)
		    flags[e->index] |= 8;
	    }
	    /* remove vertices/edges/faces on shared faces (remote process
	     * already has data) */
	    rn = g->neighbours.list + g->neighbours.displs[ii];
	    rn1 = rn + g->neighbours.counts[ii];
	    for (; rn < rn1; rn++) {
		if (rn->rank != ii)
		    continue;
		jj = rn->vertex;
		e = rn->local;
		v0 = GetFaceVertex(jj, 0);
		v1 = GetFaceVertex(jj, 1);
		v2 = GetFaceVertex(jj, 2);
		/* remove vertices on the face from the list */
		if (data_count[0] > 0) {
		    flags[e->verts[v0]] &= ~1;
		    flags[e->verts[v1]] &= ~1;
		    flags[e->verts[v2]] &= ~1;
		}
		/* remove edges on the face from the list */
		if (data_count[1] > 0) {
		    flags[e->edges[GetEdgeNo(v0, v1)]] &= ~2;
		    flags[e->edges[GetEdgeNo(v0, v2)]] &= ~2;
		    flags[e->edges[GetEdgeNo(v1, v2)]] &= ~2;
		}
		/* remove the face */
		if (data_count[2] > 0) {
		    flags[e->faces[jj]] &= ~4;
		}
	    }
	    /* count # of vertices/edges/faces/elements */
	    nvert = 0;
	    if (data_count[0] > 0) {
		for (i = 0; i < g->nvert; i++) {
		    if (flags[i] & 1)
			nvert++;
		}
	    }
	    nedge = 0;
	    if (data_count[1] > 0) {
		for (i = 0; i < g->nedge; i++) {
		    if (flags[i] & 2)
			nedge++;
		}
	    }
	    nface = 0;
	    if (data_count[2] > 0) {
		for (i = 0; i < g->nface; i++) {
		    if (flags[i] & 4)
			nface++;
		}
	    }
	    nelem = 0;
	    if (data_count[3] > 0) {
		for (i = 0; i < g->nelem; i++) {
		    if (flags[i] & 8)
			nelem++;
		}
	    }
	    /* allocate list */
	    lists[ii] = list = phgAlloc((4 + nvert + nedge + nface + nelem)
					* sizeof(*lists[ii]));
	    list[0] = nvert;
	    list[1] = nedge;
	    list[2] = nface;
	    list[3] = nelem;
	    j = 4;
	    if (nvert > 0) {
		for (i = 0; i < g->nvert; i++) {
		    k = index0[0][i];
		    if (flags[k] & 1)
			list[j++] = GlobalVertex(g, k);
		}
	    }
	    if (nedge > 0) {
		for (i = 0; i < g->nedge; i++) {
		    k = index0[1][i];
		    if (flags[k] & 2)
			list[j++] = GlobalEdge(g, k);
		}
	    }
	    if (nface > 0) {
		for (i = 0; i < g->nface; i++) {
		    k = index0[2][i];
		    if (flags[k] & 4)
			list[j++] = GlobalFace(g, k);
		}
	    }
	    if (nelem > 0) {
		for (i = 0; i < g->nelem; i++) {
		    k = index0[3][i];
		    if (flags[k] & 8)
			list[j++] = GlobalElement(g, k);
		}
	    }
	    assert(j == 4 + list[0] + list[1] + list[2] + list[3]);
	}
	phgFree(flags);

	/* save maps and types arrays */
	if (g->elem_flag) {
	    /* save L2Gmap_elem and types_elem */
	    size = (nelem = g->nelem) * sizeof(*L2Gmap_elem);
	    phgFree(L2Gmap_elem);
	    L2Gmap_elem = phgAlloc(size);
	    if (g->L2Gmap_elem != NULL) {
		memcpy(L2Gmap_elem, g->L2Gmap_elem, size);
	    }
	    else {
		for (i = 0; i < nelem; i++)
		    L2Gmap_elem[i] = i;
	    }
	    types_elem = phgAlloc(nelem * sizeof(*types_elem));
	    memcpy(types_elem, g->types_elem, nelem * sizeof(*types_elem));
	}

	/* save L2Gmap_vert and types_vert */
	size = (nvert = g->nvert) * sizeof(*L2Gmap_vert);
	phgFree(L2Gmap_vert);
	L2Gmap_vert = phgAlloc(size);
	if (g->L2Gmap_vert != NULL) {
	    memcpy(L2Gmap_vert, g->L2Gmap_vert, size);
	}
	else {
	    for (i = 0; i < nvert; i++)
		L2Gmap_vert[i] = i;
	}
	types_vert = phgAlloc(nvert * sizeof(*types_vert));
	memcpy(types_vert, g->types_vert, nvert * sizeof(*types_vert));

	if (g->edge_flag) {
	    /* save L2Gmap_edge and types_edge */
	    size = (nedge = g->nedge) * sizeof(*L2Gmap_edge);
	    phgFree(L2Gmap_edge);
	    L2Gmap_edge = phgAlloc(size);
	    if (g->L2Gmap_edge != NULL) {
		memcpy(L2Gmap_edge, g->L2Gmap_edge, size);
	    }
	    else {
		for (i = 0; i < nedge; i++)
		    L2Gmap_edge[i] = i;
	    }
	    types_edge = phgAlloc(nedge * sizeof(*types_edge));
	    memcpy(types_edge, g->types_edge, nedge * sizeof(*types_edge));
	}

	if (g->face_flag) {
	    /* save L2Gmap_face and types_face */
	    size = (nface = g->nface) * sizeof(*L2Gmap_face);
	    phgFree(L2Gmap_face);
	    L2Gmap_face = phgAlloc(size);
	    if (g->L2Gmap_face != NULL) {
		memcpy(L2Gmap_face, g->L2Gmap_face, size);
	    }
	    else {
		for (i = 0; i < nface; i++)
		    L2Gmap_face[i] = i;
	    }
	    types_face = phgAlloc(nface * sizeof(*types_face));
	    memcpy(types_face, g->types_face, nface * sizeof(*types_face));
	}

	return;
    }

    phgDebug((3, "old: nvert=%d, nedge=%d, nface=%d, nelem=%d\n",
	      n0[0], n0[1], n0[2], n0[3]));
    phgDebug((3, "new: nvert=%d, nedge=%d, nface=%d, nelem=%d\n",
	      n1[0], n1[1], n1[2], n1[3]));

    if (!g->edge_flag)
	n0[1] = n1[1] = 0;
    if (!g->face_flag)
	n0[2] = n1[2] = 0;
    if (!g->elem_flag)
	n0[3] = n1[3] = 0;

    /* sort new maps */
    for (ii = 0; ii < 4; ii++) {
	index1[ii] = sort_map_indices(n1[ii], map1[ii]);
    }

    /* exchange data counts */
    sdata_counts = phgAlloc(2 * g->nprocs * sizeof(*sdata_counts)); 
    rdata_counts = sdata_counts + g->nprocs;
    for (ii = 0; ii < g->nprocs; ii++) {
	list = lists[ii];
	sdata_counts[ii][0] = list[0];
	sdata_counts[ii][1] = list[1];
	sdata_counts[ii][2] = list[2];
	sdata_counts[ii][3] = list[3];
	sdata_counts[ii][4] = ndof_total;
    }
    MPI_Alltoall(sdata_counts, 5, MPI_INT, rdata_counts, 5, MPI_INT, g->comm);

    /* check that the number of DOFs match */
    cnt = cnt1 = rdata_counts[0][4];
    for (ii = 1; ii < g->nprocs; ii++) {
	if (cnt > rdata_counts[ii][4])
	    cnt = rdata_counts[ii][4];	/* cnt = min ndof_total */
	else if (cnt1 < rdata_counts[ii][4])
	    cnt1 = rdata_counts[ii][4];	/* cnt1 = max ndof_total */
    }
    if (cnt != cnt1) {
	if (g->rank == 0) {
	    assert(cnt1 == ndof_total);
	    phgInfo(-1, "Fatal error: DOFs mismatch between submeshes.\n");
	    phgInfo(-1, "The following DOFs were (possibly) not "
			"properly freed:\n");
	    for (ii = cnt; ii < cnt1; ii++)
		phgInfo(-1, "\t\"%s\" (created at %s:%d)\n", g->dof[ii]->name,
				g->dof[ii]->srcfile, g->dof[ii]->srcline);
	    phgInfo(-1, "Note: a DOF created after phgRedistributeGrid must be "
			"freed before the next call to phgRedistributeGrid.\n");
	    phgInfo(-1, "Abort.\n");
	}
	MPI_Barrier(g->comm);
	MPI_Abort(MPI_COMM_WORLD, 1);
	exit(1);
    }

    assert(sizeof(FLOAT) % sizeof(INT) == 0 ||
	   sizeof(INT) % sizeof(FLOAT) == 0);

    /* prepare lists and data to send */
    scounts = phgAlloc(4 * g->nprocs * sizeof(*scounts));
    sdispls = scounts + g->nprocs * 1;
    rcounts = scounts + g->nprocs * 2;
    rdispls = scounts + g->nprocs * 3;
    for (ii = 0; ii < g->nprocs; ii++) {
	/* send size */
	size =	/* size for the list */
	       (sdata_counts[ii][0] + sdata_counts[ii][1] +
		sdata_counts[ii][2] + sdata_counts[ii][3]) * sizeof(INT) + 
		/* size for the data */
	       (sdata_counts[ii][0] * data_count[0] +
		sdata_counts[ii][1] * data_count[1] +
		sdata_counts[ii][2] * data_count[2] +
		sdata_counts[ii][3] * data_count[3]) * sizeof(FLOAT);
	if ((jj = size % blksize) != 0)
	    size += blksize - jj;
	size /= blksize;
	scounts[ii] = size;
	sdispls[ii] = (ii == 0 ? 0 : sdispls[ii - 1] + scounts[ii - 1]);

	/* receive size */
	size =	/* size for the list */
	       (rdata_counts[ii][0] + rdata_counts[ii][1] +
		rdata_counts[ii][2] + rdata_counts[ii][3]) * sizeof(INT) + 
		/* size for the data */
	       (rdata_counts[ii][0] * data_count[0] +
		rdata_counts[ii][1] * data_count[1] +
		rdata_counts[ii][2] * data_count[2] +
		rdata_counts[ii][3] * data_count[3]) * sizeof(FLOAT);
	if ((jj = size % blksize) != 0)
	    size += blksize - jj;
	size /= blksize;
	rcounts[ii] = size;
	rdispls[ii] = (ii == 0 ? 0 : rdispls[ii - 1] + rcounts[ii - 1]);
    }

    /* copy lists and data to send */
    size = sdispls[g->nprocs - 1] + scounts[g->nprocs - 1];
    sbuff = phgAlloc(size * blksize);
    for (ii = 0; ii < g->nprocs; ii++) {
	if (scounts[ii] == 0)
	    continue;
	/* copy list */
	buff = sbuff + sdispls[ii] * blksize;
	size = (sdata_counts[ii][0] + sdata_counts[ii][1] +
                sdata_counts[ii][2] + sdata_counts[ii][3]) * sizeof(INT);
	memcpy(buff, lists[ii] + 4, size);
	if ((jj = size % sizeof(FLOAT)) != 0)
	    size += sizeof(FLOAT) - jj;
	/* copy DOF data */
	for (dof = g->dof; dof != NULL && *dof != NULL; dof++) {
#if IGNORE_NOACTION
	    if ((*dof)->userfunc == DofNoAction)
		continue;
#endif	/* IGNORE_NOACTION */
	    if (SpecialDofType((*dof)->type))
		continue;
	    map2[0] = (INT *)buff;
	    map2[1] = map2[0] + sdata_counts[ii][0];
	    map2[2] = map2[1] + sdata_counts[ii][1];
	    map2[3] = map2[2] + sdata_counts[ii][2];
	    copy_dof_data(*dof, (*dof)->data, n0, index0, map0, type0,
			  (FLOAT *)(buff + size), sdata_counts[ii], NULL, map2);
	    size += (sdata_counts[ii][0] * (*dof)->count_vert +
		     sdata_counts[ii][1] * (*dof)->count_edge +
		     sdata_counts[ii][2] * (*dof)->count_face +
		     sdata_counts[ii][3] * (*dof)->count_elem) * sizeof(FLOAT);
	}
	assert((size + blksize - 1) / blksize == scounts[ii]);
    }

    /* free lists */
    for (ii = 0; ii < g->nprocs; ii++)
	phgFree(lists[ii]);
    phgFree(lists);
    lists = NULL;

    /* send lists and data */
    size = rdispls[g->nprocs - 1] + rcounts[g->nprocs - 1];
    rbuff = phgAlloc(size * blksize);
    MPI_Type_contiguous(blksize, MPI_BYTE, &type);
    MPI_Type_commit(&type);
    phgAlltoallv(sbuff, scounts, sdispls, type, rbuff, rcounts, rdispls, type,
		 g->comm);
    MPI_Type_free(&type);

    /* free send buffer */
    phgFree(sbuff);

    /* copy local DOF data */
    for (dof = g->dof; dof != NULL && *dof != NULL; dof++) {
	FLOAT *old_data, *new_data;
#if IGNORE_NOACTION
	if ((*dof)->userfunc == DofNoAction) {
	    if (!SpecialDofType((*dof)->type)) {
		phgFree((*dof)->data);
		(*dof)->data = phgCalloc(DofGetDataCount(*dof),
					 sizeof(*old_data));
		(*dof)->data_vert = (*dof)->data;
		(*dof)->data_edge = (*dof)->data_vert +
						DofGetVertexDataCount(*dof);
		(*dof)->data_face = (*dof)->data_edge +
						DofGetEdgeDataCount(*dof);
		(*dof)->data_elem = (*dof)->data_face +
						DofGetFaceDataCount(*dof);
	    }
	    continue;
	}
#endif	/* IGNORE_NOACTION */
	if (SpecialDofType((*dof)->type))
	    continue;
	old_data = (*dof)->data;
	(*dof)->data = phgAlloc(DofGetDataCount(*dof) * sizeof(*old_data));
#if DEBUG
	for (i = 0; i < DofGetDataCount(*dof); i++)
	    (*dof)->data[i] = 1e99;
#endif
	(*dof)->data_vert = new_data = (*dof)->data;
	(*dof)->data_edge = (*dof)->data_vert + DofGetVertexDataCount(*dof);
	(*dof)->data_face = (*dof)->data_edge + DofGetEdgeDataCount(*dof);
	(*dof)->data_elem = (*dof)->data_face + DofGetFaceDataCount(*dof);
	copy_dof_data(*dof, old_data, n0, index0, map0, type0,
			    new_data, n1, index1, map1);
	phgFree(old_data);
    }

    /* copy received data */
    for (ii = 0; ii < g->nprocs; ii++) {
	if (rcounts[ii] == 0)
	    continue;
	/* copy list */
	buff = rbuff + rdispls[ii] * blksize;
	size = (rdata_counts[ii][0] + rdata_counts[ii][1] +
                rdata_counts[ii][2] + rdata_counts[ii][3]) * sizeof(INT);
	if ((jj = size % sizeof(FLOAT)) != 0)
	    size += sizeof(FLOAT) - jj;
	for (dof = g->dof; dof != NULL && *dof != NULL; dof++) {
#if IGNORE_NOACTION
	    if ((*dof)->userfunc == DofNoAction)
		continue;
#endif	/* IGNORE_NOACTION */
	    if (SpecialDofType((*dof)->type))
		continue;
	    map2[0] = (INT *)buff;
	    map2[1] = map2[0] + rdata_counts[ii][0];
	    map2[2] = map2[1] + rdata_counts[ii][1];
	    map2[3] = map2[2] + rdata_counts[ii][2];
	    copy_dof_data(*dof, (FLOAT *)(buff + size), rdata_counts[ii], NULL,
				map2, NULL, (*dof)->data, n1, index1, map1);
	    size += (rdata_counts[ii][0] * (*dof)->count_vert +
		     rdata_counts[ii][1] * (*dof)->count_edge +
		     rdata_counts[ii][2] * (*dof)->count_face +
		     rdata_counts[ii][3] * (*dof)->count_elem) * sizeof(FLOAT);
	}
	assert((size + blksize - 1) / blksize == rcounts[ii]);
    }

    /* free memory blocks and reset variables */
    phgFree(rbuff);
    phgFree(scounts);
    phgFree(sdata_counts);

    for (i = 0; i < 4; i++) {
	phgFree(map0[i]);
	phgFree(index0[i]);
	phgFree(type0[i]);
	phgFree(index1[i]);
    }

    L2Gmap_elem = L2Gmap_vert = L2Gmap_edge = L2Gmap_face = NULL;
    types_elem = types_vert = types_edge = types_face = NULL;
    nelem = nvert = nedge = nface = 0;
#else	/* USE_MPI */
    Unused(g);
    Unused(flag);
#endif	/* USE_MPI */

    return;
}

/*----------------- utilities ---------------*/

INT
phgDofMapE2D(DOF *u, SIMPLEX *e, int index)
/* returns DOF index of the unknown with element index 'index' on 'e' */
{
    GRID *g = u->g;
    int m, n;
    INT index0;

    if (index < (n = (m = u->count_vert) * NVert)) {
	return m * e->verts[index / m] + (index % m);
    }
    index -= n;
    index0 = m * g->nvert;

    if (index < (n = (m = u->count_edge) * NEdge)) {
	if (u->type->np_edge > 1) {
	    int k = index / m, M[u->type->np_edge];
	    phgDofMapEdgeData(u->type, e, k, M);
	    index %= m;
	    return index0 + m * e->edges[k] + M[index / u->dim] * u->dim +
			index % u->dim;
	}
	else {
	    return index0 + m * e->edges[index / m] + (index % m);
	}
    }
    index -= n;
    index0 += m * g->nedge;

    if (index < (n = (m = u->count_face) * NFace)) {
	if (u->type->np_face > 1) {
	    int k = index / m, M[u->type->np_face];
	    phgDofMapFaceData(u->type, e, k, M);
	    index %= m;
	    return index0 + m * e->faces[k] + M[index / u->dim] * u->dim +
			index % u->dim;
	}
	else {
	    return index0 + m * e->faces[index / m] + (index % m);
	}
    }
    index -= n;
    index0 += m * g->nface;

    return index0 + u->count_elem * e->index + index;
}

BTYPE
phgDofGetBoundaryType(DOF *u, INT index)
/* returns boundary type of the unknown with local index 'index' */
{
    INT n;
    DOF_TYPE *type = u->type;

    if (type != NULL && type->base_type != NULL) {
	/* Special case: type = DGn */
	SIMPLEX *e = NULL;
	assert(type->np_vert == 0 && type->np_edge == 0 && type->np_face == 0);
	n = index / u->count_elem;	/* the element index */
	if ((e = u->g->elems[n]) == NULL)
	    return UNREFERENCED;
	return phgDofGetElementBoundaryType(u, e, index % u->count_elem);
    }

    if (index < (n = DofGetVertexDataCount(u)))
	return u->g->types_vert[index / u->count_vert];
    index -= n;

    if (index < (n = DofGetEdgeDataCount(u)))
	return u->g->types_edge[index / u->count_edge];
    index -= n;

    if (index < (n = DofGetFaceDataCount(u)))
	return u->g->types_face[index / u->count_face];
    index -= n;

    assert(index < DofGetElementDataCount(u));
    return u->g->types_elem[index / u->count_elem];
}

BTYPE
phgDofGetElementBoundaryType(DOF *u, SIMPLEX *e, int index)
/* returns boundary type of the unknown with local 'index' on element 'e' */
{
    GRID *g = u->g;
    DOF_TYPE *type = u->type;
    int m, n;

    if (type != NULL && type->base_type != NULL)
	type = type->base_type;

    if (index < (n = (m = u->dim * type->np_vert) * NVert))
	return g->types_vert[e->verts[index / m]];
    index -= n;

    if (index < (n = (m = u->dim * type->np_edge) * NEdge))
	return g->types_edge[e->edges[index / m]];
    index -= n;

    if (index < (n = (m = u->dim * type->np_face) * NFace))
	return g->types_face[e->faces[index / m]];
    index -= n;

    return g->types_elem[e->index];
}

GTYPE
phgDofGetElementGeometryType(DOF *u, int index, int *eindex)
/* returns geometric type and index (in '*eindex') of the unknown of 'u'
 * with local 'index' on element 'e' */
{
    int m, n;
    DOF_TYPE *type = u->type;

    if (type != NULL && type->base_type != NULL)
	type = type->base_type;

    if (index < (n = (m = u->dim * type->np_vert) * NVert)) {
	*eindex = index / m;
	return VERTEX;
    }
    index -= n;

    if (index < (n = (m = u->dim * type->np_edge) * NEdge)) {
	*eindex = index / m;
	return EDGE;
    }
    index -= n;

    if (index < (n = (m = u->dim * type->np_face) * NFace)) {
	*eindex = index / m;
	return FACE;
    }
    index -= n;

    *eindex = index / (u->dim * type->np_elem);

    return ELEMENT;
}

FLOAT *
phgDofGetElementCoordinates(DOF *u, SIMPLEX *e, int index)
/* returns world coordinates of the unknown with local 'index' on element 'e' */
{
    static FLOAT c[Dim];
    GRID *g = u->g;
    FLOAT *c0, *c1, *c2, *c3, *L;
    int m, n;
    DOF_TYPE *type = u->type;

    if (type != NULL && type->base_type != NULL)
	type = type->base_type;

    if (index < (n = (m = u->dim * type->np_vert) * NVert))
	return g->verts[e->verts[index / m]];
    index -= n;

    if (type->points == NULL)
	phgError(1, "%s:%d, DOF points for DOF \"%s\" undefined.\n",
		 __FILE__, __LINE__, type->name);

    if (index < (n = (m = u->dim * type->np_edge) * NEdge)) {
	L = type->points + 1 + ((index / u->dim) % type->np_edge) * 2;
	c0 = g->verts[e->verts[GetEdgeVertex(index /= m, 0)]];
	c1 = g->verts[e->verts[GetEdgeVertex(index, 1)]];
	c[0] = c0[0] * L[0] + c1[0] * L[1];
	c[1] = c0[1] * L[0] + c1[1] * L[1];
	c[2] = c0[2] * L[0] + c1[2] * L[1];
	return c;
    }
    index -= n;
    L = type->points + 1 + type->np_edge * 2;

    if (index < (n = (m = u->dim * type->np_face) * NFace)) {
	L += ((index / u->dim) % type->np_face) * 3;
	c0 = g->verts[e->verts[GetFaceVertex(index /= m, 0)]];
	c1 = g->verts[e->verts[GetFaceVertex(index, 1)]];
	c2 = g->verts[e->verts[GetFaceVertex(index, 2)]];
	c[0] = c0[0] * L[0] + c1[0] * L[1] + c2[0] * L[2];
	c[1] = c0[1] * L[0] + c1[1] * L[1] + c2[1] * L[2];
	c[2] = c0[2] * L[0] + c1[2] * L[1] + c2[2] * L[2];
	return c;
    }
    index -= n;
    L += type->np_face * 3 + ((index / u->dim) % type->np_elem) * 4;

    c0 = g->verts[e->verts[0]];
    c1 = g->verts[e->verts[1]];
    c2 = g->verts[e->verts[2]];
    c3 = g->verts[e->verts[3]];
    c[0] = c0[0] * L[0] + c1[0] * L[1] + c2[0] * L[2] + c3[0] * L[3];
    c[1] = c0[1] * L[0] + c1[1] * L[1] + c2[1] * L[2] + c3[1] * L[3];
    c[2] = c0[2] * L[0] + c1[2] * L[1] + c2[2] * L[2] + c3[2] * L[3];
    return c;
}

/*----------------- functions for exchanging neighbour data ---------------*/

NEIGHBOUR_DATA *
phgDofGetNeighbourData(DOF *dof, MAP *map)
/* prepares neighbour data on.
 *
 * Note:
 *	The data are stored in the order as return by phgDofGetBasesOnFace
 *	such that the ordering is identical on all elements.
 *
 *	If 'map' != NULL then the corresponding global vector indices of the
 *	data are stored in the array 'index'
 */
{
    NEIGHBOUR_DATA *nd;
    DOF_TYPE *t;

#if USE_MPI
    GRID *g = dof->g;
    BYTE *send, *recv, *p, *q;
    SIMPLEX *e;
    size_t size, size0, size1;
    INT i, j;
    MPI_Datatype type;
    struct {
	SIMPLEX *e;
	BYTE	op_vertex;
    } *info;
#endif

    nd = phgCalloc(1, sizeof(*nd));
    nd->dof = dof;

    if (map != NULL) {
	for (nd->dof_no = 0; nd->dof_no < map->ndof; nd->dof_no++)
	    if (map->dofs[nd->dof_no] == dof)
		break;
	assert(nd->dof_no < map->ndof);
	nd->map = map;
    }

    t = dof->type;
    assert(t != NULL);
    if (t->base_type != NULL)
	t = t->base_type;

    /* Note: the test is to be removed in the future, after adding an
     * appropriate flag in DOF_TYPE indicating which bases functions are
     * nonzero on a given face */
    if (t != DOF_P0)
	nd->nbas = 3 * (t->np_vert + t->np_edge) + t->np_face;
    else
	nd->nbas = 1;		/* special case */

#if USE_MPI
    if (g->nprocs <= 1)
	return nd;

    size0 = sizeof(*nd->data) * nd->nbas * dof->dim;
    size1 = (map == NULL ? 0 : sizeof(*nd->index) * nd->nbas);
    size = size0 + size1 + sizeof(*info);
    send = phgAlloc(size * g->neighbours.count);
    recv = phgAlloc(size * g->neighbours.count);

    /* gather data from remote neighbours */
    p = send;
    q = p + size0 + sizeof(*info);
    for (i = 0; i < g->neighbours.count; i++) {
	SHORT bases[nd->nbas];
	e = g->neighbours.list[i].local;
	phgDofGetBasesOnFace(t, e, g->neighbours.list[i].vertex, bases);
	for (j = 0; j < nd->nbas; j++) {
	    INT k = phgDofMapE2D(dof, e, bases[j] * dof->dim);
	    memcpy(p, dof->data + k, dof->dim * sizeof(*dof->data));
	    p += dof->dim * sizeof(*dof->data);
	    if (map != NULL) {
		*((INT *)q) = k;
		q += sizeof(*nd->index);
	    }
	}
	info = (void *)p;
	info->e = g->neighbours.list[i].remote;
	info->op_vertex = g->neighbours.list[i].op_vertex;
	p = q;
	q += size0 + sizeof(*info);
    }
    MPI_Type_contiguous(size, MPI_BYTE, &type);
    MPI_Type_commit(&type);
    phgAlltoallv(send, g->neighbours.counts, g->neighbours.displs, type,
		 recv, g->neighbours.counts, g->neighbours.displs, type,
		 g->comm);
    MPI_Type_free(&type);
    phgFree(send);
    /* copy and reorder received data */
    nd->data = phgAlloc(size0 * g->neighbours.count);
    if (map != NULL)
	nd->index = phgAlloc(size1 * g->neighbours.count);
    for (i = 0, p = recv; i < g->neighbours.count; i++) {
	info = (void *)(p + nd->nbas * sizeof(*nd->data) * dof->dim);
	e = info->e;
	j = GetRNeighbour(g, e, info->op_vertex) - g->neighbours.list;
	memcpy(nd->data + j * nd->nbas * dof->dim, p, size0);
	p += size;
	if (map != NULL)
	    memcpy(nd->index + j * nd->nbas, p - size1, size1);
    }
    phgFree(recv);
#endif

    return nd;
}

void
phgDofReleaseNeighbourData(NEIGHBOUR_DATA **nd_ptr)
{
    NEIGHBOUR_DATA *nd;

    assert(nd_ptr != NULL);
    if ((nd = *nd_ptr) == NULL)
	return;

    phgFree(nd->data);
    phgFree(nd->index);
    phgFree(nd->plist);

    phgFree(nd);

    *nd_ptr = NULL;
}

FLOAT *
phgDofNeighbourData(NEIGHBOUR_DATA *nd, SIMPLEX *e, int face_no, int bas_no,
	INT *gindex)
/* returns the address of the DOF data for a given basis function on the face
 * 'face_no' of 'e' (note: bas_no is the basis number with respect to the face)
 *
 * If 'gindex != NULL' then it returns the global vector index of the DOF */
{
    DOF *dof = nd->dof;
    SIMPLEX *e1;
    int face_no1;

    assert(bas_no < nd->nbas);

#if USE_MPI
    if (e->bound_type[face_no] & REMOTE) {
	/* Remote neighbour */
	INT k = ((size_t)(e->neighbours[face_no])) * nd->nbas + bas_no;
	if (gindex != NULL) {
	    assert(nd->index != NULL);
	    *gindex = nd->index[k];
	}
	return nd->data + k * dof->dim;
    }
#endif

    e1 = e->neighbours[face_no];
    face_no1 = phgOppositeVertex(e, e->verts[face_no], e1);
    if (nd->peer != e1 || nd->face != face_no1) {
	if (nd->plist == NULL) {
	    nd->plist = phgAlloc(nd->nbas * sizeof(*nd->plist));
	}
	nd->peer = e1;
	nd->face = face_no1;
	/* compute and cache list of bases on the face */
	phgDofGetBasesOnFace(dof->type, e1, face_no1, nd->plist);
    }

    if (gindex != NULL) {
	assert(nd->map != NULL);
	*gindex = phgMapE2G(nd->map, nd->dof_no, e,
				nd->plist[bas_no] * dof->dim);
    }

    return dof->data + phgDofMapE2D(dof, e, nd->plist[bas_no] * dof->dim);
}
