/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2008 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: utils.c,v 1.184 2008/01/25 08:27:20 zlb Exp $ */

#define _GNU_SOURCE	/* for fcloseall() */

#include "phg.h"

#if USE_MPI
#include "hamilton.h"
#endif	/* USE_MPI */
#if USE_ZOLTAN
#include "partition-zoltan.h"
#endif	/* USE_ZOLTAN */

#include <stdlib.h>
#include <unistd.h>	/* usleep */
#include <limits.h>	/* PATH_MAX */
#include <stdarg.h>
#include <string.h>
#include <math.h>

#include <sys/time.h>
#ifdef USETIMES
# include <sys/times.h>
# include <unistd.h>
#else
# include <sys/resource.h>
#endif
#define TRUNK_SIZE 8

#if HAVE_FEENABLEEXCEPT
# include <fenv.h>
#endif	/* HAVE_FEENABLEEXCEPT */
#include <signal.h>

INT phgVerbosity = 0;
const char *phgCurrentFunctionName_ = "PHG";
INT _phg_i = 0;			/* scratch INT variable used in some macros */

/* function used by ForAllElements macro */
SIMPLEX *
_phg_step_element(GRID *g)
{
    SIMPLEX *e = NULL;

    while (_phg_i < g->nelem && (e = g->elems[_phg_i]) == NULL)
	_phg_i++;

    return e;
}

static GRID *g_traverse;	/* used by phgTraverseElements, other functions
				   shouldn't use it */
static GRID *g_;		/* for traversal callback functions */
BOOLEAN phgInitialized = FALSE;
BOOLEAN phgMasterSlave = FALSE;

char *phgLogFilename = NULL;
char *phgOutFilename = NULL;
static FILE *log_file = NULL;
static FILE *out_file = NULL;

#define LogFile	(log_file == NULL ? stderr : log_file)
#define OutFile	(out_file == NULL ? stdout : out_file)

static void
init_log_file(void)
{
    char s[PATH_MAX];

    if (phgLogFilename == NULL || log_file != NULL)
	return;

    if (strcmp(phgLogFilename, "stdout") == 0) {
	log_file = stdout;
	return;
    }
    else if (strcmp(phgLogFilename, "stderr") == 0) {
	log_file = stderr;
	return;
    }

    if (phgNProcs > 1)
	sprintf(s, "%s.%04d", phgLogFilename, phgRank);
    else
	strcpy(s, phgLogFilename);
    
    log_file = fopen(s, "w+t");
    if (log_file == NULL) {
	phgFree(phgLogFilename);
	phgLogFilename = NULL;
    }
}

static void
close_log_file(void)
{
    if (log_file != NULL && log_file != stdout && log_file != stderr) {
	fclose(log_file);
	if (phgRank > 0)
	    return;
	if (phgNProcs > 1)
	    phgPrintf("Log files \"%s.%04d\"--\"%s.%04d\" created.\n",
			phgLogFilename, 0, phgLogFilename, phgNProcs - 1);
	else
	    phgPrintf("Log file \"%s\" created.\n", phgLogFilename);
    }
    log_file = NULL;

    fflush(OutFile);
    if (out_file != NULL && out_file != stdout && out_file != stderr) {
	fclose(out_file);
    }
    out_file = NULL;
}

int
phgCompINT(const void *p0, const void *p1)
{
    INT i = *((INT *)p0) - *((INT *)p1);

    return (i < 0) ? -1 : ((i == 0) ? 0 : 1);
}

INT
phgBinarySearch(INT n, INT table[], INT key)
/* binary search, table[] is a sorted table of size n.
 *   1. if (n == 0 || key < table[0]) return -1;
 *   2. if (key > table[n - 1]) return n;
 *   3.	return i, s.t. (key >= table[i] && (i == n-1 || key < table[i + 1]); */
{
    INT i = 0, j = n - 1, k;

    if (n <= 0 || key < table[0])
	return -1;
    if (key >= table[j])
	return n - 1;

    while (j - i > 1) {
	k = (i + j) / 2;
	assert(k != i && k != j);
	if (table[k] <= key)
	    i = k;
	else
	    j = k;
    }

    return i;
}

#if HAVE_FEENABLEEXCEPT
static void
fpe_handler(int signum, siginfo_t *siginfo, void *dummy)
{
    switch (siginfo->si_code) {
	case FPE_FLTDIV:
	    phgError(1, "floating point exception: divide by zero.\n");
	    break;
	case FPE_FLTOVF:
	    phgError(1, "floating point exception: overflow.\n");
	    break;
	case FPE_FLTUND:
	    phgError(1, "floating point exception: underflow.\n");
	    break;
	case FPE_FLTRES:
	    phgError(1, "floating point exception: inexact result.\n");
	    break;
	case FPE_FLTINV:
	    phgError(1, "floating point exception: invalid operation.\n");
	    break;
	default:
	    phgError(1, "program received signal SIGFPE, abort.\n");
    }
}
#endif	/* HAVE_FEENABLEEXCEPT */

void
phgTrapSignals_(BOOLEAN init)
/* Note: should be called once with 'init == TRUE' once before phgInit,
 * and once with 'init == FALSE' to set FPE trap after initialization
 * of all external packages (like PETSc) */
{
#if HAVE_FEENABLEEXCEPT
    int flags = 0;
    static const char *fpetrap_keywords[] = {"none", "fatal", "all", NULL};
    static int fpetrap = 1;
#endif	/* HAVE_FEENABLEEXCEPT */
    static BOOLEAN reset_signals = TRUE;
    struct sigaction sa;

    if (init) {
	/* Note: the function is called once with 'init == TRUE' for
	 * registering options, then once with 'init == FALSE' after
	 * initialization of external solvers, to make sure the sigactions
	 * are called after external packages */
	phgOptionsRegisterNoArg("reset_signals", "Reset signal handlers",
		&reset_signals);
#if HAVE_FEENABLEEXCEPT
	phgOptionsRegisterKeyword("fpetrap",
		"Trap floating point exceptions",
		fpetrap_keywords, &fpetrap);
#endif	/* HAVE_FEENABLEEXCEPT */
	return;
    }

    if (reset_signals) {
	/* reset some signal handlers to default */
	sa.sa_handler = SIG_DFL;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	if (sigaction(SIGPIPE, &sa, NULL))
	    perror("sigaction");
	if (sigaction(SIGINT, &sa, NULL))
	    perror("sigaction");
	if (sigaction(SIGQUIT, &sa, NULL))
	    perror("sigaction");
	if (sigaction(SIGILL, &sa, NULL))
	    perror("sigaction");
	if (sigaction(SIGSEGV, &sa, NULL))
	    perror("sigaction");
	if (sigaction(SIGTERM, &sa, NULL))
	    perror("sigaction");
    }

#if HAVE_FEENABLEEXCEPT
    switch (fpetrap) {
	case 2:	/* all */
	    flags = FE_UNDERFLOW /*| FE_INEXACT*/;
	    /* fall through the next case */
	case 1:	/* fatal */
	    flags |= FE_DIVBYZERO | FE_INVALID | FE_OVERFLOW;
	    /* Clear current exceptions queue.
	     * Note: the initialization code of the Anasazi package from
	     * Trilinos-7.0.5 generates 'invalid operation' traps on i386 */
	    feclearexcept(flags);
	    if (feenableexcept(flags) == -1) {
		perror("");
		exit(1);
	    }
	    break;
	case 0:	/* none */
	    return;
    }

    /* install SIGFPE handler */
    sa.sa_sigaction = fpe_handler;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = SA_SIGINFO;
    if (sigaction(SIGFPE, &sa, NULL))
	perror("sigaction");
#endif	/* HAVE_FEENABLEEXCEPT */
}

BOOLEAN _phg_pause = FALSE;

void
phgInit(int *argc, char ***argv)
{
    if (phgInitialized) return;

    /* test validity of basic datatypes */
    assert(sizeof(INT) >= 4);
    assert(sizeof(UINT) >= 4);
    assert(sizeof(SHORT) >= 2);
    assert(sizeof(USHORT) >= 2);
    assert(sizeof(BYTE) == 1);

    /* Register partitioner options */
    phgRedistributeGrid(NULL);
    /* Register marking options */
    phgMarkElements_(0, NULL, 0., NULL, 0., 0, 0., 0, NULL, 0., 0, 0., 0.);
    /* Register solver options */
    phgSolverRegisterOptions();

#if USE_MPI
    phgInitMPI(argc, argv);
#else	/* USE_MPI */
    phgPrintf("Parallel Hierarchical Grid (version " PHG_VERSION ").\n");
    phgOptionsParseCmdline(argc, argv);

    /* Initialize modules after parsing all options.
     * Note: also add the calls below in phgInitMPI() in phg-mpi.c */
    phgPerfInit();			/* init phgPerf */
    phgSolverInitialize(argc, argv);	/* init phgSolver */
#if USE_ZOLTAN
    phgPartitionZoltanInitialize(*argc, *argv);
#endif	/* USE_ZOLTAN */

    phgOptionsHelp();

    if (phgVerbosity >= 2) {
	phgPrintf("sizeof(SIMPLEX)    = %d\n", sizeof(SIMPLEX));
	phgPrintf("sizeof(GRID)       = %d\n", sizeof(GRID));
    }
    fflush(OutFile);

    phgTrapSignals_(FALSE);
    phgTimerInit(FALSE);
#endif	/* USE_MPI */

    phgInitialized = TRUE;
    if (_phg_pause)
	phgPause(0);
    if (phgMasterSlave) atexit(phgFinalize);
}

#if USE_MPI
void
phgSetVerbosity_(GRID *g)
{
    ParallelFunction(phgSetVerbosity_, TRUE);

    if (phgNProcs <= 0) return;

    MPI_Bcast(&phgVerbosity, 1, MPI_INT, 0, MPI_COMM_WORLD);

    return;
}
#endif

int
phgSetVerbosity(int verbosity)
{
    int old = phgVerbosity;
    if (verbosity >= 0) phgVerbosity = verbosity;
#if USE_MPI
    phgSetVerbosity_(NULL);
#endif
    return old;
}

void
phgFinalize(void)
{
    if (!phgInitialized)
	return;

#if USE_PAPI
    phgPerfFinalize();
#endif
    phgTimerReset();

    phgSolverFinalize();
    phgQuadReset();

    close_log_file();

    phgOptionsReset();

    phgInitialized = FALSE;

#if USE_MPI
    if (!phgMasterSlave)
	phgCreateComm_(-1, NULL);	/* tell idle processes to exit */
    phgFinalizeMPI();
#endif
 
    phgFreeAtExit(NULL);

    if (phgRank > 0)
	exit(0);
}

void
phgPause(int seconds)
{
    char s[128];

    if (seconds <= 0) {
	int i;
	pid_t pid = getpid(), *pids = phgAlloc(phgNProcs * sizeof(*pids));
#if USE_MPI
	MPI_Gather(&pid, sizeof(pid), MPI_BYTE, pids, sizeof(pid), MPI_BYTE,
		   0, MPI_COMM_WORLD);
#else	/* USE_MPI */
	pids[0] = pid;
#endif	/* USE_MPI */
	if (phgRank == 0) {
            fprintf(stderr, "Program paused.\n");
	    for (i = 0; i < phgNProcs; i++) {
#if DEBUG && USE_MPI
		fprintf(stderr, "  proc %d (%s), pid = %d\n", i, 
			phgProcessorNames[i], (int)pids[i]);
#else	/* DEBUG && USE_MPI */
		fprintf(stderr, "  proc %d (%s), pid = %d\n", i,
			"unknown", (int)pids[i]);
#endif	/* DEBUG && USE_MPI */
	    }
	}
	phgFree(pids);
    }
    if (phgRank == 0) {
	if (seconds <= 0) {
            fflush(stderr);
            fprintf(stderr, "<<< Press enter to continue >>>\n");
            fflush(stderr);
            fgets(s, sizeof(s), stdin);
            fflush(stderr);
	}
	else {
	    sleep(seconds);
	}
    }
#if USE_MPI
    MPI_Barrier(MPI_COMM_WORLD);
#endif
}

static void
init_elements(SIMPLEX *s, int count)
{
    int i, j;

    for (i = 0; i < count; i++) {
	for (j = 0; j < NEdge; j++) {
	    s[i].edges[j] = -1;
#if ALLOW_CURVED_BOUNDARY
	    s[i].bound_func[j] = -1;
#endif
	}
	for (j = 0; j < NFace; j++) {
	    s[i].bound_type[j] = UNDEFINED;
#if (Dim == 3)
	    s[i].faces[j] = -1;
#endif
	}
	s[i].index = -1;
    }
}

SIMPLEX *
phgNewElements(int count)
{
    SIMPLEX *s = phgCalloc(count, sizeof(SIMPLEX));

    init_elements(s, count);
    return s;
}

SIMPLEX *
phgReallocElements(SIMPLEX *s, int oldcount, int newcount)
{
    if (newcount == oldcount)
	return s;
    s = phgRealloc(s, newcount * sizeof(SIMPLEX));
    if (newcount <= oldcount)
	return s;
    memset(s + oldcount, 0, (newcount - oldcount) * sizeof(*s));
    init_elements(s + oldcount, newcount - oldcount);

    return s;
}

COORD *
phgNewVertices(int count)
{
    COORD *v;

    v = phgAlloc(count * sizeof(COORD));
    return v;
}

static int flags_tmp;
static size_t g_list_count = 0;
static GRID **g_list = NULL;	/* list of valid grids */

void
phgNewGrid_(GRID *g)
{
    ParallelFunction(phgNewGrid_, TRUE);

#if USE_MPI
    if (phgMasterSlave && phgNProcs > 0) {
	MPI_Bcast(&flags_tmp, sizeof(flags_tmp), MPI_BYTE, 0, MPI_COMM_WORLD);
    }
#endif

    /* for Master/Slave mode in which only one grid is allowed */
    g_sys = g;
    if (g != NULL) return;

    g_sys = g = phgCalloc(1, sizeof(GRID));
    g->lif = 1.0;
    g->nelem = g->nleaf = g->nroot = g->nvert = g->nedge = g->ntree = 0;
#if Dim == 3
    g->nface = 0;
    g->nface_global = 0;
#endif
    g->verts = NULL;
    g->roots = NULL;
    g->elems = NULL;
    g->dof   = NULL;
    g->vert_flag = /*flags_tmp &*/ VERT_FLAG;	/* always needed */
    g->edge_flag = flags_tmp & EDGE_FLAG;
#if USE_ZOLTAN
    g->face_flag = FACE_FLAG;			/* needed by Zoltan */
#else	/* USE_ZOLTAN */
    g->face_flag = flags_tmp & FACE_FLAG;
#endif	/* USE_ZOLTAN */
    g->elem_flag = /*flags_tmp &*/ ELEM_FLAG;	/* always needed */
    g->geom_flag = flags_tmp & GEOM_FLAG;
    g->periodic = FALSE;
#if ALLOW_CURVED_BOUNDARY
    g->bdry_funcs = NULL;
#endif	/* ALLOW_CURVED_BOUNDARY */
    g->types_vert = NULL;
    g->types_edge = NULL;
    g->types_face = NULL;
    g->types_elem = NULL;
    g->alien = NULL;
    g->alien_map = NULL;
    g->perlist = NULL;
    g->nelem_global = g->nvert_global = g->nedge_global = 0;
#if USE_MPI
    g->L2Gmap_vert = NULL;
    g->L2Gmap_edge = NULL;
    g->L2Gmap_face = NULL;
    g->L2Gmap_elem = NULL;
#if NO_NEW_COMMUNICATOR
    g->comm = MPI_COMM_SELF;
#else	/* NO_NEW_COMMUNICATOR */
    MPI_Comm_dup(MPI_COMM_SELF, &g->comm);
#endif	/* NO_NEW_COMMUNICATOR */

    g->neighbours.counts = g->neighbours.displs = NULL;
    g->neighbours.list = NULL;
    g->neighbours.count = g->neighbours.allocated = 0;
#endif	/* USE_MPI */
    g->rank = 0;
    g->nprocs = 1;
    g->serial_no = 0;

    g->last_partitioner = -1;

    /* add g to g_list (for phgGridIsValid) */
    g_list = phgRealloc(g_list, (g_list_count + 1) * sizeof(*g_list));
    g_list[g_list_count++] = g;

    return;
}

GRID *
phgNewGrid(int flags)
{
    flags_tmp = flags;
    phgNewGrid_(NULL);

    return g_sys;
}

GRID *
phgDummyGrid(void)
{
    GRID *g = phgNewGrid(0);
#if USE_MPI
#if NO_NEW_COMMUNICATOR
    g->comm = MPI_COMM_WORLD;
#else	/* NO_NEW_COMMUNICATOR */
    MPI_Comm_free(&g->comm);
    MPI_Comm_dup(MPI_COMM_WORLD, &g->comm);
#endif	/* NO_NEW_COMMUNICATOR */
    g->nprocs = phgNProcs;
    g->rank = phgRank;
#endif	/* USE_MPI */

    return g;
}

BOOLEAN
phgGridIsValid(const GRID *g)
/* checks whether the given grid is valid */
{
    int i;

    for (i = 0; i < g_list_count; i++)
	if (g_list[i] == g) return TRUE;

    return FALSE;
}

/* Array mapping edge no to vertices, used by GetEdgeVertex macro */
BYTE GetEdgeVertex_[NEdge][2] = {{0,1}, {0,2}, {0,3}, {1,2}, {1,3}, {2,3}};
/* Array mapping face no to vertices, used by GetFaceVertex macro */
BYTE GetFaceVertex_[NFace][3] = {{1,2,3}, {0,2,3}, {0,1,3}, {0,1,2}};
/* array mapping pairs of vertices to edges */
BYTE GetEdgeNo_[NVert][NVert] = {{-1, 0, 1, 2},
				 { 0,-1, 3, 4},
				 { 1, 3,-1, 5},
				 { 2, 4, 5,-1}};

static INT index_ = 0;

static BOOLEAN
elems_callback CB_ARGS(e)
{
    int v0, v1, v2, v3;
    INT V0, V1, V2, V3;

    if (IsLeaf(e)) {
	g_->elems[e->index] = e;
    }
    else {
	/* point e->parent to parent */
	if (e->children[0] != NULL)
	    e->children[0]->parent = e;
	if (e->children[1] != NULL)
	    e->children[1]->parent = e;
    }
    if (e->generation == 0)
	e->parent = NULL;

    if (e->ordering != 0)
	return TRUE;

    /* compute ordering of vertices with increasing global indices */
    V0 = GlobalVertex(g_, e->verts[0]);
    V1 = GlobalVertex(g_, e->verts[1]);
    V2 = GlobalVertex(g_, e->verts[2]);
    V3 = GlobalVertex(g_, e->verts[3]);
    v0 = v1 = v2 = v3 = 0;
    (V0 > V1) ? v0++ : v1++;
    (V0 > V2) ? v0++ : v2++;
    (V0 > V3) ? v0++ : v3++;
    (V1 > V2) ? v1++ : v2++;
    (V1 > V3) ? v1++ : v3++;
    (V2 > V3) ? v2++ : v3++;
    e->ordering = v0 | (v1 << 2) | (v2 << 4) | (v3 << 6);

    return TRUE;
}

static BOOLEAN
types_callback CB_ARGS(e)
{
    int i, type;

    if (g_->elem_flag)
	g_->types_elem[e->index] = OWNER;

    for (i = 0; i < NFace; i++) {
	type = e->bound_type[i];
#if 0
	/* use DIRICHLET for UNDEFINED boundary type */
	if (type == UNDEFINED) type = DIRICHLET;
#endif
#if USE_MPI
	if (g_->nprocs == 1)
#endif
	type |= OWNER;

	if (g_->vert_flag) {
	    switch (i) {
		case 0:
		    g_->types_vert[e->verts[1]] |= type;
		    g_->types_vert[e->verts[2]] |= type;
		    g_->types_vert[e->verts[3]] |= type;
		    break;
		case 1:
		    g_->types_vert[e->verts[0]] |= type;
		    g_->types_vert[e->verts[2]] |= type;
		    g_->types_vert[e->verts[3]] |= type;
		    break;
		case 2:
		    g_->types_vert[e->verts[0]] |= type;
		    g_->types_vert[e->verts[1]] |= type;
		    g_->types_vert[e->verts[3]] |= type;
		    break;
		case 3:
		    g_->types_vert[e->verts[0]] |= type;
		    g_->types_vert[e->verts[1]] |= type;
		    g_->types_vert[e->verts[2]] |= type;
		    break;
	    }
	}

	if (g_->edge_flag) {
	    switch (i) {
		case 0:
		    g_->types_edge[e->edges[3]] |= type;
		    g_->types_edge[e->edges[4]] |= type;
		    g_->types_edge[e->edges[5]] |= type;
		    break;
		case 1:
		    g_->types_edge[e->edges[1]] |= type;
		    g_->types_edge[e->edges[2]] |= type;
		    g_->types_edge[e->edges[5]] |= type;
		    break;
		case 2:
		    g_->types_edge[e->edges[0]] |= type;
		    g_->types_edge[e->edges[2]] |= type;
		    g_->types_edge[e->edges[4]] |= type;
		    break;
		case 3:
		    g_->types_edge[e->edges[0]] |= type;
		    g_->types_edge[e->edges[1]] |= type;
		    g_->types_edge[e->edges[3]] |= type;
		    break;
	    }
	}

	if (g_->face_flag) {
	    g_->types_face[e->faces[i]] |= type;
	}

	if (g_->elem_flag) {
	    g_->types_elem[e->index] |= (type & REMOTE);
	}
    }

    return TRUE;
}

#if USE_MPI
typedef struct {
    INT  index;		/* global index */ 
    GTYPE gtype;	/* vertex or edge */
    BYTE glob_flag;	/* whether index is global */
    BYTE type;		/* boundary type flags */
} B_LIST;

static int
blist_comp(const void *p0, const void *p1)
{
    INT i, j, k, l;
    B_LIST *b0, *b1;

    k = (b0 = (void *)p0)->gtype;
    l = (b1 = (void *)p1)->gtype;
    if (k != l) return k - l;

    if (b0->glob_flag)
	i = b0->index;
    else
	i = k ? GlobalEdge(g_, b0->index) : GlobalVertex(g_, b0->index);
    if (b1->glob_flag)
	j = b1->index;
    else
	j = l ? GlobalEdge(g_, b1->index) : GlobalVertex(g_, b1->index);

    return i - j;
}
#endif

static void set_default_bdry_type(GRID *g);

void
phgUpdateBoundaryTypes(GRID *g)
/* Sets boundary types (g->types_xxxx), this function also updates g->elems */
{
    SIMPLEX *e;
#if USE_MPI
    int n, dst, src, sign = 0, *counts, rank0, rank1;
    INT i, j, k, l, N, M, nmax;
    MPI_Datatype type;
    MPI_Status status;
    B_LIST *blist, *bindex, *bl0, *bl1;
    MPI_Comm comm;
#else
    INT i;
#endif

    assert(g->elem_flag != FALSE); /* e->index is required by elems_callback */

    phgFree(g->elems);
    g->elems = NULL;

    if (g->nprocs <= phgRank)
	return;

    set_default_bdry_type(g);

    g_ = g;
    g->elems = phgCalloc(g->nelem, sizeof(*g->elems));
    index_ = 0;
    /* construct g->elems, update e->parent and e->odering */
    phgTraverseAllElements(g, elems_callback);

    if (!g->vert_flag && !g->edge_flag && !g->face_flag && !g->elem_flag)
	return;

    if (g->vert_flag) {
	phgFree(g->types_vert);
	g->types_vert = phgAlloc(g->nvert * sizeof(*g->types_vert));
	memset(g->types_vert, UNREFERENCED, g->nvert * sizeof(*g->types_vert));
    }

    if (g->edge_flag) {
	phgFree(g->types_edge);
	g->types_edge = phgAlloc(g->nedge * sizeof(*(g->types_edge)));
	memset(g->types_edge, UNREFERENCED, g->nedge);
    }

    if (g->face_flag) {
	phgFree(g->types_face);
	g->types_face = phgAlloc(g->nface * sizeof(*(g->types_face)));
	memset(g->types_face, UNREFERENCED, g->nface);
    }

    if (g->elem_flag) {
	phgFree(g->types_elem);
	g->types_elem = phgAlloc(g->nelem * sizeof(*(g->types_elem)));
	memset(g->types_elem, UNREFERENCED, g->nelem);
    }

    ForAllElements(g, e)
	types_callback(e);

#if USE_MPI
    if (g->nprocs == 1) return;

#if NO_NEW_COMMUNICATOR
    comm = g->comm;
#else	/* NO_NEW_COMMUNICATOR */
    MPI_Comm_dup(g->comm, &comm);
#endif	/* NO_NEW_COMMUNICATOR */

    /* FIXME: optimization: use Alltoallv instead of the circular algorithm */

    /* count number of vertices/edges which have REMOTE flag set */
    n = 0;
    if (g->vert_flag)
	for (i = 0; i < g->nvert; i++) {
	    if (g->types_vert[i] & REMOTE) n++;
	}
    if (g->edge_flag)
	for (i = 0; i < g->nedge; i++) {
	    if (g->types_edge[i] & REMOTE) n++;
	}
    counts = phgAlloc(g->nprocs * sizeof(*counts));
    MPI_Allgather(&n, 1, MPI_INT, counts, 1, MPI_INT, comm);
    nmax = 0;
    for (i = 0; i < g->nprocs; i++) {
	if (nmax < counts[i]) nmax = counts[i];
    }
    if (nmax == 0) {
	if (g->vert_flag) {
	    for (i = 0; i < g->nvert; i++)
		g->types_vert[i] |= OWNER;
	}
	if (g->edge_flag) {
	    for (i = 0; i < g->nedge; i++)
		g->types_edge[i] |= OWNER;
	}
	phgFree(counts);
	goto skip_comm;
    }
    /* collect vertices/edges which have REMOTE flag set */
    bl0 = blist = phgAlloc(2 * nmax * sizeof(*blist));
    if (g->vert_flag) {
	for (i = 0; i < g->nvert; i++) {
	    if (g->types_vert[i] == UNREFERENCED)
		continue;
	    g->types_vert[i] |= OWNER;
	    if (!(g->types_vert[i] & REMOTE))
		continue;
	    bl0->type		= (g->types_vert[i] &
					~(UNDEFINED | INTERIOR | OWNER));
	    bl0->index		= i;
	    bl0->glob_flag	= 0;
	    (bl0++)->gtype	= VERTEX;
	}
    }
    if (g->edge_flag) {
	for (i = 0; i < g->nedge; i++) {
	    if (g->types_edge[i] == UNREFERENCED)
		continue;
	    g->types_edge[i] |= OWNER;
	    if (!(g->types_edge[i] & REMOTE))
		continue;
	    bl0->index		= i;
	    bl0->type		= (g->types_edge[i] &
					~(UNDEFINED | INTERIOR | OWNER));
	    bl0->glob_flag	= 0;
	    (bl0++)->gtype	= EDGE;
	}
    }
    /* sort the list */
    if (n > 0) qsort(blist, n, sizeof(*blist), blist_comp);
    bindex = phgAlloc(n * sizeof(*bindex));
    for (i = 0, bl0 = blist, bl1 = bindex; i < n; i++, bl0++, bl1++) {
	*bl1 = *bl0;	/* save local index */
	j = bl0->index;
	bl0->index = (bl0->gtype==EDGE ? GlobalEdge(g,j) : GlobalVertex(g,j));
	bl0->glob_flag = 1;
    }
    /* circulate blist */
    MPI_Type_contiguous(sizeof(*blist), MPI_BYTE, &type);
    MPI_Type_commit(&type);
    j = g->rank;
    dst = (j == 0) ? g->nprocs - 1 : j - 1;
    src = (j == g->nprocs - 1) ? 0 : j + 1;
    for (i = 0; i < g->nprocs - 1; i++) {
	M = counts[j];
	N = counts[(j == g->nprocs - 1) ? 0 : j + 1];
	if ((i & 1) == 0) {
	    bl0 = blist + nmax;
	    bl1 = blist;
	}
	else {
	    bl0 = blist;
	    bl1 = blist + nmax;
	}
	/* circulate blist */
	MPI_Sendrecv(bl1, M, type, dst, 111,
		     bl0, N, type, src, 111, comm, &status);
#if DEBUG
#if SENDRECV_TEST
	/* Verify result of MPI_Sendrecv by sending back and comparing data */
	{
	    INT i;
	    B_LIST *back;
	    back = phgAlloc(M * sizeof(*back));
	    MPI_Sendrecv(bl0, N, type, src, 222, back, M, type, dst, 222,
			  comm, &status);
	    for (i = 0; i < M; i++) {
		if (phgMemcmp(back + i, bl1 + i, sizeof(*back))) {
		    phgInfo(-1, "%s: sent to %s: %d %d %d %d, got %d %d %d %d\n",
				phgProcessorNames[g->rank],
				phgProcessorNames[dst],
				bl1[i].index, bl1[i].gtype,
				bl1[i].glob_flag, GTypeName(bl1[i].type),
				back[i].index, back[i].gtype,
				back[i].glob_flag, GTypeName(back[i].type));
		    phgError(1, "%s:%d, testing MPI_Sendrecv failed!\n",
				__FILE__, __LINE__);
		}
	    }
	    phgFree(back);
	}
#endif	/* SENDRECV_TEST */
#endif	/* DEBUG */
	if (++j >= g->nprocs) j = 0;
	/* update local types arrays */
	k = l = 0;
	bl1 = bindex;
	while (k < N) {
	    while (l < n && (sign = blist_comp(bl1, bl0)) < 0) {
		l++;
		bl1++;
	    }
	    if (l >= n) break;
	    if (sign == 0) {
		rank0 = (g->rank + (bl0->index % g->nprocs)) % g->nprocs;
		rank1 = (j       + (bl0->index % g->nprocs)) % g->nprocs;
		if (bl1->gtype == EDGE) {
		    g->types_edge[bl1->index] |= bl0->type;
		    if (rank0 > rank1) {
			/* remote process owns the edge */
			g->types_edge[bl1->index] &= ~OWNER;
		    }
		}
		else {
		    g->types_vert[bl1->index] |= bl0->type;
		    if (rank0 > rank1) {
			/* remote process owns the vertex */
			g->types_vert[bl1->index] &= ~OWNER;
		    }
		}
		if (++l >= n) break;
		bl1++;
	    }
	    k++;
	    bl0++;
	    while (k < N && blist_comp(bl0, bl1) < 0) {
		k++;
		bl0++;
	    }
	}
    }
    MPI_Type_free(&type);
    phgFree(bindex);
    phgFree(blist);
    phgFree(counts);

skip_comm:
    /* set OWNER flag for faces */
    if (g->face_flag) {
	for (i = 0; i < g->nface; i++) {
	    if (g->types_face[i] == UNREFERENCED)
		continue;
	    g->types_face[i] |= OWNER;
	}
	for (i = 0; i < g->neighbours.count; i++) {
	    RNEIGHBOUR *rn = g->neighbours.list + i;
	    k = GlobalFace(g, j = rn->local->faces[rn->vertex]);
	    assert(g->types_face[j] != UNREFERENCED);
	    rank0 = (g->rank  + (k % g->nprocs)) % g->nprocs;
	    rank1 = (rn->rank + (k % g->nprocs)) % g->nprocs;
	    g->types_face[j] |= REMOTE;
	    if (rank0 > rank1) {
		g->types_face[j] &= ~OWNER;
	    }
	}
    }

    /* set OWNER flag for elements */
    if (g->elem_flag) {
	ForAllElements(g, e)
	    g->types_elem[e->index] |= OWNER;
    }

#if DEBUG
    if (g->edge_flag && g->nedge_global > 0) {
	n = 0;
	for (i = 0; i < g->nedge; i++) {
	    if (g->types_edge[i] & OWNER)
		n++;
	}
	rank0 = g->nedge_global;
	MPI_Allreduce(&n, &rank1, 1, MPI_INT, MPI_SUM, g->comm); 
	if (rank0 != rank1) {
	    phgInfo(-1, "total edges: %d, edges owned: %d\n", rank0, n);
#if 0
	    phgDumpGrid(g);
	    for (i = 0; i < g->nedge; i++) {
		phgInfo(-1, "types_edge[%d] = %d (owner = %d)\n",
			GlobalEdge(g, i), g->types_edge[i],
			(g->types_edge[i] & OWNER) != 0);
	    }
#endif
	    phgError(1, "%s:%d, wrong # of edges owned: %d\n",
			__FILE__, __LINE__, rank1);
	}
    }

    if (g->face_flag && g->nface_global > 0) {
	n = 0;
	for (i = 0; i < g->nface; i++) {
	    if (g->types_face[i] & OWNER)
		n++;
	}
	rank0 = g->nface_global;
	MPI_Allreduce(&n, &rank1, 1, MPI_INT, MPI_SUM, g->comm); 
	if (rank0 != rank1) {
	    phgInfo(-1, "total faces: %d, faces owned: %d\n", rank0, n);
#if 0
	    phgDumpGrid(g);
	    for (i = 0; i < g->nface; i++) {
		phgInfo(-1, "types_face[%d] = %d (owner = %d)\n",
			GlobalFace(g, i), g->types_face[i],
			(g->types_face[i] & OWNER) != 0);
	    }
#endif
	    phgError(1, "%s:%d, wrong # of faces owned: %d\n",
			__FILE__, __LINE__, rank1);
	}
    }
#endif	/* DEBUG */

#if !NO_NEW_COMMUNICATOR
    MPI_Comm_free(&comm);
#endif	/* !NO_NEW_COMMUNICATOR */

#else	/* USE_MPI */

    if (g->vert_flag)
	for (i = 0; i < g->nvert; i++)
	    g->types_vert[i] |= OWNER;
    if (g->edge_flag)
	for (i = 0; i < g->nedge; i++)
	    g->types_edge[i] |= OWNER;
    if (g->face_flag)
	for (i = 0; i < g->nface; i++)
	    g->types_face[i] |= OWNER;

    if (g->elem_flag)
	for (i = 0; i < g->nelem; i++)
	    g->types_elem[i] |= OWNER;
#endif	/* USE_MPI */
}

void *
phgAlloc(size_t size)
{
    void *ptr = (size != 0) ? malloc(size) : NULL;
    
    if (ptr == NULL && size != 0)
	phgError(1, "failed to malloc %u bytes memory.\n", size);

    return ptr;
}

void *
phgCalloc(size_t nmemb, size_t size)
{
    void *ptr = (nmemb != 0 && size != 0) ? calloc(nmemb, size) : NULL;

    if (ptr == NULL && nmemb != 0 && size != 0)
	phgError(1, "failed to calloc %u bytes memory.\n", size * nmemb);

    return ptr;
}

void *
phgRealloc(void *ptr, size_t size)
{
    void *p = realloc(ptr, size);
    
    if (p == NULL && size != 0)
	phgError(1, "failed to reallocate %u bytes memory.\n", size);

    return p;
}

void
phgFree(void *ptr)
{
    if (ptr != NULL) free(ptr);
}

int
phgMemcmp(const void *v0, const void *v1, size_t n)
/* replaces buggy memcmp() on many systems */
{
    const BYTE *s0 = v0, *s1 = v1;
    int i = 0;

    for (; n--; )
	if ((i = *(s0++) - *(s1++)) != 0)
	    break;

    return i;
}

void
phgFreeAtExit(void **ptr)
{
    static void ***ptr_table = NULL;
    static size_t table_n = 0, allocated = 0;

    if (ptr == NULL) {	/* free all registered pointers */
	int i;
	for (i = table_n - 1; i >= 0; i--)
	    phgFree(*(ptr_table[i]));
	phgFree(ptr_table);
	ptr_table = NULL;
	table_n = allocated = 0;

	return;
    }

    if (table_n >= allocated)
	ptr_table = (void***)phgRealloc(ptr_table,
				sizeof(*ptr_table) * (allocated += 128));
    ptr_table[table_n++] = ptr; 

    return;
}

double
phgGetTime(double tarray[])
/* tarray[0]=User time, tarray[1]=Sys time, tarray[2]=Elapsed time,
   all expressed in seconds.

   DAWN1000 nf77
	double F77name(dclock)();
	tarray[0]=tarray[1]=tarray[2]=F77name(dclock)(); */
{
    struct timeval tv;
  
    if (tarray != NULL) {
#ifdef USETIMES  /*==================================================*/
	struct tms ts;
	static double ClockTick = 0.0;

	if (ClockTick == 0.0)
	    ClockTick = (double)sysconf(_SC_CLK_TCK);
	if ( times(&ts) == -1 ) {
	    tarray[0]=tarray[1] = 0.0;
	} else {
	    tarray[0] = (double)ts.tms_utime / ClockTick;
	    tarray[1] = 0.0;
	}
#else		 /*==================================================*/
	/* For Solaris */
	/*#include </usr/ucbinclude/sys/rusage.h> */
	struct rusage RU;
	getrusage(RUSAGE_SELF, &RU);
	tarray[0] = RU.ru_utime.tv_sec + (double)RU.ru_utime.tv_usec * 1e-6;
	tarray[1] = RU.ru_stime.tv_sec + (double)RU.ru_stime.tv_usec * 1e-6;
#endif		 /*==================================================*/
    }

    gettimeofday(&tv, (struct timezone *)0);
    if (tarray != NULL)
	return tarray[2] = tv.tv_sec + (double)tv.tv_usec * 1e-6;
    else
	return tv.tv_sec + (double)tv.tv_usec * 1e-6;
}

void
phgWarning(const char *fmt, ...)
{
    va_list ap;
    char s[1024];
    const char *n;

    init_log_file();

    n = phgCurrentFunctionName_;
    if (strlen(n) >= 3 && !phgMemcmp(n, "phg", 3)) n += 3;

    if (phgNProcs > 1)
	sprintf(s, "*%d %s: *** WARNING: %s", phgRank, n, fmt);
    else
	sprintf(s, "* %s: *** WARNING: %s", n, fmt);
    va_start(ap, fmt);
    vfprintf(LogFile, s, ap);
    va_end(ap);
    fflush(LogFile);
}

void
phgAbort(int code)
{
    close_log_file();

    /* leave time for processes to clean up */
    if (phgNProcs > 1)
	usleep(500000);

#ifdef HAVE_FCLOSEALL
    fcloseall();
#endif

#if USE_MPI
    if (phgInitialized && phgNProcs > 1) {
	MPI_Abort(MPI_COMM_WORLD, code);
    }
#endif

    exit(code);
}

void 
phgError(int code, const char *fmt, ...)
{
    va_list ap;
#if USE_MPI
    char s[4096], pn[MPI_MAX_PROCESSOR_NAME + 4];
#else
    char s[4096], pn[1];
#endif
    const char *n;

    init_log_file();

    n = phgCurrentFunctionName_;
    if (strlen(n) >= 3 && !phgMemcmp(n, "phg", 3)) n += 3;

#if DEBUG && USE_MPI
    if (phgInitialized)
	sprintf(pn, " (%s)", phgProcessorNames[phgRank]);
    else
#endif
    pn[0] = '\0';
    if (phgNProcs > 1)
	sprintf(s, "\n*%d%s %s: *** ERROR: %s\n", phgRank, pn, n, fmt);
    else
	sprintf(s, "\n*%s %s: *** ERROR: %s\n", pn, n, fmt);
    va_start(ap, fmt);
    vfprintf(LogFile, s, ap);
    va_end(ap);
    fflush(LogFile);

    if (code == 0) return;
    phgAbort(code);
}

void
phgInfo(int verbose_level, const char *fmt, ...)
/* writes formatted message to log_file when log_file != NULL,
 * if verbose_level < 0 then writes to stdout if log_file == NULL */
{
    va_list ap;
    char s[1024];
    const char *n;

    init_log_file();

#if USE_MPI
    /*if (phgRank != 0) return;*/
#endif
    if ((verbose_level >= 0 && log_file == NULL) ||
	verbose_level > phgVerbosity) return;

    n = phgCurrentFunctionName_;
    if (strlen(n) >= 3 && !phgMemcmp(n, "phg", 3)) n += 3;

    if (phgNProcs > 1)
	sprintf(s, "*%d %s: %s", phgRank, n, fmt);
    else
	sprintf(s, "* %s: %s", n, fmt);
    va_start(ap, fmt);
    vfprintf(LogFile, s, ap);
    va_end(ap);
    fflush(LogFile);
}

#undef Log

int
phgPrintf(const char *fmt, ...)
{
    va_list ap;
    int ret = 0;

    if (phgRank == 0) {
	if (out_file == NULL && phgOutFilename != NULL) {
	    if (strcmp(phgOutFilename, "stdout") == 0) {
		out_file = stdout;
	    }
	    else if (strcmp(phgOutFilename, "stderr") == 0) {
		out_file = stderr;
	    }
	    else {
		out_file = fopen(phgOutFilename, "w+t");
		if (out_file == NULL) {
		    phgWarning("cannot open output file \"%s\".",
							phgOutFilename);
		    phgFree(phgOutFilename);
		    phgOutFilename = NULL;
		}
	    }
	}
	va_start(ap, fmt);
	ret = vfprintf(OutFile, fmt, ap);
	va_end(ap);
	fflush(OutFile);
    }

#if USE_MPI
    /*if (g->nprocs > 0)
	MPI_Bcast(&ret, sizeof(ret), MPI_BYTE, 0, g->comm);*/
#endif

    return ret;
}

FILE *
phgFOpen(GRID *g, const char *path, const char *mode)
{
#if USE_MPI
    FILE *fp;

    if (g->rank == 0)
	fp = fopen(path, mode);
# if 1
    else
	fp = (void *)-1;
# else
    if (g->nprocs > 1)
	MPI_Bcast(&fp, sizeof(fp), MPI_BYTE, 0, g->comm);
# endif
    return fp;
#else
    Unused(g);
    return fopen(path, mode);
#endif
}

int
phgFPrintf(GRID *g, FILE *fp, char *fmt, ...)
{
    va_list ap;
    int ret;

#if USE_MPI
    if (g->rank == 0) {
	va_start(ap, fmt);
	ret = vfprintf(fp, fmt, ap);
	va_end(ap);
	fflush(fp);
    }
# if 1
    else
	ret = 1;
# else
    if (g->nprocs > 1)
	MPI_Bcast(&ret, sizeof(ret), MPI_BYTE, 0, g->comm);
# endif
#else
    Unused(g);
    va_start(ap, fmt);
    ret = vfprintf(fp, fmt, ap);
    va_end(ap);
    fflush(fp);
#endif

    return ret;
}

int
phgFClose(GRID *g, FILE *fp)
{
    int ret;

#if USE_MPI
    if (g->rank == 0)
	ret = fclose(fp);
# if 1
    else
	ret = 0;
# else
    if (g->nprocs > 1)
	MPI_Bcast(&ret, sizeof(ret), MPI_BYTE, 0, g->comm);
# endif
#else
    Unused(g);
    ret = fclose(fp);
#endif

    return ret;
}

BYTE
phgOppositeVertex(SIMPLEX *e, INT v, SIMPLEX *e_op)
/* returns the vertex of e_op opposite to the vertex v (local index) of e.
   This function does not rely on the neighbours links */
{
    INT i, u0 = 0, u1 = 0, u2 = 0;

#if DEBUG
    if (e_op == NULL)
	phgError(1, "unexpected error in %s (possibly caused by inconsistent "
		 "boundary types in the input mesh file).\n", __func__);
#endif	/* DEBUG */

    switch (v) {
	case 0: u0 = e->verts[1]; u1 = e->verts[2]; u2 = e->verts[3]; break;
	case 1: u0 = e->verts[0]; u1 = e->verts[2]; u2 = e->verts[3]; break;
	case 2: u0 = e->verts[0]; u1 = e->verts[1]; u2 = e->verts[3]; break;
	case 3: u0 = e->verts[0]; u1 = e->verts[1]; u2 = e->verts[2]; break;
    }

    if ((i = e_op->verts[0]) != u0 && i != u1 && i != u2)
	return 0;
    if ((i = e_op->verts[1]) != u0 && i != u1 && i != u2)
	return 1;
    if ((i = e_op->verts[2]) != u0 && i != u1 && i != u2)
	return 2;
    if ((i = e_op->verts[3]) != u0 && i != u1 && i != u2)
	return 3;
    return 4;
}

void
phgFreeElement(SIMPLEX **eptr)
{
    SIMPLEX *e;

    if (eptr == NULL || (e = *eptr) == NULL)
	return;
#if 0
    if (e->userdata != NULL)
	phgFree(e->userdata);
#endif
    phgFree(e);
    *eptr = NULL;
}

static void
free_grid_traverse(SIMPLEX *s)
{
    if (s->children[0] != NULL)
	free_grid_traverse(s->children[0]);
    if (s->children[1] != NULL)
	free_grid_traverse(s->children[1]);
    phgFreeElement(&s);
}

void
phgFreeGrid_(GRID *g)
{
    INT i, n;

    ParallelFunction(phgFreeGrid_, g != NULL && g->nprocs > 1);

    if (g == NULL)
	return;

    for (i = 0; i < g->nroot; i++) {
	SIMPLEX *s = g->roots + i;
	if (s->children[0] != NULL)
	    free_grid_traverse(s->children[0]);
	if (s->children[1] != NULL)
	    free_grid_traverse(s->children[1]);
#if 0
	if (s->userdata != NULL)
	    phgFree(s->userdata);
#endif
    }
    phgFree(g->roots);
    phgFree(g->verts);
    phgFree(g->elems);
#if ALLOW_CURVED_BOUNDARY
    if (g->bdry_funcs != NULL) {
	EXPR **p = g->bdry_funcs;
	while (*p != NULL)
	    phgFree3DFunction(*(p++));
	phgFree(g->bdry_funcs);
    }
#endif
    phgFree(g->types_vert);
    phgFree(g->types_edge);
    phgFree(g->types_face);
    phgFree(g->types_elem);
    phgFree(g->alien_map);
    phgFree(g->perlist);
#if USE_MPI
# if !NO_NEW_COMMUNICATOR
    if (g->comm != MPI_COMM_NULL)
	MPI_Comm_free(&g->comm);
# endif	/* !NO_NEW_COMMUNICATOR */
    phgFree(g->L2Gmap_vert);
    phgFree(g->L2Gmap_edge);
# if Dim == 3
    phgFree(g->L2Gmap_face);
# endif	/* Dim == 3 */
    phgFree(g->L2Gmap_elem);
    phgFree(g->neighbours.counts);
    phgFree(g->neighbours.displs);
    phgFree(g->neighbours.list);
#endif	/* USE_MPI */

    /* avoid warning on unfreed DOF */
    if (g->geom != NULL)
	phgDofFree(&g->geom);

    /* Free DOFs attached to the grid */
    if (g->dof != NULL) {
	DOF **p;
	for (p = g->dof; *p != NULL; p++);
	i = (INT)(p - g->dof);
	while (i > 0) {
	    DOF *d = g->dof[--i];
#if DEBUG
	    /* For helping finding out memory leaks due to unfreed DOFs */
	    phgPrintf("*** WARNING: unfreed DOF \"%s\" (created at %s:%d)\n",
			d->name, d->srcfile, d->srcline);
#endif
	    d->g = NULL; /* make phgGridIsValid return FALSE in phgDofFree */
	    phgDofFree(&d);
	}
	phgFree(g->dof);
    }
    phgFree(g);

    /* remove g from g_list */
    for (i = 0; i < g_list_count; i++)
	if (g_list[i] == g) break;

    if (i >= g_list_count)
	phgError(1, "phgFreeGrid: invalid grid!\n");

    n = g_list_count - i - 1;
    if (n > 0) {
	memmove(g_list + i, g_list + i + 1, n * sizeof(*g_list));
    }
    if (--g_list_count > 0) {
	g_list = phgRealloc(g_list, g_list_count * sizeof(*g_list));
    }
    else {
	phgFree(g_list);
	g_list = NULL;
    }

    /* for Master/Slave mode in which only one grid is allowed */
    g_sys = NULL;

    return;
}

void
phgFreeGrid(GRID **gptr)
{
    GRID *g = (gptr == NULL ? NULL : *gptr);

    if (g == NULL)
	return;

    if (gptr != NULL)
	*gptr = NULL;

    phgFreeGrid_(g);
}

/* flag controlling whether to call callback function for non-leaf elements */
static BOOLEAN traverse_all_flag = FALSE;

/* Depth of the element in the tree.
   This variable is made global and may be used by the callback functions */
INT phgTraverseDepth = 0, phgTraverseIndex = 0;

static BOOLEAN
traverse_element(SIMPLEX *e, BOOLEAN(*cb) CB_ARGS(e))
{
    BOOLEAN is_leaf = TRUE;

    if (e->children[0] != NULL) {
        is_leaf = FALSE;
	phgTraverseDepth++;
	if (!traverse_element(e->children[0], cb)) return FALSE;
	phgTraverseDepth--;
    }
    if (e->children[1] != NULL) {
        is_leaf = FALSE;
	phgTraverseDepth++;
	if (!traverse_element(e->children[1], cb)) return FALSE;
	phgTraverseDepth--;
    }
    if (!is_leaf && !traverse_all_flag) {
	return TRUE;
    }
    else {
	BOOLEAN ret = cb CB_ARGS0(g_traverse, e);
	Unused(g_traverse);
	phgTraverseIndex++;
	return ret;
    }
}

void
phgTraverseElements(GRID *g, BOOLEAN(*cb) CB_ARGS(e))
/* function traversing (traverse_all_flag ? leaf : all) elements.
 * Stop traversing if callback function returns FALSE */
{
    INT i;

    if (g->nroot == 0) return;

    g_traverse = g;
    phgTraverseIndex = 0;
    for (i = 0; i < g->nroot; i++) {
	phgTraverseDepth = 0;
	if (!traverse_element(g->roots + i, cb)) break;
    }
}

void
phgTraverseAllElements(GRID *g, BOOLEAN(*cb) CB_ARGS(e))
/* function traversing all (including non leaf) elements */
{
    traverse_all_flag = TRUE;
    phgTraverseElements(g, cb);
    traverse_all_flag = FALSE;
}

static BOOLEAN conformity; /* used to check conformity during traversal */

static struct {
    SIMPLEX **elements;
    SHORT   nelem, nmax;
} *vertices, *pv, *pv0[Dim];

static BOOLEAN
find_neighbours_callback0 CB_ARGS(s)
{
    INT k;
    int j;

    memset(s->neighbours, 0, sizeof(s->neighbours));

    for (j = 0; j < NVert; j++) {
	k = s->verts[j];
	pv = vertices + k;
	if (pv->nelem >= pv->nmax)
	    pv->elements = phgRealloc(pv->elements,
				(pv->nmax+=32) * sizeof(s));
	pv->elements[pv->nelem++] = s;
    }

    if (s->index == -1)
	s->index = phgTraverseIndex;

    return TRUE;
}

#define EXTENSIVE_TEST 0

#if DEBUG && EXTENSIVE_TEST
static BOOLEAN
singular_solve(int n, int m, double a[n][m], double b[n])
/* solves the singular system AX == B (A is a singular NxM matrix),
 * returns FALSE if no solution */
{
    int i, j, k;
    double d, e;

    assert(m <= n);

    /* normalize the coefficients */
    for (i = 0; i < n; i++) {
	d = fabs(a[i][0]);
	for (j = 1; j < m; j++) {
	    if (d < (e = fabs(a[i][j])))
		d = e;
	}
	if (d == 0.)
	    continue;
	d = 1. / d;
	for (j = 1; j < m; j++)
	    a[i][j] *= d;
	b[i] *= d;
    }

    /* Gauss elimination */
    for (i = 0; i < m; i++) {
	k = i;
	d = fabs(a[i][i]);
	for (j = i + 1; j < n; j++) {
	    if (d < (e = fabs(a[j][i]))) {
		d = e;
		k = j;
	    }
	}
	if (d < 1e-12)
	    return FALSE;
	if (k != i) {
	    /* exchange row i with row k */
	    d = b[i];
	    b[i] = b[k];
	    b[k] = d;
	    for (j = 0; j < m; j++) {
		d = a[i][j];
		a[i][j] = a[k][j];
		a[k][j] = d;
	    }
	}
	/* elimination */
	d = 1. / a[i][i];
	for (j = 0; j < m; j++)
	    a[i][j] *= d;
	b[i] *= d;
	for (j = i + 1; j < n; j++) {
	    for (k = i + 1; k < m; k++)
		a[j][k] -= a[j][i] * a[i][k];
	    b[j] -= a[j][i] * b[i];
	}
    }

    /* check consistency of the system */
    for (; i < n; i++) {
	if (fabs(b[i]) > 1e-12)
	    return FALSE;
    }

    /* back substitute */
    for (i = m - 2; i >= 0; i--)
	for (j = i + 1; j < m; j++)
	    b[i] -= a[i][j] * b[j];

    return TRUE;
}

static BOOLEAN
check_intersect(COORD *v0, COORD *v1, COORD *u0, COORD *u1)
/* returns TRUE if the segments v0-v1 and u0-u1 intersect */
{
    int i;
    double a[Dim][2], b[Dim];

    for (i = 0; i < Dim; i++) {
	b[i] = (*u0)[i] - (*v0)[i];
	a[i][0] = (*v1)[i] - (*v0)[i];
	a[i][1] = -((*u1)[i] - (*u0)[i]);
    }

    if (!singular_solve(Dim, 2, a, b))
	return FALSE;

    if (b[0] < -1e-12 || b[0] > 1 + 1e-12 || b[1] < -1e-12 || b[1] > 1 + 1e-12)
	return FALSE;

    return TRUE;
}

static void
check2(SIMPLEX *s, int face, int fa, int fb, int fc, SIMPLEX *s1)
/* check the case in which 's1' has two vertices on the give face of 's'
 * (vertices 'fa' and 'fb' are shared by 's' and 's1').
 *
 * This check handles a special case in which two edges of two faces
 * sharing two common vertices intersect. This typically happens when
 * a hexahedron is adjacent to a tetrahedron and is tetrahedralized
 * inconsistently */
{
    int j;
    INT v0, v1, v2, v3;
    COORD *coor;

    if (IS_BDRY(s->bound_type[face]))
	return;

    v0 = s->verts[GetFaceVertex(face, fa)],
    v1 = s->verts[GetFaceVertex(face, fb)],
    v2 = s->verts[GetFaceVertex(face, fc)];
    coor = g_->verts;

    /* make sure edges of s1 does not cross the face */
    for (j = 0; j < NVert; j++) {
	if ((v3 = s1->verts[j]) == v0 || v3 == v1)
	    continue;
	assert(v3 != v2);
	/* (v3-v0) should not intersect with (v2-v1) and
	   (v3-v1) should not intersect with (v2-v0) */
	if (check_intersect(coor + v3, coor + v0, coor + v2, coor + v1) ||
	    check_intersect(coor + v3, coor + v1, coor + v2, coor + v0))
	    phgError(1, "nonconforming mesh, abort.\n");
    }
}

typedef struct {
    SIMPLEX *e;
    BYTE v;
} ELIST;

static int
comp_elist(const void *p0, const void *p1)
{
    const ELIST *el0 = p0, *el1 = p1;
    return (el0->e > el1->e ? 1 : (el0->e == el1->e ? 0 : -1));
}

#endif	/* DEBUG && EXTENSIVE_TEST */

static BOOLEAN
find_neighbours_callback1 CB_ARGS(s)
{
#if 0
    /* new code based on sorting the list of neighbours of local vertices.
     * FIXME: the code is much slower and breaks mesh coarsening algorithm */
    SIMPLEX *s1;
    struct {
	INT	eno;
	BYTE	v;
    } *elist = NULL;
    int i, j, n, m;
    INT eno;

    /* count number of elements in the lists of the 3 vertices */
    n = 0;
    for (j = 0; j < NVert; j++) {
	s->neighbours[j] = NULL;
	n += vertices[s->verts[j]].nelem;
    }
    n -= NVert;
    if (n <= 0)
	goto cont;
    elist = phgAlloc(n * sizeof(*elist));
    /* collect element indices */
    n = 0;
    for (j = 0; j < NVert; j++) {
	for (i = 0; i < vertices[s->verts[j]].nelem; i++) {
	    if ((s1 = vertices[s->verts[j]].elements[i]) == s)
		continue;
	    elist[n].eno = s1 - g_->roots;
	    elist[n++].v = j;
	}
    }
    qsort(elist, n, sizeof(*elist), phgCompINT);

    eno = -1;	/* current element */
    m = 0;	/* count of duplicate elements */
    for (j = 0; j <= n; j++) {
	if (j < n && eno == elist[j].eno) {
	    m++;
	    continue;
	}
	assert(m <= 3);
	if (m == 3) {
	    BYTE v0 = elist[j - 3].v, v1 = elist[j - 2].v, v2 = elist[j - 1].v;
	    /* find the face number */
	    if (v0 != 0 && v1 != 0 && v2 != 0)
		i = 0;
	    else if (v0 != 1 && v1 != 1 && v2 != 1)
		i = 1;
	    else if (v0 != 2 && v1 != 2 && v2 != 2)
		i = 2;
	    else
		i = 3;
	    s->bound_type[i] &= ~(REMOTE | UNDEFINED);
	    s->bound_type[i] |= INTERIOR;
	    s->neighbours[i] = g_->roots + eno;
	}
	else if (m == 2) {
	    /* TODO: further conformity check */
	}
	if (j >= n)
	    break;
	eno = elist[j].eno;
	m = 1;
    }

cont:
    for (i = 0; i < NVert; i++) {
	if (s->neighbours[i] != NULL)
	    continue;
	if (s->bound_type[i] & INTERIOR) {
	    if (g_->nprocs == 1) {
		conformity = FALSE;
	    }
	    else {
		s->bound_type[i] |= REMOTE;
		s->neighbours[i] = NULL;
	    }
	}
    }

    phgFree(elist);
#else	/* #if 0|1 */
    SIMPLEX *s1 = NULL;
    int face, j, l;
    INT k;

    for (face = 0; face < NFace; face++) {
	k = 0;
	for (j = 0; j < NVert; j++) {
	    if (j == face)
		continue;
	    pv0[k++] = vertices + s->verts[j];
	}
	/* element appearing in all the Dim lists is the neighbour */
	for (k = 0; k < pv0[0]->nelem; k++) {
	    s1 = pv0[0]->elements[k];
	    if (s1 == s)
		continue;
	    for (l = 0; l < pv0[1]->nelem; l++)
		if (s1 == pv0[1]->elements[l])
		    break;
	    if (l >= pv0[1]->nelem)
		continue;
#if Dim == 3
	    for (l = 0; l < pv0[2]->nelem; l++)
		if (s1 == pv0[2]->elements[l])
		    break;
	    if (l >= pv0[2]->nelem)
		continue;
#endif	/* Dim == 3 */
	    break;
	}
	if (k < pv0[0]->nelem) {
	    s->bound_type[face] &= ~(REMOTE | UNDEFINED);
	    s->bound_type[face] |= INTERIOR;
	    s->neighbours[face] = s1;
	}
	else {
	    if (s->bound_type[face] & INTERIOR) {
		if (g_->nprocs == 1) {
		    conformity = FALSE;
		}
		else {
		    s->bound_type[face] |= REMOTE;
		    s->neighbours[face] = NULL;
		}
	    }
	    /* already cleared by callback0() */
	    /* s->neighbours[face] = NULL;*/
	}
#if DEBUG && EXTENSIVE_TEST	/* Slow! don't enable unless for debugging */
	/* Check intersected edges */
	{
	    static BOOLEAN warned = FALSE;
	    ELIST *elist = NULL;
	    int i, n, m;

	    if (!warned) {
		warned = TRUE;
		phgWarning("slow code! don't enable except for debugging.\n");
	    }

	    /* count number of elements in the lists of the 3 vertices */
	    n = pv0[0]->nelem + pv0[1]->nelem + pv0[2]->nelem - 3;
	    if (n <= 0)
		continue;
	    elist = phgAlloc(n * sizeof(*elist));
	    /* collect element indices */
	    n = 0;
	    for (j = 0; j < 3; j++) {
		for (i = 0; i < pv0[j]->nelem; i++) {
		    if ((s1 = pv0[j]->elements[i]) == s)
			continue;
		    elist[n].e = s1;
		    elist[n++].v = j;
		}
	    }
	    qsort(elist, n, sizeof(*elist), comp_elist);
	
	    s1 = NULL;	/* current element */
	    m = 0;	/* count of duplicate elements */
	    for (j = 0; j <= n; j++) {
		if (j < n && s1 == elist[j].e) {
		    m++;
		    continue;
		}
		assert(m <= 3);
		if (m == 2) {
		    int v0 = elist[j - 1].v, v1 = elist[j - 2].v, v2; 
		    if (((v2 = 0) != v0 && v2 != v1) ||
			((v2 = 1) != v0 && v2 != v1) ||
			((v2 = 2) != v0 && v2 != v1))
			check2(s, face, v0, v1, v2, s1);
		}
		if (j >= n)
		    break;
		s1 = elist[j].e;
		m = 1;
	    }
	}
#endif /* DEBUG && EXTENSIVE_TEST */
    }
#endif	/* #if 0|1 */

    return TRUE;
}
	
static BOOLEAN
check_conformity_callback CB_ARGS(e)
{
    int i, j, k;
    SIMPLEX *e1;

    for (i = 0; i < NVert; i++) {
	/* check boundary types */
#if 0
	if (e->bound_type[i] == 0) {
	    phgInfo(-1, "undefined boundray type for face %d\n", i);
	    phgDumpElement(g_, e);
	    return conformity = FALSE;
	}
#endif
	/* check indices of vertices */
	if (e->verts[i] < 0 || e->verts[i] >= g_->nvert) {
	    phgInfo(-1, "invalid vertex index %d\n", e->verts[i]);
	    phgDumpElement(g_, e);
	    return conformity = FALSE;
	}
	/* check links to neighbours */
	if (!HasLocalNeighbour(e, i))
	    continue;
	e1 = e->neighbours[i];
	if (e1 == NULL) {
	    phgInfo(-1, "interior face (%d) without neighbour:\n", i);
	    phgDumpElement(g_, e);
	    if (i > 1)
		phgDumpPatch(g_, e, 0, 1);
	    else
		phgDumpPatch(g_, e, 2, 3);
	    return conformity = FALSE;
	}
	if (!IsLeaf(e1)) {
	    phgInfo(-1, "non-root neighbour:\n");
	    phgDumpElement(g_, e);
	    phgDumpElement(g_, e1);
	    if (e1->children[0] != NULL)
		phgDumpElement(g_, e1->children[0]);
	    if (e1->children[1] != NULL)
		phgDumpElement(g_, e1->children[1]);
	    return conformity = FALSE;
	}
	/* check that the vertices of the common face (edge in 2D) match */
	for (j = 0; j < NVert; j++) {
	    INT v;
	    if (j == i) continue;
	    v = e->verts[j];
	    for (k = 0; k < NFace; k++) {
		if (v == e1->verts[k]) break;
	    }
	    if (k >= NFace) {
		phgInfo(-1, "non conforming face:\n");
		phgDumpElement(g_, e);
		phgDumpElement(g_, e1);
		return conformity = FALSE;
	    }
	}
#if Dim == 3
	if (HasLocalNeighbour(e, i) && g_->face_flag) {
	    j = phgOppositeVertex(e, i, e1 = e->neighbours[i]);
	    if (e->faces[i] != e1->faces[j]) {
		phgInfo(-1, "non conforming face indices: %d <-> %d\n", i, j);
		phgDumpElement(g_, e);
		phgDumpElement(g_, e1);
		return conformity = FALSE;
	    }
	}
#endif
    }

    /* TODO: check edge indices */

    return TRUE;
}

static BOOLEAN
check_conformity_nonleaf_callback CB_ARGS(e)
{
    int fmap[NVert], emap[NEdge];
    SIMPLEX *e0, *e1;

    /* TODO: check vertices */

    if ((e0 = e->children[0]) != NULL) {
	if (g_->elem_flag && e->index != e0->index &&
	    GlobalVertex(g_, e->verts[0]) < GlobalVertex(g_, e->verts[1])) {
	    phgDumpElement(g_, e);
	    phgDumpElement(g_, e0);
	    phgInfo(-1, "index of child 0 != index of parent\n");
	    return conformity = FALSE;
	}

	if (g_->edge_flag || g_->face_flag)
	    phgMapP2C(e, fmap, emap, 0);

	if (g_->edge_flag &&
	    (e->edges[1] != e0->edges[emap[1]] ||
	     e->edges[2] != e0->edges[emap[2]] ||
	     e->edges[5] != e0->edges[emap[5]] ||
	     (GlobalVertex(g_, e->verts[0]) < GlobalVertex(g_, e->verts[1]) &&
	      e->edges[0] != e0->edges[emap[0]]))) {
	    phgDumpElement(g_, e);
	    phgDumpElement(g_, e0);
	    phgInfo(-1, "edge indices of parent and child 0 mismatch\n");
	    return conformity = FALSE;
	}

	if (g_->face_flag &&
	    (e->faces[1] != e0->faces[3] ||
	     (GlobalVertex(g_, e->verts[0]) < GlobalVertex(g_, e->verts[1]) &&
	      (e->faces[2] != e0->faces[fmap[2]] ||
	       e->faces[3] != e0->faces[fmap[3]])))) {
	    phgDumpElement(g_, e);
	    phgDumpElement(g_, e0);
	    phgInfo(-1, "face indices of parent and child 0 mismatch\n");
	    return conformity = FALSE;
	}
    }

    if ((e1 = e->children[1]) != NULL) {
	if (g_->elem_flag && e->index != e1->index &&
	    GlobalVertex(g_, e->verts[0]) > GlobalVertex(g_, e->verts[1])) {
	    phgDumpElement(g_, e);
	    phgDumpElement(g_, e0);
	    phgInfo(-1, "index of child 0 != index of parent\n");
	    return conformity = FALSE;
	}

	if (g_->edge_flag || g_->face_flag)
	    phgMapP2C(e, fmap, emap, 1);

	if (g_->edge_flag &&
	    (e->edges[3] != e1->edges[emap[3]] ||
	     e->edges[4] != e1->edges[emap[4]] ||
	     e->edges[5] != e1->edges[emap[5]] ||
	     (GlobalVertex(g_, e->verts[0]) > GlobalVertex(g_, e->verts[1]) &&
	      e->edges[0] != e1->edges[emap[0]]))) {
	    phgDumpElement(g_, e);
	    phgDumpElement(g_, e1);
	    phgInfo(-1, "edge indices of parent and child 1 mismatch\n");
	    return conformity = FALSE;
	}

	if (g_->face_flag &&
	    (e->faces[0] != e1->faces[3] ||
	     (GlobalVertex(g_, e->verts[0]) > GlobalVertex(g_, e->verts[1]) &&
	      (e->faces[2] != e1->faces[fmap[2]] ||
	       e->faces[3] != e1->faces[fmap[3]])))) {
	    phgDumpElement(g_, e);
	    phgDumpElement(g_, e1);
	    phgInfo(-1, "face indices of parent and child 1 mismatch\n");
	    return conformity = FALSE;
	}
    }

    return TRUE;
}

#if USE_MPI
static int
get_edge_no(int v0, int v1)
/* returns local number of an edge, given its two vertices */
{
    int v;
 
    assert(v0 != v1);
    if (v0 > v1) {
	v = v0;
	v0 = v1;
	v1 = v;
    }

    switch (v0 * 10 + v1) {
	case 01:
	    return 0;
	case 02:
	    return 1;
	case 03:
	    return 2;
	case 12:
	    return 3;
	case 13:
	    return 4;
	case 23:
	    return 5;
    }

    return -1;
}

static INT (*dup_list)[NVert] = NULL;

static BOOLEAN
count_duplicate_elements(SIMPLEX *e)
{
    if (IsLeaf(e)) {
	e->flag = 0;
	return FALSE;
    }

    if (e->children[0] != NULL)
	e->flag = count_duplicate_elements(e->children[0]);
    if (e->children[1] != NULL)
	e->flag |= count_duplicate_elements(e->children[1]);

    if ((e->children[0] == NULL || e->children[1] == NULL))
	e->flag = TRUE;

    if (e->flag)
	index_++;

    return e->flag;
}

static BOOLEAN
collect_duplicate_elements CB_ARGS(e)
{
    if (e->flag) {
	dup_list[index_][0] = GlobalVertex(g_, e->verts[0]);
	dup_list[index_][1] = GlobalVertex(g_, e->verts[1]);
	dup_list[index_][2] = GlobalVertex(g_, e->verts[2]);
	dup_list[index_][3] = GlobalVertex(g_, e->verts[3]);
	index_++;
    }
    return TRUE;
}

static int
duplicate_comp(const void *p0, const void *p1)
{
    int i;
    const INT *v0 = p0, *v1 = p1;
    if ((i = *(v0++) - *(v1++)) != 0)
	return i;
    if ((i = *(v0++) - *(v1++)) != 0)
	return i;
    if ((i = *(v0++) - *(v1++)) != 0)
	return i;
    return *v0 - *v1;
}
#endif

#if DEBUG && USE_MPI
static void
check_L2Gmap(INT *map, INT n, INT nglobal, const char *f, int l)
{
    INT *buf;
    INT i;

    if (map == NULL) return;

    for (i = 0; i < n; i++) {
	INT j = map[i];
	if (j < 0 || j >= nglobal)
	    phgError(1, "%s:%d, invalid entry: map[%d]=%d\n", f, l, i, j);
    }

    if (n <= 1) return;
    buf = phgAlloc(n * sizeof(*map));
    memcpy(buf, map, n * sizeof(*map));
    qsort(buf, n, sizeof(*map), phgCompINT);

    for (i = 0; i < n - 1; i++) {
	if (buf[i] == buf[i + 1])
	    phgError(1, "%s:%d, duplicate value: %d\n", f, l, buf[i]);
    }
    free(buf);
}

# define CheckL2Gmap(map, n, ng) check_L2Gmap(map, n, ng, __FILE__, __LINE__)
#else	/* DEBUG && USE_MPI */
# define CheckL2Gmap(map, n, ng)
#endif	/* DEBUG && USE_MPI */

void
phgCheckConformity(GRID *g)
{
    ParallelFunction(phgCheckConformity, g != NULL && g->nprocs > 1);

    if (g == NULL) return;

#if USE_MPI
    CheckL2Gmap(g->L2Gmap_vert, g->nvert, g->nvert_global);
    if (g->edge_flag)
	CheckL2Gmap(g->L2Gmap_edge, g->nedge, g->nedge_global);
    if (g->face_flag)
	CheckL2Gmap(g->L2Gmap_face, g->nface, g->nface_global);
    if (g->elem_flag)
	CheckL2Gmap(g->L2Gmap_elem, g->nelem, g->nelem_global);
#endif

    g_ = g;
    conformity = TRUE;
    phgTraverseElements(g, check_conformity_callback);

    if (conformity)
	phgTraverseAllElements(g, check_conformity_nonleaf_callback);

#if USE_MPI
    if (g->nprocs > 1) {
	int c0 = conformity, c;
	MPI_Allreduce(&c0, &c, 1, MPI_INT, MPI_SUM, g->comm);
	conformity = (c == g->nprocs);
    }
#endif

    if (!conformity) {
	phgError(1, "local non conformity.\n");
    }

#if USE_MPI
    if (conformity && g->nprocs > 1) {
	struct {
	    RNEIGHBOUR rn;	/* the RNEIGHBOUR data */
	    INT verts[NFace - 1]; /* global indices of the 3 shared vertices */
	    INT edges[NEdge];
#if Dim == 3
	    INT face_index;
#endif
	} *local, *remote, *p;
	RNEIGHBOUR *rlist = phgAlloc(g->neighbours.count * sizeof(*rlist));
	MPI_Datatype type;
	INT *verts;
	int i, j, k;
	int *counts = g->neighbours.counts, *displs = g->neighbours.displs;

	local = phgAlloc(g->neighbours.count * sizeof(*local));
	remote = phgAlloc(g->neighbours.count * sizeof(*remote));

	for (i = 0; i < g->neighbours.count; i++) {
	    int vertex;
	    SIMPLEX *e;

	    local[i].rn = g->neighbours.list[i];
	    vertex = local[i].rn.vertex;
	    e = local[i].rn.local;
	    verts = e->verts;
	    for (k = 0, j = 0; j < NFace; j++) {
		if (j == vertex) continue;
		local[i].verts[k++] = GlobalVertex(g, verts[j]);
	    }
	    local[i].rn.depth = 0;
	    for (j = 0; j < NEdge; j++) {
		local[i].edges[j] = GlobalEdge(g, e->edges[j]);
	    }
#if Dim == 3
	    local[i].face_index = GlobalFace(g, e->faces[vertex]);
#endif
	}

	MPI_Type_contiguous(sizeof(*local), MPI_BYTE, &type);
	MPI_Type_commit(&type);
	phgAlltoallv(local,  counts, displs, type,
		      remote, counts, displs, type, g->comm);
	MPI_Type_free(&type);

	for (i = 0; i < g->neighbours.count && conformity; i++) {
	    RNEIGHBOUR *rn = &remote[i].rn;
	    RNEIGHBOUR *rn0 = GetRNeighbour(g, rn->remote, rn->op_vertex);
	    p = remote + i;
	    if (!IsLeaf(rn->remote)) {
		phgWarning("element linked to non-leaf element.\n");
		phgWarning("remote data: local=%p, remote=%p, vertex=%d, "
			   "op=%d, rface=%d %d %d\n", rn->remote, rn->local,
			   rn->vertex, rn->op_vertex, rn->rface[0],
			   rn->rface[1], rn->rface[2]);
		phgWarning("local  data: local=%p, remote=%p, vertex=%d, "
			   "op=%d, rface=%d %d %d\n", rn0->remote, rn0->local,
			   rn0->vertex, rn0->op_vertex, rn0->rface[0],
			   rn0->rface[1], rn0->rface[2]);
		break;
	    }
	    if (local[i].rn.depth != 0) {
		phgWarning("different entries linked to same element.\n");
		break;
	    }
	    local[i].rn.depth = 1;
	    if (rn0->vertex != rn->op_vertex || rn0->op_vertex != rn->vertex ||
		rn0->local != rn->remote || rn0->remote != rn->local ||
		rn->rank != g->rank) {
		conformity = FALSE;
		phgWarning("global non conformity: linked data don't match.\n");
		break;
	    }
	    /* check whether the vertices match */
	    verts = rn0->local->verts;
	    for (j = 0; j < (NFace - 1) && conformity; j++) {
		INT v = GlobalVertex(g, verts[rn->rface[j]]);
		if (v == p->verts[j]) continue;
		conformity = FALSE;
		phgWarning("global non conformity: vertices don't match.\n");
		phgWarning("local=(%p %d %d%d%d), remote=(%p %d %d%d%d)\n",
			rn0->local, rn0->vertex, rn0->rface[0], rn0->rface[1],
			rn0->rface[2], rn0->remote, rn0->op_vertex,
			rn->rface[0], rn->rface[1], rn->rface[2]);
		phgWarning("remote vertices=%d %d %d\n",
			p->verts[0], p->verts[1], p->verts[2]);
		v = verts[rn->rface[j]];
		phgWarning("unmatched vertices: (%lf %lf %lf)\n",
			(double)g->verts[v][0], (double)g->verts[v][1],
			(double)g->verts[v][2]);
		phgDumpElement(g, rn0->local);
		break;
	    }
	    if (g->edge_flag) {
	      /* check whether the edges match */
	      int a = 0, b = 1;
	      for (j = 0; j < NVert - 1 && conformity; j++) {
		if (j == rn0->vertex) continue;
		for (k = j + 1; k < NVert; k++) {
		    int u, v;
		    u = rn0->rface[a];
		    v = rn0->rface[b];
		    if (k == rn0->vertex) continue;
		    if (GlobalEdge(g, rn0->local->edges[get_edge_no(j, k)]) !=
			remote[i].edges[get_edge_no(u, v)]) {
			conformity = FALSE;
			phgWarning("non conformity: edges don't match.\n");
			phgWarning("unmatched edge: %d-%d %d <==> %d-%d %d\n",
				j, k, get_edge_no(j, k),
				u, v, get_edge_no(u, v));
			phgWarning("remote edges: %d %d %d %d %d %d\n",
				remote[i].edges[0], remote[i].edges[1],
				remote[i].edges[2], remote[i].edges[3],
				remote[i].edges[4], remote[i].edges[5]);
			phgWarning("remote vertices=%d %d %d\n",
				p->verts[0], p->verts[1], p->verts[2]);
			phgDumpElement(g, rn0->local);
			break;
		    }
		    if (++b >= 3) {
			a++;
			b = a + 1;
		    }
		}
	      }
	    }
#if Dim == 3
	    if (g->face_flag) { /* check whether the faces match */
		SIMPLEX *e = remote[i].rn.remote;
		int vertex = remote[i].rn.op_vertex;
		if (GlobalFace(g, e->faces[vertex]) != remote[i].face_index) {
		    phgWarning("non conformity: face indices don't match.\n");
		    phgWarning("local(%p,%d)=%d, remote(%p,%d)=%d.\n",
			e, vertex, GlobalFace(g, e->faces[vertex]),
			remote[i].rn.local, remote[i].rn.vertex,
			remote[i].face_index);
		    break;
		}
	    }
#endif
	}

	for (i = 0; i < g->neighbours.count && conformity; i++) {
	    if (local[i].rn.depth == 1) continue;
	    phgWarning("global non conformity: unmatched link(s)!\n");
	    break;
	}

	phgFree(rlist);
	phgFree(local);
	phgFree(remote);
	
	i = conformity;
	MPI_Allreduce(&i, &j, 1, MPI_INT, MPI_MIN, g->comm);
	conformity = j;
    }

    /* check vertex indices on duplicate non-leaf elements */
    if (FALSE && conformity && g->nprocs > 1) {
	INT k, l, *counts, nmax;
	INT (*buffer)[NVert], (*p0)[NVert], (*p1)[NVert];
	MPI_Status status;
	MPI_Datatype type;
	int i, j, m, src, dst;

	index_ = 0;
	for (k = 0; k < g->nroot; k++)
	    count_duplicate_elements(g->roots + k);
	dup_list = phgAlloc(index_ * sizeof(*dup_list));
	counts = phgAlloc(g->nprocs * sizeof(*counts));
	MPI_Allgather(&index_, sizeof(index_), MPI_BYTE,
			counts, sizeof(index_), MPI_BYTE, g->comm);
	nmax = 0;
	for (i = 0; i < g->nprocs; i++)
	    if (nmax < counts[i])
		nmax = counts[i];
	buffer = phgAlloc(2 * nmax * sizeof(*dup_list));
	index_ = 0;
	phgTraverseAllElements(g_, collect_duplicate_elements);
	qsort(dup_list, index_, sizeof(*dup_list), duplicate_comp);
	/* circulate and compare the lists */
	j = g->rank;
	dst = (g->rank == 0) ? g->nprocs - 1 : g->rank - 1;
	src = (g->rank == g->nprocs - 1) ? 0 : g->rank + 1;
	MPI_Type_contiguous(sizeof(*buffer), MPI_BYTE, &type);
	MPI_Type_commit(&type);
	p0 = dup_list;
	p1 = buffer;
	for (i = 0; i < g->nprocs - 1; i++) {
	    p1 = buffer + (nmax - (p1 - buffer));
	    if ((l = j + 1) >= g->nprocs)
		l = 0;
	    MPI_Sendrecv(p0, counts[j], type, dst, 111,
			 p1, counts[l], type, src, 111, g->comm, &status);
	    j = l;
	    p0 = p1;
	    /* compare dup_list with p1 */
	    k = 0;
	    l = 0;
	    while (TRUE) {
		while (k < index_ && dup_list[k][0] < 0)
		    k++;
		if (k >= index_)
		    break;
		while (l < counts[j]) {
		    if ((m = duplicate_comp(p0 + l, dup_list + k)) >= 0)
			break;
		    l++;
		}
		if (l >= counts[j])
		    break;
		while (m > 0 && k < index_) {
		    if (dup_list[k][0] >=0 &&
			(m = duplicate_comp(p0 + l, dup_list + k)) <= 0)
			break;
		    k++;
		}
		if (k >= index_)
		    break;
		/* a matching pair of tetrahedra found */
		dup_list[k][0] = dup_list[k][1] = dup_list[k][2]
							= dup_list[k][3] = -1;
		k++;
		l++;
	    }
	}
	MPI_Type_free(&type);
	phgFree(buffer);
	phgFree(counts);
	/* check whether we have unmatched tetrahedra */
	for (k = 0; k < index_; k++) {
	    if (dup_list[k][0] >= 0)
		phgError(1, "%s:%d, tetrahedra: %d %d %d %d\n",
			__FILE__, __LINE__, dup_list[k][0],
			dup_list[k][1], dup_list[k][2], dup_list[k][3]);
	}
	phgFree(dup_list);
    }
#endif	/* USE_MPI */

    if (conformity) {
	if (IS_RANK0) phgInfo(1, "the grid is conforming.\n");
    }
    else {
#if 0
	phgDumpGrid(g);
#endif
	phgError(1, "%s:%d: abort.", __FILE__, __LINE__);
    }

    return;
}

#if 0
static int
elem_comp(const void *p0, const void *p1)
{
    return phgMemcmp((const SIMPLEX *)p0, (const SIMPLEX *)p1, sizeof(SIMPLEX));
}
#endif

void
phgUpdateNeighbours(GRID *g)
/* fills 'neighbours' array of all simplices. */
{
    INT i;

    FunctionEntry;

    if (g == NULL) return;

    g_ = g;

#if 0
    if (g->nprocs == 1 && g->nroot == g->nleaf) {
	/* check for duplicate tetrahedra in the initial mesh */
	INT j;
	qsort(g->roots, g->nroot, sizeof(*g->roots), elem_comp);
	j = -1;
	for (i = 0; i < g->nroot; i++) {
	    if (j < 0 || elem_comp(g->roots + j, g->roots + i) != 0)
		j++;
	    if (j != i)
		g->roots[j] = g->roots[i];
	}
	if (j < g->nroot)
	    phgWarning("duplicate tetrahedra found and removed.\n");
    }
#endif

    vertices = phgCalloc(g->nvert, sizeof(*vertices));
#if 0
    for (i = 0; i < g->nvert; i++) {
	 vertices[i].nelem = vertices[i].nmax = 0;
	 vertices[i].elements = NULL;
    }
#endif

    /* For each vertex, a list of elements to which it belongs is built */
    phgTraverseElements(g, find_neighbours_callback0);

    /* then search neighbours in the small lists of elements */
    conformity = TRUE;
    phgTraverseElements(g, find_neighbours_callback1);

    /* free memory blocks */
    for (i = 0; i < g->nvert; i++)
	if (vertices[i].elements != NULL)
	    phgFree(vertices[i].elements);
    phgFree(vertices);

    PrintTime(1);
    Return;
}

/* Note: enabling REMOVE_DEGENERATED_ELEMENTS causes PHG to crash
 * on some bad input meshes (e.g., cask1.mesh, cask.mesh, etc.)
 *
 * It is simpler to move the code to before updating neighbour links,
 * but the present code allows transferring boundary types of deleted
 * elements. */
#define REMOVE_DEGENERATED_ELEMENTS 1

void
phgRemoveDegeneratedElements(GRID *g)
/* checks and removes degenerated elements and unused vertices.
 *
 * Note: this function should be called after updating the neighbour links,
 *	 necessary for transfering boundary types of deleted elements */
{
    INT *vmap;
    INT i, j;
    SIMPLEX *e;
#if REMOVE_DEGENERATED_ELEMENTS
    INT k, u, v, w;
    SIMPLEX *e0, *e1;
    int a, b, c;
    BOOLEAN flag;
#endif	/* REMOVE_DEGENERATED_ELEMENTS */

    FunctionEntry;

#if USE_MPI
    assert(g->nprocs <= 1);
    assert(g->L2Gmap_vert == NULL);
#endif
    assert(g->serial_no == 0);

    if (g->nvert <= 0)
	Return;

    vmap = phgAlloc(g->nvert * sizeof(*vmap));

    /* check and remove tetrahedra with zero length edges.
     * TODO: check for other types of degenerated tetrahedra */

#if REMOVE_DEGENERATED_ELEMENTS
    for (j = 0; j < g->nvert; j++)
	vmap[j] = j;
    flag = FALSE;
    k = 0;
    for (j = 0; j < g->nroot; j++) {
	e = g->roots + j;
	for (i = 0; i < NEdge; i++) {
	    u = e->verts[a = GetEdgeVertex(i, 0)];
	    v = e->verts[b = GetEdgeVertex(i, 1)];
	    if (g->verts[u][0] == g->verts[v][0] &&
		g->verts[u][1] == g->verts[v][1] &&
		g->verts[u][2] == g->verts[v][2])
		break;
	}
	if (i >= NEdge) {
	    /* no zero length edge in this element */
	    e->index = k++;
	    continue;
	}
	e->index = -1;	/* mark e as deleted */
	/* mark the vertex with larger value as deleted */
	w = (u > v) ? v : u;
	while (vmap[w] != w)
	    w = vmap[w];
	vmap[u] = vmap[v] = w;
	if (flag)
	    continue;
	/* update neighbour links and boundary types of the neighbours */
	e0 = e->neighbours[a];
	e1 = e->neighbours[b];
	if (e0 != NULL) {
	    for (c = 0; c < NFace; c++)
		if (e0->neighbours[c] == e)
		    break;
	    if (c >= NVert) {
		flag = TRUE;
		continue;
	    }
	    e0->neighbours[c] = e1;
	    e0->bound_type[c] = e->bound_type[b];
	}
	if (e1 != NULL) {
	    for (c = 0; c < NFace; c++)
		if (e1->neighbours[c] == e)
		    break;
	    if (c >= NVert) {
		flag = TRUE;
		continue;
	    }
	    e1->neighbours[c] = e0;
	    e1->bound_type[c] = e->bound_type[a];
	}
    }

    /* update vmap[] */
    if (k < g->nroot) {
	for (j = 0; j < g->nvert; j++) {
	    v = j;
	    while (v != vmap[v])
		v = vmap[v];
	    vmap[j] = v;
	}
    }

    /* remove tetrahedra with duplicate vertices */
    k = 0;
    for (j = 0; j < g->nroot; j++) {
	INT verts[NVert];
	e = g->roots + j;
	if (e->index == -1)
	    continue;
	for (i = 0; i < NVert; i++)
	    verts[i] = e->verts[i] = vmap[e->verts[i]];
	qsort(verts, NVert, sizeof(*verts), phgCompINT);
	if (verts[0] < 0 || verts[NVert - 1] >= g->nvert) {
	    flag = TRUE;
	    e->index = -1;
	    continue;
	}
	for (i = 1; i < NVert; i++) {
	    if (verts[i] == verts[i - 1])
		break;
	}
	if (i < NVert) {
	    flag = TRUE;
	    e->index = -1;
	    continue;
	}
	e->index = k++;
    }

    if (k < g->nroot) {
	/* some elements are to be deleted, update neighbours and vertices */
	i = g->nroot - k;
	phgInfo(2, "deleting %d degenerated element%s\n", i, i > 1 ? "s" : "");
	for (j = 0; j < g->nroot; j++) {
	    e = g->roots + j;
	    if (e->index == -1)
		continue;
	    for (i = 0; i < NVert; i++) {
		if ((e0 = e->neighbours[i]) != NULL) {
		    if (!flag) {
			assert(e0->index >= 0);
			e->neighbours[i] = g->roots + (e0->index);
		    }
		    else {
			e->neighbours[i] = NULL;
		    }
		}
	    }
	}
	for (j = k = 0; j < g->nroot; j++) {
	    if (g->roots[j].index == -1)
		continue;
	    if (j != k)
		g->roots[k] = g->roots[j];
	    k++;
	}
	g->ntree = g->nleaf = g->nelem = g->nroot = g->nelem_global = k;
	if (flag) {
	    phgWarning("bdry type of deleted elements may not be "
			"correctly transferred.\n");
	    phgUpdateNeighbours(g);
	}
    }
#endif	/* REMOVE_DEGENERATED_ELEMENTS */

    /* remove unused vertices, they may either come from the imported file
     * or from the deleted elements */

    memset(vmap, 0, g->nvert * sizeof(*vmap));
    for (j = 0; j < g->nroot; j++) {
	e = g->roots + j;
	for (i = 0; i < NVert; i++)
	    vmap[e->verts[i]] = 1;
    }

    /* compute new index in vmap and update g->verts */
    j = 0;
    for (i = 0; i < g->nvert; i++) {
	if (vmap[i] == 0) {
	    /* vertex unused */
	    vmap[i] = -1;
	    continue;
	}
	vmap[i] = j;
	if (j != i)
	    memcpy(g->verts + j, g->verts + i, sizeof(COORD));
	j++;
    }

    if (j == g->nvert) {
	/* no unused vertices */
	phgFree(vmap);
	Return;
    }

    i = g->nvert - j;
    phgInfo(2, "deleting %d unused vert%s\n", i, i > 1 ? "ices" : "ex");
    g->nvert = g->nvert_global = j;
    if (j == 0)
	g->verts = NULL;
    else {
	g->verts = phgRealloc(g->verts, j * sizeof(*g->verts));
	for (j = 0; j < g->nroot; j++) {
	    e = g->roots + j;
	    for (i = 0; i < NVert; i++)
		e->verts[i] = vmap[e->verts[i]];
	}
    }

    phgFree(vmap);
    Return;
}

static SIMPLEX *
find_edge_neighbour(SIMPLEX *e, INT u0, INT u1, INT *u)
/* Returns the neighbour not containing the face {u0,u1,u}, but containing
 * the edge u0-u1, where u0, u1, *u are global vetex indices. On return
 * *u is set to the local index of the third vertex of the common face. */
{
    int i, j;

    /* one of the vertices must be equal to "*u" */
    if (e->verts[i = 0] != *u && e->verts[++i] != *u &&
	e->verts[++i] != *u && e->verts[++i] != *u) {
	phgInfo(-1, "u0=%d, u1=%d, *u=%d\n", u0, u1, *u);
	phgDumpElement(NULL, e);
	phgError(1, "%s:%d, unexpected error!\n", __FILE__, __LINE__);
    }

    for (j = 0; j == i || e->verts[j] == u0 || e->verts[j] == u1; j++);
    *u = e->verts[j];

    /* it is assumed that e->neighbours[i] == NULL for all non-interior and
     * REMOTE neighbours */
    return IsLeaf(e) ? GetNeighbour(e, i) : e->neighbours[i];
}

INT
phgGetPatch(GRID *g, SIMPLEX *e, BYTE v0, BYTE v1, SIMPLEX ***p)
/* constructs the list of elements constituting the patch around
 * the edge v0-v1 (local indices) of e. returns number of elements
 * in the patch and list of elements through 'p' */
/* TODO: patch composed of several non contiguous parts */
{
    INT u0 = e->verts[v0], u1 = e->verts[v1], u;
    /* gcc is unhappy if v2, v3 are not initialized :( */
    BYTE v2 = 0, v3 = 0;
    INT count = 0, size;
    SIMPLEX *e1;

    *p = phgAlloc((size = 6) * sizeof(**p));
    (*p)[count++] = e;

    /* find out the two opposite vertices, v2 and v3, of the edge */
    switch (v0 * 4 + v1) {
	case 0 * 4 + 1: v2 = 2; v3 = 3; break;
	case 0 * 4 + 2: v2 = 1; v3 = 3; break;
	case 0 * 4 + 3: v2 = 1; v3 = 2; break;
	case 1 * 4 + 2: v2 = 0; v3 = 3; break;
	case 1 * 4 + 3: v2 = 0; v3 = 2; break;
	case 2 * 4 + 3: v2 = 0; v3 = 1; break;
    }
    
    /* Note: refine_patch depends on the direction, don't change! */

    /* search in first direction */
    e1 = IsLeaf(e) ? GetNeighbour(e, v2) : e->neighbours[v2];
    u = e->verts[v3];
    while (e1 != e && e1 != NULL) {
	if (count >= size) {
	    if (count > 100)
		phgError(1, "%s:%d, unexpected error.\n", __FILE__, __LINE__);
	    *p = phgRealloc(*p, (size += 6) * sizeof(**p));
	}
	(*p)[count++] = e1;
	e1 = find_edge_neighbour(e1, u0, u1, &u);
    }
    if (e1 == e) {
	phgDebug((4, "circular patch, size = %d\n", count));
	return count;
    }
    /* search in second direction */
#if 0
{
    phgInt count0 = count;	/* save position */
#endif
    e1 = IsLeaf(e) ? GetNeighbour(e, v3) : e->neighbours[v3];
    u = e->verts[v2];
    /* Note: e1 may come back to e when dealing with non-leaf elements
     * on a distributed mesh */
    while (e1 != e && e1 != NULL) {
	if (count >= size) {
	    if (count > 100)
		phgError(1, "%s:%d, unexpected error.\n", __FILE__, __LINE__);
	    *p = phgRealloc(*p, (size += 6) * sizeof(**p));
	}
	(*p)[count++] = e1;
	e1 = find_edge_neighbour(e1, u0, u1, &u);
#if DEBUG
	if (e1 == e) {
	    int i;
	    phgInfo(-1, "unexpected case: e1 == e,\n");
	    for (i = 0; i < count; i++)
		phgDumpElement(g, (*p)[i]);
	    phgError(1, "abort.\n");
	}
#endif
    }

#if 0
    /* We have the list of elements around the patch in the following order:
     *		*p[0] -> *p[1] -> ... -> *p[count0-1]
     *		*p[0] <- *p[count0] <- ... <- *p[count-1]
     *
     * We first commute (*p[0], *p[count0-1]), (*p[1], *p[count0-2]), etc.,
     * the we commute (*p[0], *p[count-1]), (*p[1], *p[count-2]), etc.,
     * such that the elements in the list are chained in the right direction */
    for (u0 = 0, u1 = count0 - 1; u0 < u1; u0++, u1--) {
	e = (*p)[u0];
	(*p)[u0] = (*p)[u1];
	(*p)[u1] = e;
    }
    for (u0 = 0, u1 = count - 1; u0 < u1; u0++, u1--) {
	e = (*p)[u0];
	(*p)[u0] = (*p)[u1];
	(*p)[u1] = e;
    }
}
#endif

    phgDebug((4, "open patch, size = %d\n", count));
    return count;    
}

#if USE_MPI
/* functions for managing global edge numbers */

static INT nedge_save, nedge_global_save;

static struct {
    SIMPLEX *e;		/* first element on the patch */
    BYTE v0, v1;	/* vertices of the edge */
    BYTE shared;	/* whether the edge is shared */
} *edges_local = NULL;
static int edges_local_count = 0, edges_local_allocated = 0;

static struct {
    INT index;		/* index to edges_local */
    INT v0, v1;		/* global indices of the two vertices */
} *edges_shared = NULL, *edges_gathered = NULL;
static int edges_shared_count = 0, edges_shared_allocated = 0;

static void
add_local_edge(SIMPLEX *e, int v0, int v1, BOOLEAN shared)
{
    if (edges_local_count >= edges_local_allocated)
	edges_local = phgRealloc(edges_local, sizeof(*edges_local) *
			(edges_local_allocated += 4096));
    edges_local[edges_local_count].e = e;
    edges_local[edges_local_count].v0 = v0;
    edges_local[edges_local_count].v1 = v1;
    edges_local[edges_local_count++].shared = shared;
}

static int
comp_edge0(const void *p0, const void *p1)
{
    int i = *((const int *)p0), j = *((const int *)p1);
    INT v0 = edges_local[i].e->verts[edges_local[i].v0];
    INT v1 = edges_local[i].e->verts[edges_local[i].v1];
    INT u0 = edges_local[j].e->verts[edges_local[j].v0];
    INT u1 = edges_local[j].e->verts[edges_local[j].v1];
    INT c;

    if (v0 > v1) {
	c = v0; v0 = v1; v1 = c;
    }

    if (u0 > u1) {
	c = u0; u0 = u1; u1 = c;
    }

    if ((c = v0 - u0) != 0)
	return (c > 0) ? 1 : -1;
    return ((c = v1 - u1) > 0) ? 1 : (c < 0) ? -1 : 0;
}

static int
comp_edge(const void *p0, const void *p1)
{
    int i;

    if ((i = comp_edge0(p0, p1)) != 0)
	return i;
    return edges_local[*((const int *)p1)].shared -
	   edges_local[*((const int *)p0)].shared;
}

static void
add_shared_edge(int index, INT v0, INT v1)
{
    if (edges_shared_count >= edges_shared_allocated)
	edges_shared = phgRealloc(edges_shared, sizeof(*edges_shared) *
			(edges_shared_allocated += 4096));

    if (v0 < v1) {
	edges_shared[edges_shared_count].v0 = v0;
	edges_shared[edges_shared_count].v1 = v1;
    }
    else {
	edges_shared[edges_shared_count].v0 = v1;
	edges_shared[edges_shared_count].v1 = v0;
    }
    edges_shared[edges_shared_count++].index = index;
}

static int
comp_edge_index(const void *p0, const void *p1)
{
    int i = *((const int *)p0), j = *((const int *)p1);
    INT v0 = edges_gathered[i].v0, v1 = edges_gathered[i].v1;
    INT u0 = edges_gathered[j].v0, u1 = edges_gathered[j].v1;
    INT c;

    if ((c = v0 - u0) != 0)
	return (c > 0) ? 1 : -1;
    return ((c = v1 - u1) > 0) ? 1 : (c < 0) ? -1 : 0;
}

static void
update_edge_no(int index, INT no)
/* update an edge number */
{
    int j, k, n;
    SIMPLEX **p, *e;
    INT v0, v1;

    e = edges_local[index].e;
    v0 = e->verts[edges_local[index].v0];
    v1 = e->verts[edges_local[index].v1];
    n = phgGetPatch(g_, e, edges_local[index].v0, edges_local[index].v1, &p);
    for (j = 0; j < n; j++) {
	e = p[j];
	for (k = 0; k < NEdge; k++) {
	    INT u0 = e->verts[GetEdgeVertex(k, 0)],
		u1 = e->verts[GetEdgeVertex(k, 1)];
	    if ((v0 == u0 && v1 == u1) || (v0 == u1 && v1 == u0))
		break;
	}
	assert(k < NEdge);
	e->edges[k] = no;
	phgDebug((4, "update_edge_no: e=%p, edge=%d, new=%d\n", e, k, no));
    }
    phgFree(p);
}

static void
sync_edges(void)
/* update saved edge numbers. */
{
    int *nbuffer, *counts, *displs, tmp[2], *index;
    int i, j, flag, local_count, total_local, total_shared;
    INT edge_no_local = 0, edge_no_global = 0;
    MPI_Datatype type;
    void *p;

    /* sort indices of edges_local */
    index = NULL;
    if (edges_local_count > 0) {
	index = phgAlloc(edges_local_count * sizeof(*index));
	for (i = 0; i < edges_local_count; i++)
	    index[i] = i;
	qsort(index, edges_local_count, sizeof(*index), comp_edge);
    }

    /* count # of private edges and build list of shared edges */
    i = local_count = 0;
    while (i < edges_local_count) {
	if (edges_local[j = index[i]].shared) {
 	    add_shared_edge(j,
		GlobalVertex(g_, edges_local[j].e->verts[edges_local[j].v0]),
		GlobalVertex(g_, edges_local[j].e->verts[edges_local[j].v1]));
	    i++;
	    continue;
	}
	j = i;
	while (++i < edges_local_count
		&& comp_edge0(index + j, index + i) == 0);
	local_count++;
    }

    phgDebug((3, "edges: local=%d (%d), shared=%d, total=%d\n", local_count,
			edges_local_count,
			edges_shared_count, local_count + edges_shared_count));

    nbuffer = phgAlloc(2 * g_->nprocs * sizeof(*counts));
    counts = nbuffer;
    displs = nbuffer + g_->nprocs;

    p = g_->L2Gmap_edge;
    g_->L2Gmap_edge = phgRealloc(g_->L2Gmap_edge, (nedge_save + local_count +
			edges_shared_count) * sizeof(*(g_->L2Gmap_edge)));
    if (p == NULL) {
	for (i = 0; i < nedge_save; i++)
	    g_->L2Gmap_edge[i] = i;
    }

    tmp[0] = local_count;
    tmp[1] = edges_shared_count;
    MPI_Allgather(tmp, 2, MPI_INT, nbuffer, 2, MPI_INT, g_->comm);

    edge_no_local = nedge_save;
    total_local = total_shared = 0;
    for (i = 0; i < g_->nprocs; i++) {
	if (i == g_->rank) edge_no_global = total_local + nedge_global_save;
	total_local += nbuffer[2 * i];
	total_shared += (counts[i] = nbuffer[2 * i + 1]);
    }

    j = 0;
    for (i = 0; i < g_->nprocs; i++) {
	displs[i] = j;
	j += counts[i];
    }

    /* update private edges */
    phgDebug((3, "private edges range: %d-%d\n", edge_no_global,
		edge_no_global + local_count - 1));
    i = 0;
    while (i < edges_local_count) {
	if (edges_local[index[i]].shared) {
	    i++;
	    continue;
	}
	j = i;
	while (++j < edges_local_count
		&& comp_edge0(index + i, index + j) == 0);
	while (i < j)
	    update_edge_no(index[i++], edge_no_local);
	g_->L2Gmap_edge[edge_no_local++] = edge_no_global++;
    }

    if (index != NULL)
	phgFree(index);

    edge_no_global = nedge_global_save + total_local;

    if (total_shared == 0)
	goto end;

    /* update shared edges */
    edges_gathered = phgAlloc(total_shared * sizeof(*edges_gathered));
    index = phgAlloc(total_shared * sizeof(*index));
    MPI_Type_contiguous(sizeof(*edges_gathered), MPI_BYTE, &type);
    MPI_Type_commit(&type);
    MPI_Allgatherv(edges_shared, edges_shared_count, type,
		   edges_gathered, counts, displs, type, g_->comm);
    MPI_Type_free(&type);
    for (i = 0; i < total_shared; i++)
	index[i] = i;
    qsort(index, total_shared, sizeof(*index), comp_edge_index);

    phgDebug((3, "update shared edges\n"));
    i = 0;
    flag = 0;
    while (i < total_shared) {
	j = index[i];
	if (j >= displs[g_->rank] && j < displs[g_->rank] + counts[g_->rank]) {
	    flag = 1;
	    update_edge_no(edges_gathered[j].index, edge_no_local);
	}
	j = i++;
	if (i >= total_shared || comp_edge_index(index + i, index + j) > 0) {
	    if (flag) {
		g_->L2Gmap_edge[edge_no_local++] = edge_no_global;
		flag = 0;
	    }
	    edge_no_global++;
	}
    }

    phgFree(index);
    phgFree(edges_gathered);

end:
    g_->nedge = edge_no_local;
    g_->nedge_global = edge_no_global;
    phgDebug((3, "shared edges range: %d-%d\n", nedge_global_save + total_local,
		edge_no_global - 1));

    phgFree(nbuffer);

    edges_local_count = edges_shared_count = 0;
    if (edges_local_allocated > 0) {
	phgFree(edges_local);
	edges_local = NULL;
	edges_local_allocated = 0;
    }
    if (edges_shared_allocated > 0) {
	phgFree(edges_shared);
	edges_shared = NULL;
	edges_shared_allocated = 0;
    }
}

#endif	/* USE_MPI */

static BOOLEAN
update_edges_callback CB_ARGS(e)
{
    int i, j, k, count;
    SIMPLEX **p, *e1;
    BOOLEAN flag;	/* indicate whether a shared edge */

    /* loop on all edges of the element */
    for (i = 0; i < NEdge; i++) {
	int v0 = GetEdgeVertex(i, 0), v1 = GetEdgeVertex(i, 1);
	INT g0, g1;	/* global indices of the two vertices */
	if (e->edges[i] != -1)
	    continue;
	/* new edge, assign it to all elements sharing it */
	g0 = e->verts[v0];
	g1 = e->verts[v1];
	/* loop on elements in the patch */
	count = phgGetPatch(g_, e, v0, v1, &p);
	flag = FALSE;
	for (j = 0; j < count; j++) {
#if USE_MPI
	    int u0 = 0, u1 = 0;
#endif
	    /* loop on edges of p[j] to find the one matching current edge */
	    e1 = p[j];
	    for (k = 0; k < NEdge; k++) {
		INT h0, h1;
		if (e1->edges[k] != -1) continue;
		h0 = e1->verts[GetEdgeVertex(k, 0)];
		h1 = e1->verts[GetEdgeVertex(k, 1)];
		if ((h0 == g0 && h1 == g1) || (h0 == g1 && h1 == g0)) break;
	    }
	    if (k >= NEdge) {
		phgDumpElement(g_, e1);
		phgInfo(-1, "common edge: %d %d\n", GlobalVertex(g_, g0),
			GlobalVertex(g_, g1));
		phgError(1, "%s:%d: unexpected error.\n", __FILE__, __LINE__);
	    }
	    e1->edges[k] = g_->nedge;
#if USE_MPI
	    if (g_->nprocs > 1) {
		/* determine whether it's an interface edge */
		switch (k) {
		    case 0:	/* 0-1 */
			u0 = 2; u1 = 3;
			break;
		    case 1:	/* 0-2 */
			u0 = 1; u1 = 3;
			break;
		    case 2:	/* 0-3 */
			u0 = 1; u1 = 2;
			break;
		    case 3:	/* 1-2 */
			u0 = 0; u1 = 3;
			break;
		    case 4:	/* 1-3 */
			u0 = 0; u1 = 2;
			break;
		    case 5:	/* 2-3 */
			u0 = 0; u1 = 1;
			break;
		}
		if ((e1->bound_type[u0] | e1->bound_type[u1]) & REMOTE)
		    flag = TRUE;
	    }
#endif
	}
	phgFree(p);
	g_->nedge++;
#if USE_MPI
	if (g_->nprocs > 1) {	/* no discontinuous patches if undistributed */
	    add_local_edge(e, v0, v1, flag);
	}
#endif
    }
    return TRUE;
}

void
phgUpdateEdges(GRID *g)
{
    FunctionEntry;

    if (g == NULL) return;

#if USE_MPI
    if (g->nprocs > 1) {
	nedge_global_save = g->nedge_global;
	nedge_save = g->nedge;
    }
#endif

    /* traverse all edges */
    phgDebug((3, "collecting edges\n"));
    g_ = g;
    phgTraverseElements(g, update_edges_callback);

#if USE_MPI
    phgDebug((3, "number of edges collected: %d\n", edges_local_count));
    if (g->nprocs > 1)
	sync_edges();
    else
#endif
    g->nedge_global = g->nedge;

    PrintTime(1);
    phgInfo(2, "number of edges = %d\n", g->nedge);

    Return;
}

#if Dim == 3
static INT faces_local, faces_shared, gindex;

static BOOLEAN
update_faces_callback0 CB_ARGS(e)
{
    int i;

    /* loop on all faces of the element */
    for (i = 0; i < NFace; i++) {
	if (e->faces[i] != -1) continue;
	if (HasLocalNeighbour(e, i)) {
	    SIMPLEX *e1 = e->neighbours[i];
	    if (e->verts[i] < e1->verts[phgOppositeVertex(e, i, e1)])
		faces_local++;
	}
	else if (!HasRemoteNeighbour(e, i)) {
	    faces_local++;
	}
	else {
	    faces_shared++;
	}
    }
    return TRUE;
}

static BOOLEAN
update_faces_callback1 CB_ARGS(e)
{
    int i, j;
    SIMPLEX *e1;

    /* loop on all faces of the element */
    for (i = 0; i < NFace; i++) {
	if (e->faces[i] != -1 || (e->bound_type[i] & REMOTE))
	    continue;
# if USE_MPI
	if (g_->L2Gmap_face != NULL)
	    g_->L2Gmap_face[g_->nface] = gindex;
#endif
	if (HasLocalNeighbour(e, i)) {
	    j = phgOppositeVertex(e, i, e1 = e->neighbours[i]);
	    if (e->verts[i] < e1->verts[j]) {
		e->faces[i] = e1->faces[j] = g_->nface++;
		gindex++;
	    }
	}
	else {
	    e->faces[i] = g_->nface++;
	    gindex++;
	}
    }
    return TRUE;
}

void
phgUpdateFaces(GRID *g)
{
    FunctionEntry;

    if (g == NULL) Return;
    g_ = g;

    /* count number of unassigned faces (those having index -1) */
    faces_local = faces_shared = 0;
    phgTraverseElements(g, update_faces_callback0);
    phgDebug((3, "faces: local=%d, shared=%d\n", faces_local, faces_shared));
    gindex = g->nface_global;
# if USE_MPI
    if (g->nprocs > 1) {
	int i, j, total_local, local0 = 0, total_shared;
	int *nbuffer, *counts, *displs, *counts0;
	struct {
	    SIMPLEX	*local, *remote;
	    INT		gindex;		/* global vertex index */
	    SHORT	rank;
	    BYTE	vertex, op_vertex;
	} *local, *remote;
	MPI_Datatype type;

	nbuffer = phgAlloc(3 * g->nprocs * sizeof(*nbuffer));
	counts = nbuffer;
	displs = nbuffer + g->nprocs;
	counts0 = nbuffer + 2 * g->nprocs;	/* nbr of rfaces owned */
	local = (void *)g->L2Gmap_face;
	g->L2Gmap_face = phgRealloc(g->L2Gmap_face, sizeof(*(g->L2Gmap_face)) *
				(g->nface + faces_local + faces_shared));
	if (local == NULL) {
	    for (i = 0; i < g->nface; i++)
		g_->L2Gmap_face[i] = i;
	}
	/* gather counts of local faces */
	i = faces_local;
	MPI_Allgather(&i, 1, MPI_INT, counts, 1, MPI_INT, g->comm);
	total_local = 0;
	for (i = 0; i < g->nprocs; i++) {
	    if (i == g->rank) local0 = total_local;
	    total_local += counts[i];
	    counts[i] = 0;
	}
	local = phgAlloc(sizeof(*local) * faces_shared);
	remote = phgAlloc(sizeof(*remote) * faces_shared);
	/* copy unassigned shared faces and count the counts for each process */
	j = 0;
	for (i = 0; i < g->neighbours.count; i++) {
	    RNEIGHBOUR *rn = g->neighbours.list + i;
	    if ((rn->local)->faces[rn->vertex] >= 0) continue;
	    local[j].local = rn->local;
	    local[j].remote = rn->remote;
	    counts[local[j].rank = rn->rank]++;
	    local[j].gindex = GlobalVertex(g, (rn->local)->verts[rn->vertex]);
	    local[j].vertex = rn->vertex;
	    local[j].op_vertex = rn->op_vertex;
	    j++;
	}
	j = 0;
	for (i = 0; i < g->nprocs; i++) {
	    displs[i] = j;
	    j += counts[i];
	}
	MPI_Type_contiguous(sizeof(*local), MPI_BYTE, &type);
	MPI_Type_commit(&type);
	phgAlltoallv(local,  counts, displs, type,
		     remote, counts, displs, type, g->comm);
	/* get number of shared faces which should be assigned by the process */
	j = 0;
	for (i = 0; i < faces_shared; i++) {
	    if (GlobalVertex(g, remote[i].remote->verts[remote[i].op_vertex])
		< remote[i].gindex) j++;
	}
	MPI_Allgather(&j, 1, MPI_INT, counts0, 1, MPI_INT, g->comm);
	total_shared = 0;
	for (i = 0; i < g->nprocs; i++) {
	    if (i == g->rank) gindex += total_local + total_shared;
	    total_shared += counts0[i];
	}
	/* assign indices to shared faces */
	for (i = 0; i < faces_shared; i++) {
	    if (GlobalVertex(g, remote[i].remote->verts[remote[i].op_vertex])
		< remote[i].gindex) {
		if (g->L2Gmap_face != NULL) g->L2Gmap_face[g->nface] = gindex;
		remote[i].gindex = gindex;
		remote[i].remote->faces[remote[i].op_vertex] = g->nface++;
		gindex++;
	    }
	    else {
		remote[i].gindex = -1;
	    }
	}
	phgAlltoallv(remote, counts, displs, type,
		     local,  counts, displs, type, g->comm);
	MPI_Type_free(&type);
	for (i = 0; i < faces_shared; i++) {
	    if (local[i].gindex == -1) continue;
	    if (g->L2Gmap_face != NULL)
		g->L2Gmap_face[g->nface] = local[i].gindex;
	    local[i].local->faces[local[i].vertex] = g->nface++;
	}
	gindex = g->nface_global + local0;
	g->nface_global += total_local + total_shared;
	phgFree(remote);
	phgFree(local);
	phgFree(nbuffer);
    }
# endif

    /* assign local indices */
    phgTraverseElements(g, update_faces_callback1);

    if (g->nprocs <= 1)
	g->nface_global = g->nface;

    PrintTime(1);
    Return;
}
#endif	/* Dim == 3 */

void
phgUpdateElementIndices(GRID *g, INT nelem_old, INT nelem_global_old)
{
#if USE_MPI
    int i, index0, n;
#endif

    FunctionEntry;

#if USE_MPI
    if (g == NULL || g->nprocs <= 1) Return;
    g_ = g;

    n = g->nelem - nelem_old;
    MPI_Scan(&n, &index0, 1, MPI_INT, MPI_SUM, g->comm);
    index0 += nelem_global_old - n;

    if (g->L2Gmap_elem == NULL) {
	g->L2Gmap_elem = phgAlloc(g->nelem * sizeof(*g->L2Gmap_elem));
	for (i = 0; i < nelem_old; i++) g->L2Gmap_elem[i] = i;
    }
    else {
	g->L2Gmap_elem = phgRealloc(g->L2Gmap_elem,
					g->nelem * sizeof(*g->L2Gmap_elem));
    }

    /* assign global indices */
    for (i = 0; i < n; i++)
	g->L2Gmap_elem[nelem_old + i] = index0 + i;

    PrintTime(1);
#else	/* USE_MPI */
    (void)g;
    (void)nelem_old;
    (void)nelem_global_old;
#endif	/* USE_MPI */

    Return;
}

void
phgDumpElement(GRID *g, SIMPLEX *e)
{
    int i;
    char s[128];
#ifdef VL_
# undef VL_
#endif
#define VL_(i)	(e->verts[i])
#ifdef EL_
# undef EL_
#endif
#define EL_(i)	(e->edges[i])
#ifdef FL_
# undef FL_
#endif
#define FL_(i)	(e->faces[i])
#if USE_MPI
    RNEIGHBOUR *rn;
# ifdef VG_
#  undef VG_
# endif
# define VG_(i)	GlobalVertex(g, e->verts[i])
# ifdef EG_
#  undef EG_
# endif
# define EG_(i)	GlobalEdge(g, e->edges[i])
# ifdef FG_
#  undef FG_
# endif
# define FG_(i)	GlobalFace(g, e->faces[i])
#endif

    if (e == NULL) {
	phgWarning("phgDumpElement: e == NULL\n", e);
	return;
    }

#if Dim == 2
    phgInfo(-1, "E(%d): %p, %s, mark=%d, bdry=%x:%x:%x, " "gen=%d\n",
		g->elem_flag ? GlobalElement(g, e->index) : -1,
		e, GTypeName(e->type), e->mark,
		e->bound_type[0], e->bound_type[1], e->bound_type[2],
		e->generation);
#if USE_MPI
    phgInfo(-1, "  children=%p %p, parent=%p, flag=%d\n",
		e->children[0], e->children[1], e->parent, e->flag);
    phgInfo(-1, "  verts=%d %d %d (%d %d %d)\n",
		VL_(0), VL_(1), VL_(2), VG_(0), VG_(1), VG_(2));
    phgInfo(-1, "  edges=%d %d %d (%d %d %d)\n",
		EL_(0), EL_(1), EL_(2), EG_(0), EG_(1), EG_(2));
#else
    phgInfo(-1, "  children=%p %p, parent=%p\n",
		e->children[0], e->children[1]);
    phgInfo(-1, "  verts=%d %d %d\n", VL_(0), VL_(1), VL_(2));
    phgInfo(-1, "  edges=%d %d %d\n", e->edges[0], e->edges[1], e->edges[2]);
#endif
#else	/* Dim */
    phgInfo(-1, "E(%d): %p, %s, mark=%d, bdry=%x:%x:%x:%x, " "gen=%d\n",
		g->elem_flag ? GlobalElement(g, e->index) : -1,
		e, GTypeName(e->type), e->mark,
		e->bound_type[0], e->bound_type[1], e->bound_type[2],
		e->bound_type[3], e->generation);
#if USE_MPI
    phgInfo(-1, "  children=%p %p, parent=%p, flag=%d\n",
		e->children[0], e->children[1], e->parent, e->flag);
    phgInfo(-1, "  verts=%d %d %d %d (%d %d %d %d)\n",
		VL_(0), VL_(1), VL_(2), VL_(3), VG_(0), VG_(1), VG_(2), VG_(3));
    phgInfo(-1, "  edges=%d %d %d %d %d %d (%d %d %d %d %d %d)\n",
		EL_(0), EL_(1), EL_(2), EL_(3), EL_(4), EL_(5),
		EG_(0), EG_(1), EG_(2), EG_(3), EG_(4), EG_(5));
    phgInfo(-1, "  faces=%d %d %d %d (%d %d %d %d)\n",
		FL_(0), FL_(1), FL_(2), FL_(3),
		FG_(0), FG_(1), FG_(2), FG_(3));
#else
    phgInfo(-1, "  children=%p %p, parent=%p\n",
		e->children[0], e->children[1]);
    phgInfo(-1, "  verts=%d %d %d %d\n", VL_(0), VL_(1), VL_(2), VL_(3));
    phgInfo(-1, "  edges=%d %d %d %d %d %d\n", e->edges[0], e->edges[1],
		e->edges[2], e->edges[3], e->edges[4], e->edges[5]);
    phgInfo(-1, "  faces=%d %d %d %d\n", e->faces[0], e->faces[1],
		e->faces[2], e->faces[3]);
#endif
#endif	/* Dim */
#undef V_
    s[0] = '\0';
    for (i = 0; i < NFace; i++) {
	if (HasLocalNeighbour(e, i))
	    sprintf(s+strlen(s), "%p ", e->neighbours[i]);
	else if ((e->bound_type[i] & DIRICHLET))
	    sprintf(s+strlen(s), "D ");
	else if ((e->bound_type[i] & NEUMANN))
	    sprintf(s+strlen(s), "N ");
	else if ((e->bound_type[i] & UNDEFINED))
	    sprintf(s+strlen(s), "U ");
#if USE_MPI
	else if ((e->bound_type[i] & REMOTE) && IsLeaf(e)) {
	    if (g != NULL && (rn = GetRNeighbour(g, e, i)) != NULL)
		sprintf(s+strlen(s), "%p@p%d ", rn->remote, rn->rank);
	    else
		sprintf(s+strlen(s), "unset ");
	}
#endif
	else
	    sprintf(s+strlen(s), "undef ");
    }
    phgInfo(-1, "  neighbours=%s\n", s);

    if (phgVerbosity < 2) return;

#if USE_MPI
    if (IsLeaf(e))
      for (i = 0; i < NFace; i++) {
	if (!(e->bound_type[i] & REMOTE) || g == NULL ||
		(rn = GetRNeighbour(g,e,i)) == NULL)
	    continue;
	phgInfo(-1, "  RNeighbour%d: vertex=%d, op_vertex=%d, face=%d %d %d\n",
			i, rn->vertex, rn->op_vertex, 
			rn->rface[0], rn->rface[1], rn->rface[2]);
    }
#endif
}

void
phgDumpPatch(GRID *g, SIMPLEX *e, int v0, int v1)
/* dumps out a patch */
{
    int j, n;
    SIMPLEX **p;

    phgInfo(-1, "========= the patch:\n");
    n = phgGetPatch(g, e, v0, v1, &p);
    for (j = 0; j < n; j++) phgDumpElement(g, p[j]);
    phgFree(p);
    phgInfo(-1, "=========\n");
}

static BOOLEAN
dump_callback CB_ARGS(e)
{
    phgDumpElement(g_, e);
    return TRUE;
}

static void
dump_head(GRID *g)
{
    phgInfo(-1, "vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv\n");
    phgInfo(-1, "nroot = %d\n", g->nroot);
    phgInfo(-1, "nleaf = %d\n", g->nleaf);
    phgInfo(-1, "ntree = %d\n", g->ntree);
    phgInfo(-1, "nvert = %d\n", g->nvert);
    phgInfo(-1, "nedge = %d\n", g->nedge);
#if Dim == 3
    phgInfo(-1, "nface = %d\n", g->nface);
#endif
    phgInfo(-1, "nelem = %d\n", g->nelem);
    /*if (g->nprocs > 1)*/ {
	phgInfo(-1, "nvert_global = %d\n", g->nvert_global);
	phgInfo(-1, "nedge_global = %d\n", g->nedge_global);
#if Dim == 3
	phgInfo(-1, "nface_global = %d\n", g->nface_global);
#endif
	phgInfo(-1, "nelem_global = %d\n", g->nelem_global);
    }
#if USE_MPI
    phgInfo(-1, "neighbours.count = %d\n", g->neighbours.count);
#endif
}

static void
dump_tail(GRID *g)
{
    Unused(g);
    phgInfo(-1, "^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n");
}

void
phgDumpGridInfo(GRID *g)
{
    ParallelFunction(phgDumpGridInfo, g != NULL && g->nprocs > 1);
    if (g == NULL) return;
    dump_head(g_ = g);
    dump_tail(g);
}

void
phgDumpGrid(GRID *g)
{
    ParallelFunction(phgDumpGrid, g != NULL && g->nprocs > 1);
    if (g == NULL) return;
    dump_head(g_ = g);
    phgTraverseElements(g, dump_callback);
    dump_tail(g);
}

void
phgDumpTree(GRID *g)
{
    ParallelFunction(phgDumpTree, g != NULL && g->nprocs > 1);
    if (g == NULL) return;
    dump_head(g_ = g);
    phgTraverseAllElements(g, dump_callback);
    dump_tail(g);
}

void
phgDumpBranch(GRID *g, SIMPLEX *e)
{
    phgDumpElement(g, e);
    if (e->children[0] != NULL)
	phgDumpBranch(g, e->children[0]);
    if (e->children[1] != NULL)
	phgDumpBranch(g, e->children[1]);
}

#if defined(GZIP_PROG) || defined(BZIP2_PROG)
#define MAX_FILES 16
static struct {
    FILE *fp;
    BOOLEAN is_pipe;
} open_files[MAX_FILES];
#endif

void
phgCloseInputFile_(FILE *fp)
{
#if defined(GZIP_PROG) || defined(BZIP2_PROG)
    int i;
    BOOLEAN is_pipe = FALSE;
    for (i = 0; i < MAX_FILES; i++) {
	if (open_files[i].fp == fp) {
	    is_pipe = open_files[i].is_pipe;
	    open_files[i].fp = NULL;
	    break;
	}
    }
    (is_pipe ? pclose : fclose)(fp);
#else
    if (fp != NULL) fclose(fp);
#endif
}

FILE *
phgOpenInputFile_(const char *filename)
{
    FILE *fp;
#if defined(GZIP_PROG) || defined(BZIP2_PROG)
    char cmd[PATH_MAX + 16];
    int len = strlen(filename);
    BOOLEAN is_pipe = FALSE;
    static BOOLEAN initialized = FALSE;

    if (!initialized) {
	int i;
	initialized = TRUE;
	for (i = 0; i < MAX_FILES; i++) open_files[i].fp = NULL;
    }
#endif

#ifdef GZIP_PROG
    if (len > 3 && !phgMemcmp(filename + len - 3, ".gz", 3)) {
	if ((fp = fopen(filename, "r")) != NULL) {
	    fclose(fp);
	    phgInfo(1, "gzipped file.\n");
	    sprintf(cmd, GZIP_PROG " -dc %s 2>/dev/null", filename);
	    fp = popen(cmd, "r");
	    is_pipe = TRUE;
	}
    } else
#endif
#ifdef BZIP2_PROG
    if (len > 4 && !phgMemcmp(filename + len - 4, ".bz2", 4)) {
	if ((fp = fopen(filename, "r")) != NULL) {
	    fclose(fp);
	    phgInfo(1, "bzip2 file.\n");
	    sprintf(cmd, BZIP2_PROG " -dc %s 2>/dev/null", filename);
	    fp = popen(cmd, "r");
	    is_pipe = TRUE;
	}
    } else
#endif
    {
	fp = fopen(filename, "r");
    }
    if (fp == NULL)
	phgWarning("can't open file \"%s\"!\n", filename);
#if defined(GZIP_PROG) || defined(BZIP2_PROG)
    else {
	int i;
	for (i = 0; i < MAX_FILES; i++)
	    if (open_files[i].fp == NULL) break;
	if (i >= MAX_FILES) {
	    phgWarning("too many open files.\n");
	    (is_pipe ? pclose : fclose)(fp);
	    fp = NULL;
	}
	else {
	    open_files[i].fp = fp;
	    open_files[i].is_pipe = is_pipe;
	}
    }
#endif

    return fp;
}

BDRY_MAP_FUNC phgBdryMapFunc_ = NULL;

void
phgImportSetBdryMapFunc(BDRY_MAP_FUNC func)
{
    phgBdryMapFunc_ = func;
}

static int btype_index = -1;

static const char *btype_names[] = {
    "dirichlet", "neumann", "user1", "user2", "undefined", NULL
};
static BTYPE btype_values[] = {
    DIRICHLET, NEUMANN, BDRY_USER1, BDRY_USER2, UNDEFINED
};

BTYPE
phgImportSetDefaultBdryType(BTYPE type)
{
    BTYPE t, oldtype = (btype_index<0 ? UNDEFINED:btype_values[btype_index]);
    int i;

    btype_index = -1;
    t = type & BDRY_MASK;
    for (i = 0; i < sizeof(btype_values) / sizeof(btype_values[0]); i++) {
	if (t == btype_values[i]) {
	    btype_index = i;
	    break;
	}
    }

    if (btype_index == -1)
	phgError(1, "%s: invalid boundary type %d.\n", __func__, type);

    return oldtype;
}

static BOOLEAN
default_bdry_callback CB_ARGS(e)
{
    int i;

    for (i = 0; i < NFace; i++) {
	if (e->bound_type[i] == UNDEFINED) {
	    e->bound_type[i] &= ~BDRY_MASK;
	    e->bound_type[i] |= btype_values[btype_index];
	}
    }

    return TRUE;
}

static void
set_default_bdry_type(GRID *g)
/* sets UNDEFINED boundary type to default type */
{
    if (btype_values[btype_index] == UNDEFINED)
	return;

    /* Note: can't use ForAllElements because g->elems is not up to date */
    phgTraverseElements(g, default_bdry_callback);
}

BOOLEAN
phgImport(GRID *g, const char *filename, BOOLEAN distr)
{
    FILE *fp;
    char line[128], magic[128];
    BOOLEAN ret = FALSE;
    int i, flag;
    static BOOLEAN initialized = FALSE;

    if (!initialized) {
	initialized = TRUE;
	phgImportSetDefaultBdryType(UNDEFINED);
	phgOptionsRegisterKeyword("-default_bdry_type", "Default boundary type",
		btype_names, &btype_index);
	return TRUE;
    }

    if (g == NULL)
	phgError(1, "invalid grid pointer.\n");

    if (g->nleaf_global > 0) {
	phgWarning("g already contains a valid grid, new grid not imported.\n");
	return ret;
    }

    if (phgRank > 0)
	goto hamilton;

    if (phgVerbosity > 0)
	phgPrintf("Default boundary type: %s.\n",
			BTypeName(btype_values[btype_index]));

    if ((fp = phgOpenInputFile_(filename)) == NULL) {
	phgError(1, "cannot read file \"%s\".\n", filename);
	goto hamilton;
    }

    if (fgets(line, sizeof(line), fp) == NULL ||
	sscanf(line, "%s", magic) != 1) {
	phgCloseInputFile_(fp);
	goto hamilton;
    }
    phgCloseInputFile_(fp);

    /* Note:
     *	flag < 0:	input error,
     *	flag == 0: 	the elements don't have refinement types,
     *	flag > 0:  	the elements already have refinement types */
    if (!strncmp(magic, "DIM:", 4) || !strncmp(magic, "DIM_OF_WORLD:", 13))
	flag = phgImportALBERT(g, filename, FALSE);
    else if (!strcmp(magic, "MeshVersionFormatted"))
	flag = phgImportMedit(g, filename, FALSE);
    else {
	flag = -1;
	phgWarning("can't determine type of input file, mesh not imported.\n");
    }

    ret = (flag >= 0);

    if (ret) {
	if (phgRank == 0)
	    phgInfo(1, "%d elements loaded.\n", g->nleaf);\

	phgUpdateNeighbours(g);
	phgRemoveDegeneratedElements(g);
	phgUpdateEdges(g);
#if Dim == 3
	phgUpdateFaces(g);
#endif
	phgCheckConformity(g);

#if ALLOW_CURVED_BOUNDARY
	phgInitBoundaryFunctions(g);
#endif

	/* compute BoundingBox of the domain */
	if (g->nvert == 0) {
	    for (i = 0; i < Dim; i++) {
		g->bbox[0][i] = g->bbox[1][i] = 0.;
	    }
	}
	else {
	    int j;
	    FLOAT f;

	    for (j = 0; j < Dim; j++) {
		g->bbox[0][j] = g->bbox[1][j] = g->verts[0][j];
	    }
	    for (i = 1; i < g->nvert; i++) {
		COORD *c = g->verts + i;
		for (j = 0; j < Dim; j++) {
		if ((f = (*c)[j]) < g->bbox[0][j]) g->bbox[0][j] = f;
		if (f > g->bbox[1][j]) g->bbox[1][j] = f;
		}
	    }
	}

	phgRefineInit(g, !flag);

hamilton:
#if USE_MPI
	if (phgNProcs > 1) {
	    if (phgRank == 0) {
		/* compute hamiltonian path and reorder the root elements.
		 * if fails, then avoid to use RTK repartitioner */
		flag = phgHamiltonPath(g, FALSE);
		if (flag != 0) {
        	    phgWarning("Error in computing the Hamilton path,\n");
        	    phgWarning("Hamilton path not created.\n");
		}
	    }
	    if (!phgMasterSlave)
		MPI_Bcast(&flag, 1, MPI_INT, 0, MPI_COMM_WORLD);
	    if (flag != 0) {
		if (!strcmp(phgOptionsGetKeyword("-partitioner"), "rtk")) {
#if USE_PARMETIS
		    if (!phgMasterSlave) {
			if (phgRank == 0)
        		    phgWarning("switching to ParMETIS partitioner.\n");
			phgOptionsSetKeyword("-partitioner", "metis");
		    }
#elif USE_ZOLTAN
		    if (!phgMasterSlave) {
			if (phgRank == 0)
        		    phgWarning("switching to Zoltan partitioner.\n");
			phgOptionsSetKeyword("-partitioner", "zoltan");
		    }
#else
		    if (phgRank == 0)
			phgWarning("poor partitioning results expected.\n");
#endif
		}
            }
	}
	if (phgRank > 0)
	    goto end;
#endif	/* USE_MPI */

	/* phgUpdateBoundaryTypes might have been called in RefineInit
	 * if pre-refinement is enabled */
	if ((g->vert_flag || g->edge_flag || g->face_flag || g->elem_flag)
	    && g->types_vert == NULL)
	    phgUpdateBoundaryTypes(g);
    }

#if USE_MPI
end:
    if (!phgMasterSlave && phgNProcs > 1) {
	MPI_Bcast(&ret, sizeof(ret), MPI_BYTE, 0, MPI_COMM_WORLD);
	if (distr) /* force grid distribution */
	    phgRedistributeGrid(g);
    }
#endif

    if (g->geom_flag)
	phgGeomInit(g);

    if (ret)
	g->serial_no++;

    return ret;
}

#if ALLOW_CURVED_BOUNDARY

void phgInitBoundaryFunctions(GRID *g)
{
    EXPR **funcs;
    BYTE *flags;
    SIMPLEX *e;
    int i, j;

    if ((funcs = g->bdry_funcs) == NULL)
	return;

    flags = phgCalloc(g->nvert, sizeof(*flags));

    /* mark boundary vertices (bit 0) */
    for (i = 0; i < g->nroot; i++) {
	e = g->roots + i;
	for (j = 0; j < NFace; j++) {
	    if (e->bound_type[j] & BDRY_MASK) {
		int k;
		for (k = 0; k < NFace; k++)
		    if (k != j)
			flags[e->verts[k]] = 1;
	    }
	}
    }

    while (*funcs != NULL) {
	for (j = 0; j < g->nvert; j++) {
	    COORD *c;
	    if (!flags[j])
		continue;
	    c = g->verts + j;
	    if (fabs(phgEvaluate3DFunction(*funcs, (*c)[0], (*c)[1], (*c)[2],
			0.0)) < 1e-4)
		flags[j] |= 2;
	    else
		flags[j] &= 1;
	}
	for (i = 0; i < g->nroot; i++) {
	    e = g->roots + i;
	    for (j = 0; j < NEdge; j++) {
		if (e->bound_func[j] != -1)
		    continue;
		if ((flags[e->verts[GetEdgeVertex(j, 0)]] & 2) != 0 &&
		    (flags[e->verts[GetEdgeVertex(j, 1)]] & 2) != 0)
		    e->bound_func[j] = (funcs - g->bdry_funcs) / 4;
	    }
	}
	funcs += 4;
    }

    phgFree(flags);
}

FLOAT
phgGetBoundaryError(GRID *g, SIMPLEX *e)
/* returns the maximum square of the distances between the middle points of
 * the edges and their projections onto curved boundary */
{
    EXPR **f;
    COORD *v0, *v1;
    FLOAT d, dmax, x, y, z, dx, dy, dz;
    int edge;

    dmax = 0.0;
    for (edge = 0; edge < NEdge; edge++) {
	if (e->bound_func[edge] == -1)
	    continue;
    
	f = g->bdry_funcs + e->bound_func[edge] * 4;
	v0 = g->verts + e->verts[GetEdgeVertex(edge, 0)];
	v1 = g->verts + e->verts[GetEdgeVertex(edge, 1)];
	x = ((*v0)[0] + (*v1)[0]) * .5;
	y = ((*v0)[1] + (*v1)[1]) * .5;
	z = ((*v0)[2] + (*v1)[2]) * .5;
	dx = phgEvaluate3DFunction(f[1], x, y, z, 0.0) - x;
	dy = phgEvaluate3DFunction(f[2], x, y, z, 0.0) - y;
	dz = phgEvaluate3DFunction(f[3], x, y, z, 0.0) - z;
	/* FIXME: sum or max? */
	d = dx * dx + dy * dy + dz * dz;
	if (dmax < d)
	    dmax = d;
    }

    return dmax;
}

#else	/* ALLOW_CURVED_BOUNDARY */

FLOAT
phgGetBoundaryError(GRID *g, SIMPLEX *e)
{
    Unused(g);
    Unused(e);
    return 0.0;
}

#endif	/* ALLOW_CURVED_BOUNDARY */

#if USE_MPI

static int
comp_rneighbour(const void *p1, const void *p2)
/* Note: the function is made global since it's also used by refine.c */
{
    int i;
    const RNEIGHBOUR *ptr1 = p1, *ptr2 = p2;

#if 0	/* no longer necessary */
    if (ptr1->local->children[0] != NULL || ptr2->local->children[0] != NULL) {
	/* move non-leaf element(s) to the end of list */
	if (ptr1->local->children[0] == NULL) return -1;
	else if (ptr2->local->children[0] == NULL) return 1;
	else return 0;
    }
#endif

    if ((i = ptr1->rank - ptr2->rank) != 0) return i;
    if (ptr1->local != ptr2->local) return (ptr1->local < ptr2->local) ? -1 : 1;
    return ptr1->vertex - ptr2->vertex;
}

INT
phgCountRNeighbours(int nprocs, RNEIGHBOUR *list, INT count, int *counts,
			int *displs)
/* counts number of elements for different processes, and removes entries on
   non leaf elements */
{
    int i, j;
    RNEIGHBOUR *p0, *p1;
    SIMPLEX *e;

    if (counts == NULL) return count;

    /* First, remove entries for non leaf elements */
    p0 = list;
    while (p0 < list + count) {
	e = p0->local;
	if (!IsLeaf(e))
	    break;
	p0++;
    }
    p1 = p0;
    while (p1 < list + count) {
	do {
	    p1->local->neighbours[p1->vertex] = (void *)-1;
	} while (++p1 < list + count && ((e = p1->local)->children[0] != NULL
		|| e->children[1] != NULL));
	do {
	    *(p0++) = *(p1++);
	} while (p1 < list + count && (e = p1->local)->children[0] == NULL
		&& e->children[1] == NULL);
    }
    count = p0 - list;

    /* Then sort the list */
    if (count > 0) qsort(list, count, sizeof(RNEIGHBOUR), comp_rneighbour);

    /* Finally, count number of entries for each process */
    p0 = list;
    j = 0;
    for (i = 0; i < nprocs; i++) {
	while (p0 - list < count && p0->rank < i /*&& 
		p0->local->children[0] == NULL*/) p0++;
	p1 = p0;
	while (p1 - list < count && p1->rank == i /*&&
		p1->local->children[0] == NULL*/) p1++;
	counts[i] = p1 - p0;
	displs[i] = j;
	j += p1 - p0;
	p0 = p1;
    }

    return j;
}

#endif	/* USE_MPI */
